# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 29.3 `tsa` Chapter.

``tsa`` - Time Series Analysis

This module provides facilities for investigating and modelling the
statistical structure of series of observations collected at points in
time.
The models may then be used to forecast the series.

The module covers the following models and approaches:

(1) univariate time series analysis, including autocorrelation functions
    and autoregressive moving average (ARMA) models;

(#) univariate spectral analysis;

(#) transfer function (multi-input) modelling, in which one time series
    is dependent on other time series;

(#) bivariate spectral methods including coherency, gain and input
    response functions;

(#) vector ARMA models for multivariate time series;

(#) Kalman filter models (linear and nonlinear);

(#) GARCH models for volatility;

(#) inhomogeneous time series.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13intro.html

See Also
--------
``naginterfaces.library.examples.tsa`` :
    This subpackage contains examples for the ``tsa`` module.
    See also the :ref:`library_tsa_ex` subsection.
"""

# NAG Copyright 2017-2023.

from __future__ import absolute_import

import collections as _collections

from math import ceil
from math import log
from numpy import ascontiguousarray as _ascontiguousarray

from ..base import tsa as _b_tsa
from ..base import utils as _b_utils
from ..library import _utils

UniDiffReturnData = _collections.namedtuple(
    'UniDiffReturnData',
    ['xd', 'nxd'],
)

def uni_diff(x, nd, nds, ns):
    u"""
    Univariate time series, seasonal and non-seasonal differencing.

    ``uni_diff`` carries out non-seasonal and seasonal differencing on a
    time series.
    Information which allows the original series to be reconstituted
    from the differenced series is also produced.
    This information is required in time series forecasting.

    For full information please refer to the NAG Library document for
    g13aa

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13aaf.html

    Parameters
    ----------
    x : float, array-like, shape (nx)
        The undifferenced time series, x_i, for i = 1,2,...,n.

    nd : int
        d, the order of non-seasonal differencing.

    nds : int
        D, the order of seasonal differencing.

    ns : int
        s, the seasonality.

    Returns
    -------
    xd : float, ndarray, shape (nx)
        The differenced values in elements 0 to `nxd`-1, and
        reconstitution data in the remainder of the array.

    nxd : int
        The number of differenced values in the array `xd`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nds` = *<value>*.

            Constraint: if `ns` = 0, `nds` <= 0.

        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 0.

        (`errno` 1)
            On entry, `nds` = *<value>*.

            Constraint: `nds` >= 0.

        (`errno` 1)
            On entry, `nd` = *<value>*.

            Constraint: `nd` >= 0.

        (`errno` 2)
            On entry, nx = *<value>*, `nd` = *<value>*, `nds` =
            *<value>* and `ns` = *<value>*.

            Constraint: nx > `nd`+(`nds`*`ns`).

    Notes
    -----
    Let nabla^d nabla_s^Dx_i be the ith value of a time series x_i, for
    i = 1,2,...,n after non-seasonal differencing of order d and
    seasonal differencing of order D (with period or seasonality s).
    In general,

    [table omitted]

    Non-seasonal differencing up to the required order d is obtained
    using

    [table omitted]

    Seasonal differencing up to the required order D is then obtained
    using

    [table omitted]

    Mathematically, the sequence in which the differencing operations
    are performed does not affect the final resulting series of m =
    n-d-D*s values.
    """

    fun_name = 'naginterfaces.library.tsa.uni_diff'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _xd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nx_py,),
    )
    _return_scalar = _b_tsa.uni_diff(
        _x,
        _nx,
        nd,
        nds,
        ns,
        _xd,
    )
    _xd_py = _xd
    _nxd_py = _return_scalar
    _return_data = UniDiffReturnData(
        _xd_py,
        _nxd_py,
    )
    return _return_data

UniAutocorrReturnData = _collections.namedtuple(
    'UniAutocorrReturnData',
    ['xm', 'xv', 'r', 'stat'],
)

def uni_autocorr(x, nk):
    u"""
    Univariate time series, sample autocorrelation function.

    ``uni_autocorr`` computes the sample autocorrelation function of a
    time series.
    It also computes the sample mean, the sample variance and a
    statistic which may be used to test the hypothesis that the true
    autocorrelation function is zero.

    For full information please refer to the NAG Library document for
    g13ab

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13abf.html

    Parameters
    ----------
    x : float, array-like, shape (nx)
        The time series, x_i, for i = 1,2,...,n.

    nk : int
        K, the number of lags for which the autocorrelations are
        required. The lags range from 1 to K and do not include zero.

    Returns
    -------
    xm : float
        The sample mean of the input time series.

    xv : float
        The sample variance of the input time series.

    r : float, ndarray, shape (`nk`)
        The sample autocorrelation coefficient relating to lag k, for k
        = 1,2,...,K.

    stat : float
        The statistic used to test the hypothesis that the true
        autocorrelation function of the time series is identically zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nx = *<value>* and `nk` = *<value>*.

            Constraint: nx > `nk`.

        (`errno` 1)
            On entry, nx = *<value>*.

            Constraint: nx > 1.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, all values of `x` are practically identical.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The data consists of n observations x_i, for i = 1,2,...,n from a
    time series.

    The quantities calculated are

    (a) The sample mean

            x_= sum_i = 1^nx_i/n.

    (#) The sample variance (for n >= 2)

            s^2 = sum_i = 1^n(x_i-x_)^2/((n-1)).

    (#) The sample autocorrelation coefficients of lags k = 1,2,...,K,
        where K is a user-specified maximum lag, and K < n, n > 1.

        The coefficient of lag k is defined as

            r_k = sum_i = 1^n-k(x_i-x_)(x_{i+k}-x_)/sum_i =
            1^n(x_i-x_)^2.

        See page 496 of Box and Jenkins (1976) for further details.

    (#) A test statistic defined as

            `stat` = nsum_k = 1^Kr_k^2,

        which can be used to test the hypothesis that the true
        autocorrelation function is identically zero.

        If n is large and K is much smaller than n, `stat` has a chi_K^2
        distribution under the hypothesis of a zero autocorrelation
        function.
        Values of `stat` in the upper tail of the distribution provide
        evidence against the hypothesis; ``stat.prob_chisq`` can be used
        to compute the tail probability.

        Section 8.2.2 of Box and Jenkins (1976) provides further details
        of the use of `stat`.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.uni_autocorr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _nk = _b_utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = _b_utils._EngineIntScalarType.to_py(_nk)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nk_py,),
    )
    try:
        _return_scalars = _b_tsa.uni_autocorr(
            _x,
            _nx,
            _nk,
            _r,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xm_py = _return_scalars.xm
    _xv_py = _return_scalars.xv
    _r_py = _r
    _stat_py = _return_scalars.stat
    _return_data = UniAutocorrReturnData(
        _xm_py,
        _xv_py,
        _r_py,
        _stat_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniAutocorrPartReturnData = _collections.namedtuple(
    'UniAutocorrPartReturnData',
    ['p', 'v', 'ar', 'nvl'],
)

def uni_autocorr_part(r, nl):
    u"""
    Univariate time series, partial autocorrelations from
    autocorrelations.

    ``uni_autocorr_part`` calculates partial autocorrelation
    coefficients given a set of autocorrelation coefficients.
    It also calculates the predictor error variance ratios for
    increasing order of finite lag autoregressive predictor, and the
    autoregressive parameters associated with the predictor of maximum
    order.

    For full information please refer to the NAG Library document for
    g13ac

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13acf.html

    Parameters
    ----------
    r : float, array-like, shape (nk)
        The autocorrelation coefficient relating to lag k, for k =
        1,2,...,K.

    nl : int
        L, the number of partial autocorrelation coefficients required.

    Returns
    -------
    p : float, ndarray, shape (`nl`)
        `p`[l-1] contains the partial autocorrelation coefficient at lag
        l, p_{l,l}, for l = 1,2,...,`nvl`.

    v : float, ndarray, shape (`nl`)
        `v`[l-1] contains the predictor error variance ratio v_l, for l
        = 1,2,...,`nvl`.

    ar : float, ndarray, shape (`nl`)
        The autoregressive parameters of maximum order, i.e., p_{Lj} if
        no exception or warning is raised, or p_{l_0-1,j} if `errno` =
        3, for j = 1,2,...,`nvl`.

    nvl : int
        The number of valid values in each of `p`, `v` and `ar`. Thus in
        the case of premature termination at iteration l_0 (see
        [Notes]), `nvl` is returned as l_0-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nk = *<value>* and `nl` = *<value>*.

            Constraint: nk >= `nl`.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` > 0.

        (`errno` 1)
            On entry, nk = *<value>*.

            Constraint: nk > 0.

        (`errno` 2)
            On entry, the autocorrelation coefficient of lag 1 has
            absolute value greater than or equal to 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The autocorrelation coefficients do not form a positive
            definite sequence.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The data consist of values of autocorrelation coefficients
    r_1,r_2,...,r_K, relating to lags 1,2,...,K.
    These will generally (but not necessarily) be sample values such as
    may be obtained from a time series x_t using ``uni_autocorr``.

    The partial autocorrelation coefficient at lag l may be identified
    with the parameter p_{l,l} in the autoregression

        x_t =
        c_l+p_{l,1}x_{t-1}+p_{l,2}x_{t-2}+...+p_{l,l}x_{t-l}+e_{l,t}

    where e_{l,t} is the predictor error.

    The first subscript l of p_{l,l} and e_{l,t} emphasizes the fact
    that the parameters will in general alter as further terms are
    introduced into the equation (i.e., as l is increased).

    The parameters are determined from the autocorrelation coefficients
    by the Yule--Walker equations

        r_i = p_{l,1}r_{i-1}+p_{l,2}r_{i-2}+...+p_{l,l}r_{i-l}, i =
        1,2,...,l

    taking r_j = r_|j| when j < 0, and r_0 = 1.

    The predictor error variance ratio v_l = var(e_{l,t})/var(x_t) is
    defined by

        v_l = 1-p_{l,1}r_1-p_{l,2}r_2-...-p_{l,l}r_l.

    The above sets of equations are solved by a recursive method (the
    Durbin--Levinson algorithm).
    The recursive cycle applied for l = 1,2,...,(L-1), where L is the
    number of partial autocorrelation coefficients required, is
    initialized by setting p_{1,1} = r_1 and v_1 = 1-r_1^2.

    Then

        [table omitted]

    If the condition |p_{l,l}| >= 1 occurs, say when l = l_0, it
    indicates that the supplied autocorrelation coefficients do not form
    a positive definite sequence (see Hannan (1960)), and the recursion
    is not continued.
    The autoregressive parameters are overwritten at each recursive
    step, so that upon completion the only available values are p_{Lj},
    for j = 1,2,...,L, or p_{l_0-1,j} if the recursion has been
    prematurely halted.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Durbin, J, 1960, `The fitting of time series models`, Rev. Inst.
    Internat. Stat. (28), 233

    Hannan, E J, 1960, `Time Series Analysis`, Methuen
    """

    fun_name = 'naginterfaces.library.tsa.uni_autocorr_part'

    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nk_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _nk = _b_utils._EngineIntScalarType(
        _nk_py, {'fun_name': fun_name, 'entity': '_nk'},
    ).to_ctypes()
    _nl = _b_utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = _b_utils._EngineIntScalarType.to_py(_nl)
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    _ar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    try:
        _return_scalar = _b_tsa.uni_autocorr_part(
            _r,
            _nk,
            _nl,
            _p,
            _v,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _p_py = _p
    _v_py = _v
    _ar_py = _ar
    _nvl_py = _return_scalar
    _return_data = UniAutocorrPartReturnData(
        _p_py,
        _v_py,
        _ar_py,
        _nvl_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniArimaPrelimReturnData = _collections.namedtuple(
    'UniArimaPrelimReturnData',
    ['par', 'rv', 'isf'],
)

def uni_arima_prelim(mr, r, xv):
    u"""
    Univariate time series, preliminary estimation, seasonal ARIMA
    model.

    ``uni_arima_prelim`` calculates preliminary estimates of the
    parameters of an autoregressive integrated moving average (ARIMA)
    model from the autocorrelation function of the appropriately
    differenced times series.

    For full information please refer to the NAG Library document for
    g13ad

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13adf.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    r : float, array-like, shape (nk)
        The autocorrelations (starting at lag 1), which must have been
        calculated after the time series has been appropriately
        differenced.

    xv : float
        The series sample variance, calculated after appropriate
        differencing has been applied to the series.

    Returns
    -------
    par : float, ndarray, shape (:)
        The first npar elements of `par` contain the preliminary
        estimates of the ARIMA model parameters, in standard order.

    rv : float
        An estimate of the residual variance of the preliminarily
        estimated model.

    isf : int, ndarray, shape (4)
        Contains success/failure indicators, one for each of the four
        types of parameter (autoregressive, moving average, seasonal
        autoregressive, seasonal moving average).

        The indicator has the interpretation:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 2)
            On entry, nk = *<value>*.

            Constraint: nk >= max(p+q, s*(P+Q)).

        (`errno` 3)
            On entry, `r`[*<value>*] = *<value>*.

            Constraint: -1.0 <= `r`[i] <= 1.0, for i = 1,2,...nk.

        (`errno` 4)
            On entry, `xv` = *<value>*.

            Constraint: `xv` > 0.0.

        (`errno` 5)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    Preliminary estimates of the p non-seasonal autoregressive
    parameters phi_1,phi_2,...,phi_p and the q non-seasonal moving
    average parameters theta_1,theta_2,...,theta_q may be obtained from
    the sample autocorrelations relating to lags 1 to p+q, i.e.,
    r_1,...,r_{p+q}, of the differenced nabla^d nabla_s^Dx_t, where x_t
    is assumed to follow a (possibly) seasonal ARIMA model (see [Notes
    for uni_arima_estim] for the specification of an ARIMA model).

    Taking r_0 = 1 and r_{-k} = r_k, the phi_i, for i = 1,2,...,p are
    the solutions to the equations

        r_{q+i-1}phi_1+r_{q+i-2}phi_2+...+r_{q+i-p}phi_p = r_{q+i}, i =
        1,2,...,p.

    The theta_j, for j = 1,2,...,q, are obtained from the solutions to
    the equations

        c_j = tau_0tau_j+tau_1tau_{j+1}+...+tau_{q+j}tau_q, j =
        0,1,...,q

    (Cramer Wold-factorization), by setting

        theta_j = -(tau_j)/(tau_0),

    where c_j are the 'covariances' modified in a two stage process by
    the autoregressive parameters.

    Stage 1:

        [table omitted]

    Stage 2:

        c_j = d_j-phi_1 d_{j+1}-phi_2 d_{j+2}-...-phi_p d_{j+p}, j =
        0,1,...,q.

    The P seasonal autoregressive parameters Phi_1,Phi_2,...,Phi_P and
    the Q seasonal moving average parameters Theta_1,Theta_2,...,Theta_Q
    are estimated in the same way as the non-seasonal parameters, but
    each r_j is replaced in the calculation by r_{s*j}, where s is the
    seasonal period.

    An estimate of the residual variance is obtained by successively
    reducing the sample variance, first for non-seasonal, and then for
    seasonal, parameter estimates.
    If moving average parameters are estimated, the variance is reduced
    by a multiplying factor of tau_0^2, but otherwise by c_0.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.uni_arima_prelim'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nk_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _nk = _b_utils._EngineIntScalarType(
        _nk_py, {'fun_name': fun_name, 'entity': '_nk'},
    ).to_ctypes()
    _npar = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,)), {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _npar_py = _b_utils._EngineIntScalarType.to_py(_npar)
    _par = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalar = _b_tsa.uni_arima_prelim(
            _mr,
            _r,
            _nk,
            xv,
            _npar,
            _par,
            _isf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _par_py = _par
    _rv_py = _return_scalar
    _isf_py = _isf
    _return_data = UniArimaPrelimReturnData(
        _par_py,
        _rv_py,
        _isf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniArimaEstimReturnData = _collections.namedtuple(
    'UniArimaEstimReturnData',
    ['par', 'c', 'icount', 'ex', 'exr', 'al', 's', 'g', 'sd', 'h', 'st', 'nst', 'itc', 'zsp', 'isf'],
)

def uni_arima_estim(mr, par, c, x, iex, igh, ist, kpiv, zsp, kzsp, kfc=1, piv=None, nit=100, data=None, io_manager=None):
    u"""
    Univariate time series, estimation, seasonal ARIMA model
    (comprehensive).

    ``uni_arima_estim`` fits a seasonal autoregressive integrated moving
    average (ARIMA) model to an observed time series, using a nonlinear
    least squares procedure incorporating backforecasting. Parameter
    estimates are obtained, together with appropriate standard errors.
    The residual series is returned, and information for use in
    forecasting the time series is produced for use by the functions
    ``uni_arima_update`` and ``uni_arima_forecast_state``.

    The estimation procedure is iterative, starting with initial
    parameter values such as may be obtained using ``uni_arima_prelim``.
    It continues until a specified convergence criterion is satisfied,
    or until a specified number of iterations has been carried out.
    The progress of the procedure can be monitored by means of a
    user-supplied function.

    For full information please refer to the NAG Library document for
    g13ae

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13aef.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    par : float, array-like, shape (npar)
        The initial estimates of the p values of the phi parameters, the
        q values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters, in that
        order.

    c : float
        If `kfc` = 0, `c` must contain the expected value, c, of the
        differenced series.

        If `kfc` = 1, `c` must contain an initial estimate of c.

    x : float, array-like, shape (nx)
        The n values of the original undifferenced time series.

    iex : int
        The dimension of the arrays `ex`, `exr` and `al`.

    igh : int
        The dimension of the arrays `g` and `sd`.

        The second dimension of the arrays `h` and `hc`.

    ist : int
        The dimension of the array `st`.

    kpiv : int
        Must be nonzero if the progress of the optimization is to be
        monitored using `piv`. Otherwise `kpiv` must contain 0.

    zsp : float, array-like, shape (4)
        When `kzsp` = 1, the first four elements of `zsp` must contain
        the four values used to guide the search procedure. These are as
        follows.

        `zsp`[0] contains alpha, the value used to constrain the
        magnitude of the search procedure steps.

        `zsp`[1] contains beta, the multiplier which regulates the value
        alpha.

        `zsp`[2] contains delta, the value of the stationarity and
        invertibility test tolerance factor.

        `zsp`[3] contains gamma, the value of the convergence criterion.

        If `kzsp` != 1 on entry, default values for `zsp` are supplied
        by the function.

        These are 0.001, 10.0, 1000.0 and max(100*machine
        precision,0.0000001) respectively.

    kzsp : int
        The value 1 if the function is to use the input values of `zsp`,
        and any other value if the default values of `zsp` are to be
        used.

    kfc : int, optional
        Must be set to 1 if the constant, c, is to be estimated and 0 if
        it is to be held fixed at its initial value.

    piv : None or callable piv(mr, par, c, kfc, icount, s, g, h, itc,
        zsp, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `piv` is used to monitor the progress of the optimization.

        Parameters
        ~~~~~~~~~~
        mr : int, ndarray, shape (7)
            `mr`, defined as for ``uni_arima_estim``.

        par : float, ndarray, shape (npar)
            `par`, defined as for ``uni_arima_estim``.

        c : float
            `c`, defined as for ``uni_arima_estim``.

        kfc : int
            `kfc`, defined as for ``uni_arima_estim``.

        icount : int, ndarray, shape (6)
            `icount`, defined as for ``uni_arima_estim``.

        s : float
            `s`, defined as for ``uni_arima_estim``.

        g : float, ndarray, shape (igh)
            `g`, defined as for ``uni_arima_estim``.

        h : float, ndarray, shape (:, igh)
            `h`, defined as for ``uni_arima_estim``.

        itc : int
            `itc`, defined as for ``uni_arima_estim``.

        zsp : float, ndarray, shape (4)
            `zsp`, defined as for ``uni_arima_estim``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    nit : int, optional
        The maximum number of iterations to be performed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    par : float, ndarray, shape (npar)
        The latest values of the estimates of these parameters.

    c : float
        If `kfc` = 0, `c` is unchanged.

        If `kfc` = 1, `c` contains the latest estimate of c.

        Therefore, if `c` and `kfc` are both zero on entry, there is no
        constant correction.

    icount : int, ndarray, shape (6)
        Size of various output arrays.

        `icount`[0]

            Contains q+(Q*s), the number of backforecasts.

        `icount`[1]

            Contains n-d-(D*s), the number of differenced values.

        `icount`[2]

            Contains d+(D*s), the number of values of reconstitution
            information.

        `icount`[3]

            Contains n+q+(Q*s), the number of values held in each of the
            series `ex`, `exr` and `al`.

        `icount`[4]

            Contains n-d-(D*s)-p-q-P-Q-`kfc`, the number of degrees of
            freedom associated with S.

        `icount`[5]

            Contains `icount`[0]+npar+`kfc`, the number of parameters
            being estimated.

        These values are always computed regardless of the exit value of
        errno.

    ex : float, ndarray, shape (`iex`)
        The extended differenced series which is made up of:

        `icount`[0] backforecast values of the differenced series.

        `icount`[1] actual values of the differenced series.

        `icount`[2] values of reconstitution information.

        The total number of these values held in `ex` is `icount`[3].

        If the function exits because of a faulty input parameter, the
        contents of `ex` will be indeterminate.

    exr : float, ndarray, shape (`iex`)
        The values of the model residuals which is made up of:

        `icount`[0] residuals corresponding to the backforecasts in the
        differenced series.

        `icount`[1] residuals corresponding to the actual values in the
        differenced series.

        The remaining `icount`[2] values contain zeros.

        If the function exits with errno holding a value other than 0 or
        9, the contents of `exr` will be indeterminate.

    al : float, ndarray, shape (`iex`)
        The intermediate series which is made up of:

        `icount`[0] intermediate series values corresponding to the
        backforecasts in the differenced series.

        `icount`[1] intermediate series values corresponding to the
        actual values in the differenced series.

        The remaining `icount`[2] values contain zeros.

        If an exception is raised, the contents of `al` will be
        indeterminate.

    s : float
        The residual sum of squares after the latest series of parameter
        estimates has been incorporated into the model. If the function
        exits with a faulty input parameter, `s` contains zero.

    g : float, ndarray, shape (`igh`)
        The latest value of the derivatives of S with respect to each of
        the parameters being estimated (backforecasts, `par` parameters,
        and where relevant the constant -- in that order). The contents
        of `g` will be indeterminate if the function exits with a faulty
        input parameter.

    sd : float, ndarray, shape (`igh`)
        The standard deviations corresponding to each of the parameters
        being estimated (backforecasts, `par` parameters, and where
        relevant the constant, in that order).

        If the function exits with errno containing a value other than 0
        or 9, or if the required number of iterations is zero, the
        contents of `sd` will be indeterminate.

    h : float, ndarray, shape (1+q+(Q*s)+npar+`kfc`, `igh`)
        The second derivative of S and correlation coefficients.

        (a) the latest values of an approximation to the second
            derivative of S with respect to each of the
            (q+Q*s+npar+`kfc`) parameters being estimated
            (backforecasts, `par` parameters, and where relevant the
            constant -- in that order), and

        (#) the correlation coefficients relating to each pair of these
            parameters.

        These are held in a matrix defined by the first
        (q+Q*s+npar+`kfc`) rows and the first (q+Q*s+npar+`kfc`) columns
        of `h`. (Note that `icount`[5] contains the value of this
        expression.) The values of (a) are contained in the upper
        triangle, and the values of (b) in the strictly lower triangle.

        These correlation coefficients are zero during intermediate
        printout using `piv`, and indeterminate if errno contains on
        exit a value other than 0 or 9.

        All the contents of `h` are indeterminate if the required number
        of iterations are zero.

        The (q+(Q*s)+npar+`kfc`+1)th row of `h` is used internally as
        workspace.

    st : float, ndarray, shape (`ist`)
        The `nst` values of the state set array. If the function exits
        with errno containing a value other than 0 or 9, the contents of
        `st` will be indeterminate.

    nst : int
        The number of values in the state set array `st`.

    itc : int
        The number of iterations performed.

    zsp : float, ndarray, shape (4)
        `zsp` contains the values, default or otherwise, used by the
        function.

    isf : int, ndarray, shape (4)
        Contains success/failure indicators, one for each of the four
        types of parameter in the model (autoregressive, moving average,
        seasonal autoregressive, seasonal moving average), in that
        order.

        Each indicator has the interpretation:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 3)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 3)
            On entry, `zsp`[3] = *<value>*.

            Constraint: 0.0 <= `zsp`[3] < 1.0.

        (`errno` 3)
            On entry, `zsp`[2] = *<value>*.

            Constraint: `zsp`[2] >= 1.0.

        (`errno` 3)
            On entry, `zsp`[1] = *<value>*.

            Constraint: `zsp`[1] > 1.0.

        (`errno` 3)
            On entry, `zsp`[0] = *<value>*.

            Constraint: `zsp`[0] > 0.0.

        (`errno` 4)
            On entry, `ist` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ist` >= (P*s)+d+(D*s)+q+max(p, Q*s).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            On entry, ldh = *<value>*.

            Constraint: ldh >= q*(Q*s)+npar+`kfc`.

        (`errno` 6)
            On entry, `igh` = *<value>*.

            Constraint: `igh` >= q*(Q*s)+npar+`kfc`.

        (`errno` 6)
            On entry, `iex` = *<value>*.

            Constraint: `iex` >= q*(Q*s)+nx.

        (`errno` 7)
            A failure in the search procedure has occurred.

        (`errno` 8)
            Failure to invert H.

        (`errno` 9)
            Unable to calculate the latest estimates of the
            backforecasts.

        (`errno` 10)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The time series x_1,x_2,...,x_n supplied to ``uni_arima_estim`` is
    assumed to follow a seasonal autoregressive integrated moving
    average (ARIMA) model defined as follows:

        nabla^d nabla_s^Dx_t-c = w_t,

    where nabla^d nabla_s^Dx_t is the result of applying non-seasonal
    differencing of order d and seasonal differencing of seasonality s
    and order D to the series x_t, as outlined in the description of
    ``uni_diff``.
    The differenced series is then of length N = n-d', where d' =
    d+(D*s) is the generalized order of differencing.
    The scalar c is the expected value of the differenced series, and
    the series w_1,w_2,...,w_N follows a zero-mean stationary
    autoregressive moving average (ARMA) model defined by a pair of
    recurrence equations.
    These express w_t in terms of an uncorrelated series a_t, via an
    intermediate series e_t.
    The first equation describes the seasonal structure:

        w_t = Phi_1 w_{t-s}+Phi_2 w_{t-2*s}+...+Phi_P
        w_{t-P*s}+e_t-Theta_1 e_{t-s}-Theta_2 e_{t-2*s}-...-Theta_Q
        e_{t-Q*s}.

    The second equation describes the non-seasonal structure.
    If the model is purely non-seasonal the first equation is redundant
    and e_t above is equated with w_t:

        e_t = phi_1 e_{t-1}+phi_2 e_{t-2}+...+phi_p e_{t-p}+a_t-theta_1
        a_{t-1}-theta_2 a_{t-2}-...-theta_q a_{t-q}.

    Estimates of the model parameters defined by

phi_1,phi_2,...,phi_p,theta_1,theta_2,...,theta_q,;Phi_1,Phi_2,...,Phi_P,Theta_1,Theta_2,...,Theta_Q

    and (optionally) c are obtained by minimizing a quadratic form in
    the vector w = (w_1,w_2,...,w_N)'.

    This is QF = w'V^-1w, where V is the covariance matrix of w, and is
    a function of the model parameters.
    This matrix is not explicitly evaluated, since QF may be expressed
    as a 'sum of squares' function.
    When moving average parameters theta_i or Theta_i are present, so
    that the generalized moving average order q' = q+s*Q is positive,
    backforecasts w_{1-q'},w_{2-q'},...,w_0 are introduced as nuisance
    parameters.
    The 'sum of squares' function may then be written as

        S(pm) = sum_t = 1-q'^Na_t^2-sum_t = 1-q'-p'^-q'b_t^2,

    where pm is a combined vector of parameters, consisting of the
    backforecasts followed by the ARMA model parameters.

    The terms a_t correspond to the ARMA model residual series a_t, and
    p' = p+s*P is the generalized autoregressive order.
    The terms b_t are only present if autoregressive parameters are in
    the model, and serve to correct for transient errors introduced at
    the start of the autoregression.

    The equations defining a_t and b_t are precisely:

        e_t = w_t-Phi_1 w_{t-s}-Phi_2 w_{t-2*s}-...-Phi_P
        w_{t-P*s}+Theta_1 e_{t-s}+Theta_2 e_{t-2*s}+...+Theta_Q
        e_{t-Q*s},
        for t = 1-q',2-q',...,n.

        a_t = e_t-phi_1 e_{t-1}-phi_2 e_{t-2}-...-phi_p e_{t-p}+theta_1
        a_{t-1}+theta_2 a_{t-2}+...+theta_q a_{t-q},
        for t = 1-q',2-q',...,n.

        f_t = w_t-Phi_1 w_{t+s}-Phi_2 w_{t+2*s}-...-Phi_P
        w_{t+P*s}+Theta_1 f_{t-s}+Theta_2 f_{t-2*s}+...+Theta_Q
        f_{t-Q*s},
        for t = (1-q'-s*P),(2-q'-s*P),...,(-q'+P)

        b_t = f_t-phi_1 f_{t+1}-phi_2 f_{t+2}-...-phi_p f_{t+p}+theta_1
        b_{t-1}+theta_2 b_{t-2}+...+theta_q b_{t-q},
        for t = (1-q'-p'),(2-q'-p'),...,(-q').

    For all four of these equations, the following conditions hold:

        w_i = 0 if i < 1-q'

        e_i = 0 if i < 1-q'

        a_i = 0 if i < 1-q'

        f_i = 0 if i < 1-q'-s*P

        b_i = 0 if i < 1-q'-p'

    Minimization of S with respect to pm uses an extension of the
    algorithm of Marquardt (1963).

    The first derivatives of S with respect to the parameters are
    calculated as

        2*suma_t*a_{t,i}-2sumb_t*b_{t,i} = 2*G_i,

    where a_{t,i} and b_{t,i} are derivatives of a_t and b_t with
    respect to the ith parameter.

    The second derivative of S is approximated by

        2*suma_{t,i}*a_{t,j}-2*sumb_{t,i}*b_{t,j} = 2*H_{ij}.

    Successive parameter iterates are obtained by calculating a vector
    of corrections dpm by solving the equations

        (H+alpha*D)*dpm = -G,

    where G is a vector with elements G_i, H is a matrix with elements
    H_{ij}, alpha is a scalar used to control the search and D is the
    diagonal matrix of H.

    The new parameter values are then pm+dpm.

    The scalar alpha controls the step size, to which it is inversely
    related.

    If a step results in new parameter values which give a reduced value
    of S, then alpha is reduced by a factor beta.
    If a step results in new parameter values which give an increased
    value of S, or in ARMA model parameters which in any way contravene
    the stationarity and invertibility conditions, then the new
    parameters are rejected, alpha is increased by the factor beta, and
    the revised equations are solved for a new parameter correction.

    This action is repeated until either a reduced value of S is
    obtained, or alpha reaches the limit of 10^9, which is used to
    indicate a failure of the search procedure.

    This failure may be due to a badly conditioned sum of squares
    function or to too strict a convergence criterion.
    Convergence is deemed to have occurred if the fractional reduction
    in the residual sum of squares in successive iterations is less than
    a value gamma, while alpha < 1.0.

    The stationarity and invertibility conditions are tested to within a
    specified tolerance multiple delta of machine accuracy.
    Upon convergence, or completion of the specified maximum number of
    iterations without convergence, statistical properties of the
    estimates are derived.
    In the latter case the sequence of iterates should be checked to
    ensure that convergence is adequate for practical purposes,
    otherwise these properties are not reliable.

    The estimated residual variance is

        erv = S_min/df,

    where S_min is the final value of S, and the residual number of
    degrees of freedom is given by

        df = N-p-q-P-Q (-1 if c is estimated).

    The covariance matrix of the vector of estimates pm is given by

        erv*H^-1,

    where H is evaluated at the final parameter values.

    From this expression are derived the vector of standard deviations,
    and the correlation matrix for the whole parameter set.
    These are asymptotic approximations.

    The differenced series w_t (now uncorrected for the constant),
    intermediate series e_t and residual series a_t are all available
    upon completion of the iterations over the range (extended by
    backforecasts)

        t = 1-q',2-q',...,N.

    The values a_t can only properly be interpreted as residuals for t
    >= 1+p'-q', as the earlier values are corrupted by transients if p'
    > 0.

    In consequence of the manner in which differencing is implemented,
    the residual a_t is the one step ahead forecast error for x_{t+d'}.

    For convenient application in forecasting, the following quantities
    constitute the 'state set', which contains the minimum amount of
    time series information needed to construct forecasts:

    (i) the differenced series w_t, for (N-s*P) < t <= N,

    (#) the d' values required to reconstitute the original series x_t
        from the differenced series w_t,

    (#) the intermediate series e_t, for (N-max(p,Q*s)) < t <= N,

    (#) the residual series a_t, for (N-q) < t <= N.

    This state set is available upon completion of the iterations.
    The function may be used purely for the construction of this state
    set, given a previously estimated model and time series x_t, by
    requesting zero iterations.
    Backforecasts are estimated, but the model parameter values are
    unchanged.
    If later observations become available and it is desired to update
    the state set, ``uni_arima_update`` can be used.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Marquardt, D W, 1963, `An algorithm for least squares estimation of
    nonlinear parameters`, J. Soc. Indust. Appl. Math. (11), 431
    """

    def cb_piv(mr, par, npar, c, kfc, icount, s, g, h, igh, itc, zsp, data=None):
        """Wrapping user's piv. Called as base piv."""
        if data is not None:
            piv(
                mr,
                par,
                c,
                kfc,
                icount,
                s,
                g,
                h,
                itc,
                zsp,
                data,
            )
        else:
            piv(
                mr,
                par,
                c,
                kfc,
                icount,
                s,
                g,
                h,
                itc,
                zsp,
            )

    fun_name = 'naginterfaces.library.tsa.uni_arima_estim'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _icount = _b_utils._EngineIntArrayType.empty_ndarray(
        (6,),
    )
    _iex = _b_utils._EngineIntScalarType(
        iex, {'fun_name': fun_name, 'entity': 'iex'},
    ).to_ctypes()
    _iex_py = _b_utils._EngineIntScalarType.to_py(_iex)
    _ex = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_iex_py,),
    )
    _exr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_iex_py,),
    )
    _al = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_iex_py,),
    )
    _igh = _b_utils._EngineIntScalarType(
        igh, {'fun_name': fun_name, 'entity': 'igh'},
    ).to_ctypes()
    _igh_py = _b_utils._EngineIntScalarType.to_py(_igh)
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_igh_py,),
    )
    _sd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_igh_py,),
    )
    _vl_s_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _kfc = _b_utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _kfc_py = _b_utils._EngineIntScalarType.to_py(_kfc)
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+_vl_q_py+(_vu_q_py*_vl_s_py)+_npar_py+_kfc_py, _igh_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _ist = _b_utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _ist_py = _b_utils._EngineIntScalarType.to_py(_ist)
    _st = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ist_py,),
    )
    _b_utils._check_callable(piv, {'fun_name': fun_name, 'entity': 'piv'}, allow_none=True)
    _piv = None if piv is None else cb_piv
    _zsp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'zsp'},
        zsp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalars = _b_tsa.uni_arima_estim(
            _mr,
            _par,
            _npar,
            c,
            _kfc,
            _x,
            _nx,
            _icount,
            _ex,
            _exr,
            _al,
            _iex,
            _g,
            _igh,
            _sd,
            _h,
            _st,
            _ist,
            _piv,
            kpiv,
            nit,
            _zsp,
            kzsp,
            _isf,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _par_py = _par
    _c_py = _return_scalars.c
    _zsp_py = _zsp
    _icount_py = _icount
    _ex_py = _ex
    _exr_py = _exr
    _al_py = _al
    _s_py = _return_scalars.s
    _g_py = _g
    _sd_py = _sd
    _h_py = _h
    _st_py = _st
    _nst_py = _return_scalars.nst
    _itc_py = _return_scalars.itc
    _isf_py = _isf
    _return_data = UniArimaEstimReturnData(
        _par_py,
        _c_py,
        _icount_py,
        _ex_py,
        _exr_py,
        _al_py,
        _s_py,
        _g_py,
        _sd_py,
        _h_py,
        _st_py,
        _nst_py,
        _itc_py,
        _zsp_py,
        _isf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniArimaEstimEasyReturnData = _collections.namedtuple(
    'UniArimaEstimEasyReturnData',
    ['par', 'c', 's', 'ndf', 'sd', 'cm', 'st', 'nst', 'itc', 'isf', 'res', 'nres'],
)

def uni_arima_estim_easy(mr, par, c, x, nppc, kfc=1, kpiv=0, nit=100, ires=None, io_manager=None):
    u"""
    Univariate time series, estimation, seasonal ARIMA model
    (easy-to-use).

    ``uni_arima_estim_easy`` is an easy-to-use version of
    ``uni_arima_estim``.
    It fits a seasonal autoregressive integrated moving average (ARIMA)
    model to an observed time series, using a nonlinear least squares
    procedure incorporating backforecasting. Parameter estimates are
    obtained, together with appropriate standard errors.
    The residual series is returned, and information for use in
    forecasting the time series is produced for use in
    ``uni_arima_update`` and ``uni_arima_forecast_state``.

    The estimation procedure is iterative, starting with initial
    parameter values such as may be obtained using ``uni_arima_prelim``.
    It continues until a specified convergence criterion is satisfied or
    until a specified number of iterations have been carried out.
    The progress of the iteration can be monitored by means of an
    optional printing facility.

    For full information please refer to the NAG Library document for
    g13af

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13aff.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    par : float, array-like, shape (npar)
        The initial estimates of the p values of the phi parameters, the
        q values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters, in that
        order.

    c : float
        If `kfc` = 0, `c` must contain the expected value, c, of the
        differenced series.

        If `kfc` = 1, `c` must contain an initial estimate of c.

        Therefore, if `c` and `kfc` are both zero on entry, there is no
        constant correction.

    x : float, array-like, shape (nx)
        The n values of the original undifferenced time series.

    nppc : int
        The number of phi, theta, Phi, Theta and c parameters to be
        estimated. `nppc` = p+q+P+Q+1 if the constant is being estimated
        and `nppc` = p+q+P+Q if not.

    kfc : int, optional
        Must be set to 1 if the constant, c, is to be estimated and 0 if
        it is to be held fixed at its initial value.

    kpiv : int, optional
        Must be nonzero if the progress of the optimization is to be
        monitored using the built-in printing facility. Otherwise `kpiv`
        must contain zero. If selected, monitoring output will be sent
        to the file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). For each
        iteration, the heading

        ::

          G13AFZ MONITORING OUTPUT - ITERATION n

        followed by the argument values, and residual sum of squares,
        are printed.

    nit : int, optional
        The maximum number of iterations to be performed.

    ires : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows:
        15*(q+Q*s)+11*nx+13*(npar+`kfc`)+8*(p+P*s)+12+2*(q+Q*s+npar+`kfc`)^2.

        The dimension of the array `res`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    par : float, ndarray, shape (npar)
        Contains the latest values of the estimates of these parameters.

    c : float
        If `kfc` = 0, `c` is unchanged.

        If `kfc` = 1, `c` contains the latest estimate of c.

    s : float
        The residual sum of squares after the latest series of parameter
        estimates has been incorporated into the model. If the function
        exits with a faulty input parameter, `s` contains zero.

    ndf : int
        The number of degrees of freedom associated with `s`,

        `ndf` = n-d-D*s-p-q-P-Q-`kfc`.

    sd : float, ndarray, shape (`nppc`)
        The standard deviations corresponding to the parameters in the
        model (p autoregressive, q moving average, P seasonal
        autoregressive, Q seasonal moving average and c, if estimated,
        in that order). If the function exits with errno containing a
        value other than 0 or 9, or if the required number of iterations
        is zero, the contents of `sd` will be indeterminate.

    cm : float, ndarray, shape (`nppc`, `nppc`)
        The correlation coefficients associated with each pair of the
        `nppc` parameters. These are held in the first `nppc` rows and
        the first `nppc` columns of `cm`. These correlation coefficients
        are indeterminate if errno contains on exit a value other than 0
        or 9, or if the required number of iterations is zero.

    st : float, ndarray, shape (nx)
        The value of the state set in its first `nst` elements. If the
        function exits with errno containing a value other than 0 or 9,
        the contents of `st` will be indeterminate.

    nst : int
        The size of the state set. `nst` = P*s+D*s+d+q+max(p, Q*s).

        `nst` should be used subsequently in ``uni_arima_update`` and
        ``uni_arima_forecast_state`` as the dimension of `st`.

    itc : int
        The number of iterations performed.

    isf : int, ndarray, shape (4)
        The first four elements of `isf` contain success/failure
        indicators, one for each of the four types of parameter in the
        model (autoregressive, moving average, seasonal autoregressive,
        seasonal moving average), in that order.

        Each indicator has the interpretation:

        [table omitted]

    res : float, ndarray, shape (`ires`)
        The first `nres` elements of `res` contain the model residuals
        derived from the differenced series. If the function exits with
        errno holding a value other than 0 or 9, these elements of `res`
        will be indeterminate. The rest of the array `res` is used as
        workspace.

    nres : int
        The number of model residuals returned in `res`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 3)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 4)
            On entry, nx = *<value>* and minimum required size for the
            state set array = *<value>*.

        (`errno` 5)
            On entry, `ires` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ires` >=
            15*Q'+11n+13*`nppc`+8*P'+12+2*(Q'+`nppc`)^2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            A failure in the search procedure has occurred.

        (`errno` 8)
            The inversion of the Hessian matrix in the calculation of
            the covariance matrix of the parameter estimates has failed.

        (`errno` 9)
            Failure whilst calculating backforecasts.

        (`errno` 10)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The time series x_1,x_2,...,x_n supplied to the function is assumed
    to follow a seasonal autoregressive integrated moving average
    (ARIMA) model defined as follows:

        nabla^d nabla_s^Dx_t-c = w_t,

    where nabla^d nabla_s^Dx_t is the result of applying non-seasonal
    differencing of order d and seasonal differencing of seasonality s
    and order D to the series x_t, as outlined in the description of
    ``uni_diff``.
    The differenced series is then of length N = n-d', where d' =
    d+(D*s) is the generalized order of differencing.
    The scalar c is the expected value of the differenced series, and
    the series w_1,w_2,...,w_N follows a zero-mean stationary
    autoregressive moving average (ARMA) model defined by a pair of
    recurrence equations.
    These express w_t in terms of an uncorrelated series a_t, via an
    intermediate series e_t.
    The first equation describes the seasonal structure:

        w_t = Phi_1 w_{t-s}+Phi_2 w_{t-2*s}+...+Phi_P
        w_{t-P*s}+e_t-Theta_1 e_{t-s}-Theta_2 e_{t-2*s}-...-Theta_Q
        e_{t-Q*s}.

    The second equation describes the non-seasonal structure.
    If the model is purely non-seasonal the first equation is redundant
    and e_t above is equated with w_t:

        e_t = phi_1 e_{t-1}+phi_2 e_{t-2}+...+phi_p e_{t-p}+a_t-theta_1
        a_{t-1}-theta_2 a_{t-2}-...-theta_q a_{t-q}.

    Estimates of the model parameters defined by

phi_1,phi_2,...,phi_p,theta_1,theta_2,...,theta_q,;Phi_1,Phi_2,...,Phi_P,Theta_1,Theta_2,...,Theta_Q

    and (optionally) c are obtained by minimizing a quadratic form in
    the vector w = (w_1,w_2,...,w_N)'.

    The minimization process is iterative, iterations being performed
    until convergence is achieved (see [Notes for uni_arima_estim] for
    full details), or until the user-specified maximum number of
    iterations are completed.

    The final values of the residual sum of squares and the parameter
    estimates are used to obtain asymptotic approximations to the
    standard deviations of the parameters, and the correlation matrix
    for the parameters.
    The 'state set' array of information required by forecasting is also
    returned.

    **Note:** if the maximum number of iterations are performed without
    convergence, these quantities may not be reliable.
    In this case, the sequence of iterates should be checked, using the
    optional monitoring function, to verify that convergence is adequate
    for practical purposes.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Marquardt, D W, 1963, `An algorithm for least squares estimation of
    nonlinear parameters`, J. Soc. Indust. Appl. Math. (11), 431
    """

    fun_name = 'naginterfaces.library.tsa.uni_arima_estim_easy'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _nppc = _b_utils._EngineIntScalarType(
        nppc, {'fun_name': fun_name, 'entity': 'nppc'},
    ).to_ctypes()
    _nppc_py = _b_utils._EngineIntScalarType.to_py(_nppc)
    _sd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nppc_py,),
    )
    _cm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nppc_py, _nppc_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _st = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nx_py,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    _kfc = _b_utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _kfc_py = _b_utils._EngineIntScalarType.to_py(_kfc)
    _vl_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _vl_s_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))
    _vu_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))
    _ires = 15*(_vl_q_py+_vu_q_py*_vl_s_py)+11*_nx_py+13*(_npar_py+_kfc_py)+8*(_vl_p_py+_vu_p_py*_vl_s_py)+12+2*(_vl_q_py+_vu_q_py*_vl_s_py+_npar_py+_kfc_py)**2 if ires is None else ires
    _ires = _b_utils._EngineIntScalarType(
        _ires, {'fun_name': fun_name, 'entity': '_ires'},
    ).to_ctypes()
    _ires_py = _b_utils._EngineIntScalarType.to_py(_ires)
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ires_py,),
    )
    try:
        _return_scalars = _b_tsa.uni_arima_estim_easy(
            _mr,
            _par,
            _npar,
            c,
            _kfc,
            _x,
            _nx,
            _sd,
            _nppc,
            _cm,
            _st,
            kpiv,
            nit,
            _isf,
            _res,
            _ires,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _par_py = _par
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _ndf_py = _return_scalars.ndf
    _sd_py = _sd
    _cm_py = _cm
    _st_py = _st
    _nst_py = _return_scalars.nst
    _itc_py = _return_scalars.itc
    _isf_py = _isf
    _res_py = _res
    _nres_py = _return_scalars.nres
    _return_data = UniArimaEstimEasyReturnData(
        _par_py,
        _c_py,
        _s_py,
        _ndf_py,
        _sd_py,
        _cm_py,
        _st_py,
        _nst_py,
        _itc_py,
        _isf_py,
        _res_py,
        _nres_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniArimaUpdateReturnData = _collections.namedtuple(
    'UniArimaUpdateReturnData',
    ['st', 'anexr'],
)

def uni_arima_update(st, mr, par, c, anx):
    u"""
    Univariate time series, update state set for forecasting.

    ``uni_arima_update`` accepts a series of new observations of a time
    series, the model of which is already fully specified, and updates
    the 'state set' information for use in constructing further
    forecasts.
    The previous specifications of the time series model should have
    been obtained by using ``uni_arima_estim`` or
    ``uni_arima_estim_easy`` to estimate the relevant parameters.
    The supplied state set will originally have been produced by
    ``uni_arima_estim`` or ``uni_arima_estim_easy``, but may since have
    been updated by earlier calls to ``uni_arima_update``.

    A set of residuals corresponding to the new observations is
    returned.
    These may be of use in checking that the new observations conform to
    the previously fitted model.

    For full information please refer to the NAG Library document for
    g13ag

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13agf.html

    Parameters
    ----------
    st : float, array-like, shape (nst)
        The state set derived from ``uni_arima_estim`` or
        ``uni_arima_estim_easy``, or as modified using earlier calls of
        ``uni_arima_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, array-like, shape (npar)
        The estimates of the p values of the phi parameters, the q
        values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters in the model
        -- in that order, using the usual notation.

    c : float
        The constant to be subtracted from the differenced data.

    anx : float, array-like, shape (nuv)
        The new undifferenced observations which are to be used to
        update `st`.

    Returns
    -------
    st : float, ndarray, shape (nst)
        The updated values of the state set.

    anexr : float, ndarray, shape (nuv)
        The residuals corresponding to the new observations in `anx`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            On entry, nst = *<value>*.

            Constraint: nst = P*s+D*s+d+q+max(Q*s, p).

        (`errno` 3)
            On entry, nuv = *<value>*.

            Constraint: nuv > 0.

        (`errno` 4)
            On entry, nwa = *<value>* and the minimum size required =
            *<value>*.

            Constraint: nwa >= 4*npar+3*nst.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The time series model is specified as outlined in [Notes for
    uni_arima_estim] and [Notes for uni_arima_estim_easy].
    This also describes how the state set, which contains the minimum
    amount of time series information needed to construct forecasts, is
    made up of

    (i) the differenced series w_t (uncorrected for the constant c), for
        (N-P*s) < t <= N,

    (#) the d' values required to reconstitute the original series x_t
        from the differenced series w_t,

    (#) the intermediate series e_t, for (N-max(p,Q*s)) < t <= N, and

    (#) the residual series a_t, for (N-q) < t <= N.

    If the number of original undifferenced observations was n, then d'
    = d+(D*s) and N = n-d'.

    To update the state set, given a number of new undifferenced
    observations x_t, t = n+1,n+2,...,n+k, the four series above are
    first reconstituted.

    Differencing and residual calculation operations are then applied to
    the new observations and k new values of w_t,e_t and a_t are
    derived.

    The first k values in these three series are then discarded and a
    new state set is obtained.

    The residuals in the a_t series corresponding to the k new
    observations are preserved in an output array.
    The parameters of the time series model are not changed in this
    function.
    """

    fun_name = 'naginterfaces.library.tsa.uni_arima_update'

    _st = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nst_py = _b_utils._EngineFloat64ArrayType.get_shape(_st, {'fun_name': fun_name, 'entity': 'st'}, exp_rank=1)[0]
    _nst = _b_utils._EngineIntScalarType(
        _nst_py, {'fun_name': fun_name, 'entity': '_nst'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _anx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'anx'},
        anx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nuv_py = _b_utils._EngineFloat64ArrayType.get_shape(_anx, {'fun_name': fun_name, 'entity': 'anx'}, exp_rank=1)[0]
    _nuv = _b_utils._EngineIntScalarType(
        _nuv_py, {'fun_name': fun_name, 'entity': '_nuv'},
    ).to_ctypes()
    _anexr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nuv_py,),
    )
    _b_tsa.uni_arima_update(
        _st,
        _nst,
        _mr,
        _par,
        _npar,
        c,
        _anx,
        _nuv,
        _anexr,
    )
    _st_py = _st
    _anexr_py = _anexr
    _return_data = UniArimaUpdateReturnData(
        _st_py,
        _anexr_py,
    )
    return _return_data

UniArimaForecastStateReturnData = _collections.namedtuple(
    'UniArimaForecastStateReturnData',
    ['fva', 'fsd'],
)

def uni_arima_forecast_state(st, mr, par, c, rms, nfv):
    u"""
    Univariate time series, forecasting from state set.

    ``uni_arima_forecast_state`` produces forecasts of a time series,
    given a time series model which has already been fitted to the time
    series using ``uni_arima_estim`` or ``uni_arima_estim_easy``.
    The original observations are not required, since
    ``uni_arima_forecast_state`` uses as input either the original state
    set produced by ``uni_arima_estim`` or ``uni_arima_estim_easy`` or
    the state set updated by a series of new observations using
    ``uni_arima_update``.
    Standard errors of the forecasts are also provided.

    For full information please refer to the NAG Library document for
    g13ah

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ahf.html

    Parameters
    ----------
    st : float, array-like, shape (nst)
        The state set derived from ``uni_arima_estim`` or
        ``uni_arima_estim_easy`` originally, or as modified using
        earlier calls of ``uni_arima_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, array-like, shape (npar)
        The estimates of the p values of the phi parameters, the q
        values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters which
        specify the model and which were output originally by
        ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    c : float
        c, the value of the model constant. This will have been output
        by ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    rms : float
        V, the residual variance associated with the model.

        If ``uni_arima_estim_easy`` was used to estimate the model,
        `rms` should be set to s/ndf, where s and ndf were output by
        ``uni_arima_estim_easy``.

        If ``uni_arima_estim`` was used to estimate the model, `rms`
        should be set to s/icount[4], where s and icount[4] were output
        by ``uni_arima_estim``.

    nfv : int
        L, the required number of forecasts.

    Returns
    -------
    fva : float, ndarray, shape (`nfv`)
        `nfv` forecast values relating to the original undifferenced
        series.

    fsd : float, ndarray, shape (`nfv`)
        The standard errors associated with each of the `nfv` forecast
        values in `fva`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            On entry, nst = *<value>*.

            Constraint: nst = P*s+D*s+d+q+max(Q*s, p).

        (`errno` 3)
            On entry, `nfv` = *<value>*.

            Constraint: `nfv` > 0.

        (`errno` 4)
            On entry, nwa = *<value>* and the minimum size required =
            *<value>*.

            Constraint: nwa >= 4*npar+3*nst.

        (`errno` 5)
            On entry, `rms` = *<value>*.

            Constraint: `rms` >= 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The original time series is x_t, for t = 1,2,...,n and parameters
    have been fitted to the model of this time series using
    ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    Forecasts of x_t, for t = n+1,...,n+L, are calculated in five
    stages, as follows:

    (i) set a_t = 0 for t = N+1,N+2,...,N+L, where N = n-d-(D*s) is the
        number of differenced values in the series;

    (#) calculate the values of e_t, for t = N+1,...,N+L, and e_t =
        phi_1*e_{t-1}+... +
        phi_p*e_{t-p}+a_t-theta_1*a_{t-1}-...-theta_q*a_{t-q};

    (#) calculate the values of w_t, for t = N+1,...,N+L, where w_t =
        Phi_1*w_{t-s}+... +
        Phi_P*w_{t-s*P}+e_t-Theta_1*e_{t-s}-...-Theta_Q*e_{t-s*Q} and
        w_t for t <= N are the first s*P values in the state set,
        corrected for the constant;

    (#) add the constant term c to give the differenced series nabla^d
        nabla_s^Dx_t = w_t+c, for t = N+1,...,N+L;

    (#) the differencing operations are reversed to reconstitute x_t,
        for t = n+1,...,n+L.

    The standard errors of these forecasts are given by s_t =
    [V*(psi_0^2+psi_1^2+...+psi_t-n-1^2)]^(1/2), for t = n+1,...,n+L,
    where psi_0 = 1, V is the residual variance of a_t, and psi_j is the
    coefficient expressing the dependence of x_t on a_{t-j}.

    To calculate psi_j, for j = 1,2,...,(L-1), the following device is
    used.

    A copy of the state set is initialized to zero throughout and the
    calculations outlined above for the construction of forecasts are
    carried out with the settings a_{N+1} = 1, and a_t = 0, for t =
    N+2,...,N+L.

    The resulting quantities corresponding to the sequence
    x_{N+1},x_{N+2},...,x_{N+L} are precisely 1,
    psi_1,psi_2,...,psi_{L-1}.

    The supplied time series model is used throughout these
    calculations, with the exception that the constant term c is taken
    to be zero.
    """

    fun_name = 'naginterfaces.library.tsa.uni_arima_forecast_state'

    _st = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nst_py = _b_utils._EngineFloat64ArrayType.get_shape(_st, {'fun_name': fun_name, 'entity': 'st'}, exp_rank=1)[0]
    _nst = _b_utils._EngineIntScalarType(
        _nst_py, {'fun_name': fun_name, 'entity': '_nst'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _nfv = _b_utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = _b_utils._EngineIntScalarType.to_py(_nfv)
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _b_tsa.uni_arima_forecast_state(
        _st,
        _nst,
        _mr,
        _par,
        _npar,
        c,
        rms,
        _nfv,
        _fva,
        _fsd,
    )
    _fva_py = _fva
    _fsd_py = _fsd
    _return_data = UniArimaForecastStateReturnData(
        _fva_py,
        _fsd_py,
    )
    return _return_data

UniArimaForcecastReturnData = _collections.namedtuple(
    'UniArimaForcecastReturnData',
    ['rms', 'st', 'nst', 'fva', 'fsd', 'isf'],
)

def uni_arima_forcecast(mr, par, c, kfc, x, ist, nfv, ifv):
    u"""
    Univariate time series, state set and forecasts, from fully
    specified seasonal ARIMA model.

    ``uni_arima_forcecast`` applies a fully specified seasonal ARIMA
    model to an observed time series, generates the state set for
    forecasting and (optionally) derives a specified number of forecasts
    together with their standard deviations.

    For full information please refer to the NAG Library document for
    g13aj

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ajf.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, array-like, shape (npar)
        The p values of the phi parameters, the q values of the theta
        parameters, the P values of the Phi parameters, and the Q values
        of the Theta parameters, in that order.

    c : float
        c, the expected value of the differenced series (i.e., c is the
        constant correction). Where there is no constant term, `c` must
        be set to 0.0.

    kfc : int
        Must be set to 0 if `c` was not estimated, and 1 if `c` was
        estimated. This is irrespective of whether or not `c` = 0.0. The
        only effect is that the residual degrees of freedom are one
        greater when `kfc` = 0. Assuming the supplied time series to be
        the same as that to which the model was originally fitted, this
        ensures an unbiased estimate of the residual mean-square.

    x : float, array-like, shape (nx)
        The n values of the original undifferenced time series.

    ist : int
        The dimension of the array `st`.

    nfv : int
        The required number of forecasts. If `nfv` <= 0, no forecasts
        will be computed.

    ifv : int
        The dimension of the arrays `fva` and `fsd`.

    Returns
    -------
    rms : float
        The residual variance (mean square) associated with the model.

    st : float, ndarray, shape (`ist`)
        The `nst` values of the state set.

    nst : int
        The number of values in the state set array `st`.

    fva : float, ndarray, shape (`ifv`)
        If `nfv` > 0, `fva` contains the `nfv` forecast values relating
        to the original undifferenced time series.

    fsd : float, ndarray, shape (`ifv`)
        If `nfv` > 0, `fsd` contains the estimated standard errors of
        the `nfv` forecast values.

    isf : int, ndarray, shape (4)
        Contains validity indicators, one for each of the four possible
        parameter types in the model (autoregressive, moving average,
        seasonal autoregressive, seasonal moving average), in that
        order.

        Each indicator has the interpretation:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 4)
            On entry, `ist` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ist` >= (P*s)+d+(D*s)+q+max(p, Q*s).

        (`errno` 5)
            Unable to calculate the latest estimates of the
            backforecasts.

        (`errno` 6)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

        (`errno` 7)
            On entry, `ifv` = *<value>* and `nfv` = *<value>*.

            Constraint: `ifv` >= max(1, `nfv`).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The time series x_1,x_2,...,x_n supplied to the function is assumed
    to follow a seasonal autoregressive integrated moving average
    (ARIMA) model with known parameters.

    The model is defined by the following relations.

    (a) nabla^d nabla_s^Dx_t-c = w_t where nabla^d nabla_s^Dx_t is the
        result of applying non-seasonal differencing of order d and
        seasonal differencing of seasonality s and order D to the series
        x_t, and c is a constant.

    (#) w_t = Phi_1 w_{t-s}+Phi_2 w_{t-2*s}+...+Phi_P
        w_{t-P*s}+e_t-Theta_1 e_{t-s}-Theta_2 e_{t-2*s}-...-Theta_Q
        e_{t-Q*s}.

        This equation describes the seasonal structure with seasonal
        period s; in the absence of seasonality it reduces to w_t = e_t.

    (#) e_t = phi_1 e_{t-1}+phi_2 e_{t-2}+...+phi_p e_{t-p}+a_t-theta_1
        a_{t-1}-theta_2 a_{t-2}-...-theta_q a_{t-q}.

        This equation describes the non-seasonal structure.

    Given the series, the constant c, and the model parameters Phi,
    Theta, phi, theta, the function computes the following.

    (a) The state set required for forecasting. This contains the
        minimum amount of information required for forecasting and
        comprises:

        (i) the differenced series w_t, for (N-s*P) <= t <= N;

        (#) the (d+D*s) values required to reconstitute the original
            series x_t from the differenced series w_t;

        (#) the intermediate series e_t, for N-max(p, Q*s) < t <= N;

        (#) the residual series a_t, for (N-q) < t <= N, where N =
            n-(d+D*s).

    (#) A set of L forecasts of x_t and their estimated standard errors,
        s_t, for t = n+1,...,n+L (L may be zero).

        The forecasts and estimated standard errors are generated from
        the state set, and are identical to those that would be produced
        from the same state set by ``uni_arima_forecast_state``.

    Use of ``uni_arima_forcecast`` should be confined to situations in
    which the state set for forecasting is unknown.
    Forecasting from the series requires recalculation of the state set
    and this is relatively expensive.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.uni_arima_forcecast'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _ist = _b_utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _ist_py = _b_utils._EngineIntScalarType.to_py(_ist)
    _st = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ist_py,),
    )
    _ifv = _b_utils._EngineIntScalarType(
        ifv, {'fun_name': fun_name, 'entity': 'ifv'},
    ).to_ctypes()
    _ifv_py = _b_utils._EngineIntScalarType.to_py(_ifv)
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ifv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ifv_py,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    _return_scalars = _b_tsa.uni_arima_forcecast(
        _mr,
        _par,
        _npar,
        c,
        kfc,
        _x,
        _nx,
        _st,
        _ist,
        nfv,
        _fva,
        _fsd,
        _ifv,
        _isf,
    )
    _rms_py = _return_scalars.rms
    _st_py = _st
    _nst_py = _return_scalars.nst
    _fva_py = _fva
    _fsd_py = _fsd
    _isf_py = _isf
    _return_data = UniArimaForcecastReturnData(
        _rms_py,
        _st_py,
        _nst_py,
        _fva_py,
        _fsd_py,
        _isf_py,
    )
    return _return_data

UniSmoothExpReturnData = _collections.namedtuple(
    'UniSmoothExpReturnData',
    ['init', 'fv', 'fse', 'yhat', 'res', 'dv', 'ad'],
)

def uni_smooth_exp(mode, itype, p, param, y, k, init, nf, comm):
    u"""
    Univariate time series, exponential smoothing.

    ``uni_smooth_exp`` performs exponential smoothing using either
    single exponential, double exponential or a Holt--Winters method.

    For full information please refer to the NAG Library document for
    g13am

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13amf.html

    Parameters
    ----------
    mode : int
        Indicates if ``uni_smooth_exp`` is continuing from a previous
        call or, if not, how the initial values are computed.

        `mode` = 0

            Required values for m_0, r_0 and s_{-j}, for j =
            0,1,...,p-1, are supplied in `init`.

        `mode` = 1

            ``uni_smooth_exp`` continues from a previous call using
            values that are supplied in `comm`['r'].

        `mode` = 2

            Required values for m_0, r_0 and s_{-j}, for j =
            0,1,...,p-1, are estimated using the first k observations.

    itype : int
        The smoothing function.

        `itype` = 1

            Single exponential.

        `itype` = 2

            Brown double exponential.

        `itype` = 3

            Linear Holt.

        `itype` = 4

            Additive Holt--Winters.

        `itype` = 5

            Multiplicative Holt--Winters.

    p : int
        If `itype` = 4 or 5, the seasonal order, p, otherwise `p` is not
        referenced.

    param : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2): 1; if `itype`=3: 3; if `itype` in
        (4, 5): 4; otherwise: 0.

        The smoothing parameters.

        If `itype` = 1 or 2, `param`[0] = alpha and any remaining
        elements of `param` are not referenced.

        If `itype` = 3, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = phi and any remaining elements of `param` are not
        referenced.

        If `itype` = 4 or 5, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = beta and `param`[3] = phi.

    y : float, array-like, shape (n)
        The time series.

    k : int
        If `mode` = 2, the number of observations used to initialize the
        smoothing.

        If `mode` != 2, `k` is not referenced.

    init : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype`=1: 1; if `itype` in (2, 3): 2; if `itype` in
        (4, 5): 2+`p`; otherwise: 0.

        If `mode` = 0, the initial values for m_0, r_0 and s_{-j}, for j
        = 0,1,...,p-1, used to initialize the smoothing.

        If `itype` = 1, `init`[0] = m_0 and the remaining elements of
        `init` are not referenced.

        If `itype` = 2 or 3, `init`[0] = m_0 and `init`[1] = r_0 and the
        remaining elements of `init` are not referenced.

        If `itype` = 4 or 5, `init`[0] = m_0, `init`[1] = r_0 and
        `init`[2] to `init`[p+1] hold the values for s_{-j}, for j =
        0,1,...,p-1.

        The remaining elements of `init` are not referenced.

    nf : int
        The number of forecasts required beyond the end of the series.
        Note, the one step ahead forecast is always produced.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``uni_smooth_exp``.

    Returns
    -------
    init : float, ndarray, shape (:)
        If `mode` != 1, the values used to initialize the smoothing.
        These are in the same order as described above.

    fv : float, ndarray, shape (`nf`)
        y^_{t+f}, for f = 1,2,...,`nf`, the next `nf` step forecasts.
        Where t = n, if `mode` != 1, else t is the total number of
        smoothed and forecast values already produced.

    fse : float, ndarray, shape (`nf`)
        The forecast standard errors for the values given in `fv`.

    yhat : float, ndarray, shape (n)
        y^_{t+1}, for t = 1,2,...,n, the one step ahead forecast values,
        with `yhat`[i-1] being the one step ahead forecast of `y`[i-2].

    res : float, ndarray, shape (n)
        The residuals, (y_{t+1}-y^_{t+1}), for t = 1,2,...,n.

    dv : float
        The square root of the mean deviation.

    ad : float
        The mean absolute deviation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1, 2, 3, 4 or 5.

        (`errno` 3)
            On entry, `p` = *<value>*.

            Constraint: if `itype` = 4 or 5, `p` > 1.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: 0.0 <= `param`[*<value>*] <= 1.0.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: if `itype` = 2, 0.0 < `param`[*<value>*] <= 1.0.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: `param`[*<value>*] >= 0.0.

        (`errno` 5)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 6)
            A multiplicative Holt--Winters model cannot be used with the
            supplied data.

        (`errno` 7)
            On entry, `k` = *<value>* and n = *<value>*.

            Constraint: if `mode` = 2 and `itype` = 4 or 5, 1 <= `k` <=
            n.

        (`errno` 7)
            On entry, `k` = *<value>*, 2*`p` = *<value>*.

            Constraint: if `mode` = 2 and `itype` = 4 or 5, 2*`p` <=
            `k`.

        (`errno` 9)
            On entry, `nf` = *<value>*.

            Constraint: `nf` >= 0.

        (`errno` 16)
            On entry, the array `comm`['r'] has not been initialized
            correctly.

    Notes
    -----
    Exponential smoothing is a relatively simple method of short term
    forecasting for a time series. ``uni_smooth_exp`` provides five
    types of exponential smoothing; single exponential, Brown's double
    exponential, linear Holt (also called double exponential smoothing
    in some references), additive Holt--Winters and multiplicative
    Holt--Winters.
    The choice of smoothing method used depends on the characteristics
    of the time series.
    If the mean of the series is only slowly changing then single
    exponential smoothing may be suitable.
    If there is a trend in the time series, which itself may be slowly
    changing, then double exponential smoothing may be suitable.
    If there is a seasonal component to the time series, e.g., daily or
    monthly data, then one of the two Holt--Winters methods may be
    suitable.

    For a time series y_t, for t = 1,2,...,n, the five smoothing
    functions are defined by the following:

        Single Exponential Smoothing

                [table omitted]

        Brown Double Exponential Smoothing

                [table omitted]

        Linear Holt Smoothing

                [table omitted]

        Additive Holt--Winters Smoothing

                [table omitted]

        Multiplicative Holt--Winters Smoothing

                [table omitted]

        and psi is defined as in the additive Holt--Winters smoothing,

    where m_t is the mean, r_t is the trend and s_t is the seasonal
    component at time t with p being the seasonal order.
    The f-step ahead forecasts are given by y^_{t+f} and their variances
    by var(y^_{t+f}).
    The term var(epsilon_t) is estimated as the mean deviation.

    The parameters, alpha, beta and gamma control the amount of
    smoothing.
    The nearer these parameters are to one, the greater the emphasis on
    the current data point.
    Generally these parameters take values in the range 0.1 to 0.3.
    The linear Holt and two Holt--Winters smoothers include an
    additional parameter, phi, which acts as a trend dampener.
    For 0.0 < phi < 1.0 the trend is dampened and for phi > 1.0 the
    forecast function has an exponential trend, phi = 0.0 removes the
    trend term from the forecast function and phi = 1.0 does not dampen
    the trend.

    For all methods, values for alpha, beta, gamma and psi can be chosen
    by trying different values and then visually comparing the results
    by plotting the fitted values along side the original data.
    Alternatively, for single exponential smoothing a suitable value for
    alpha can be obtained by fitting an ARIMA(0,1,1) model (see
    ``multi_inputmod_estim``).
    For Brown's double exponential smoothing and linear Holt smoothing
    with no dampening, (i.e., phi = 1.0), suitable values for alpha and
    gamma can be obtained by fitting an ARIMA(0,2,2) model.
    Similarly, the linear Holt method, with phi != 1.0, can be expressed
    as an ARIMA(1,2,2) model and the additive Holt--Winters, with no
    dampening, (phi = 1.0), can be expressed as a seasonal ARIMA model
    with order p of the form ARIMA(0,1,p+1)(0,1,0).
    There is no similar procedure for obtaining parameter values for the
    multiplicative Holt--Winters method, or the additive Holt--Winters
    method with phi != 1.0.
    In these cases parameters could be selected by minimizing a measure
    of fit using one of the nonlinear optimization functions in
    submodule ``opt``.

    In addition to values for alpha, beta, gamma and psi, initial
    values, m_0, r_0 and s_{-j}, for j = 0,1,...,p-1, are required to
    start the smoothing process.
    You can either supply these or they can be calculated by
    ``uni_smooth_exp`` from the first k observations.
    For single exponential smoothing the mean of the observations is
    used to estimate m_0.
    For Brown double exponential smoothing and linear Holt smoothing, a
    simple linear regression is carried out with the series as the
    dependent variable and the sequence 1,2,...,k as the independent
    variable.
    The intercept is then used to estimate m_0 and the slope to estimate
    r_0.
    In the case of the additive Holt--Winters method, the same
    regression is carried out, but a separate intercept is used for each
    of the p seasonal groupings.
    The slope gives an estimate for r_0 and the mean of the p intercepts
    is used as the estimate of m_0.
    The seasonal parameters s_{-j}, for j = 0,1,...,p-1, are estimated
    as the p intercepts -- m_0.
    A similar approach is adopted for the multiplicative Holt--Winter's
    method.

    One step ahead forecasts, y^_{t+1} are supplied along with the
    residuals computed as (y_{t+1}-y^_{t+1}).
    In addition, two measures of fit are provided.
    The mean absolute deviation,

        1/n sum(|y_t-y^_t|)

    and the square root of the mean deviation

        sqrt(1/n sum((y_t-y^_t)^2)).

    References
    ----------
    Chatfield, C, 1980, `The Analysis of Time Series`, Chapman and Hall
    """

    fun_name = 'naginterfaces.library.tsa.uni_smooth_exp'

    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _init = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'init'},
        init,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nf = _b_utils._EngineIntScalarType(
        nf, {'fun_name': fun_name, 'entity': 'nf'},
    ).to_ctypes()
    _nf_py = _b_utils._EngineIntScalarType.to_py(_nf)
    _fv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nf_py,),
    )
    _fse = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nf_py,),
    )
    _yhat = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_tsa.uni_smooth_exp(
        mode,
        itype,
        p,
        _param,
        _n,
        _y,
        k,
        _init,
        _nf,
        _fv,
        _fse,
        _yhat,
        _res,
        comm,
    )
    _init_py = _init
    _fv_py = _fv
    _fse_py = _fse
    _yhat_py = _yhat
    _res_py = _res
    _dv_py = _return_scalars.dv
    _ad_py = _return_scalars.ad
    _return_data = UniSmoothExpReturnData(
        _init_py,
        _fv_py,
        _fse_py,
        _yhat_py,
        _res_py,
        _dv_py,
        _ad_py,
    )
    return _return_data

UniArimaResidReturnData = _collections.namedtuple(
    'UniArimaResidReturnData',
    ['r', 'rcm', 'chi', 'idf', 'siglev'],
)

def uni_arima_resid(v, mr, m, par, ishow, io_manager=None):
    u"""
    Univariate time series, diagnostic checking of residuals, following
    ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    ``uni_arima_resid`` is a diagnostic checking function suitable for
    use after fitting a Box--Jenkins ARMA model to a univariate time
    series using ``uni_arima_estim`` or ``uni_arima_estim_easy``. The
    residual autocorrelation function is returned along with an estimate
    of its asymptotic standard errors and correlations.
    Also, ``uni_arima_resid`` calculates the Box--Ljung portmanteau
    statistic and its significance level for testing model adequacy.

    For full information please refer to the NAG Library document for
    g13as

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13asf.html

    Parameters
    ----------
    v : float, array-like, shape (n)
        `v`[t-1] must contain an estimate of epsilon_t, for t =
        1,2,...,n.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim`` then the actual argument `v` must be
        exr[icount[0]] as returned by ``uni_arima_estim``.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim_easy`` then the actual argument `v` must be
        res as returned by ``uni_arima_estim_easy``.

    mr : int, array-like, shape (7)
        The orders vector (p, d, q, P, D, Q, s) as supplied to
        ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    m : int
        The value of m, the number of residual autocorrelations to be
        computed. See [Further Comments] for advice on the value of `m`.

    par : float, array-like, shape (npar)
        The parameter estimates in the order phi_1,phi_2,...,phi_p,
        theta_1,theta_2,...,theta_q, Phi_1,Phi_2,...,Phi_P,
        Theta_1,Theta_2,...,Theta_Q only.

    ishow : int
        Must be nonzero if the residual autocorrelations, their standard
        errors and the portmanteau statistics are to be printed and zero
        otherwise.

        These quantities are available also as output variables in `r`,
        `rcm`, `chi`, `idf` and `siglev`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    r : float, ndarray, shape (`m`)
        An estimate of the residual autocorrelation coefficient at lag
        l, for l = 1,2,...,m. If `errno` = 3 on exit then all elements
        of `r` are set to zero.

    rcm : float, ndarray, shape (`m`, `m`)
        The estimated standard errors and correlations of the elements
        in the array `r`. The correlation between `r`[i-1] and `r`[j-1]
        is returned as `rcm`[i-1,j-1] except that if i = j then
        `rcm`[i-1,j-1] contains the standard error of `r`[i-1]. If on
        exit, `errno` >= 5, then all off-diagonal elements of `rcm` are
        set to zero and all diagonal elements are set to 1/sqrt(n).

    chi : float
        The value of the portmanteau statistic, Q_(m). If `errno` = 3 on
        exit then `chi` is returned as zero.

    idf : int
        The number of degrees of freedom of `chi`.

    siglev : float
        The significance level of `chi` based on `idf` degrees of
        freedom. If `errno` = 3 on exit, `siglev` is returned as one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` < n.

        (`errno` 1)
            On entry, `m` = *<value>* and npar = *<value>*.

            Constraint: `m` > npar.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar != 0 and npar = p+q+P+Q.

        (`errno` 1)
            On entry, `mr`[3] = *<value>* and `mr`[5] = *<value>*.

            Constraint: `mr`[3] = 0 and `mr`[5] = 0.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*.

            Constraint: `mr`[6] >= 0.

        (`errno` 1)
            On entry, `mr`[5] = *<value>*.

            Constraint: `mr`[5] >= 0.

        (`errno` 1)
            On entry, `mr`[3] = *<value>*.

            Constraint: `mr`[3] >= 0.

        (`errno` 1)
            On entry, `mr`[2] = *<value>*.

            Constraint: `mr`[2] >= 0.

        (`errno` 1)
            On entry, `mr`[0] = *<value>*.

            Constraint: `mr`[0] >= 0.

        (`errno` 2)
            On entry, the seasonal moving average operator is
            non-invertible.

        (`errno` 2)
            On entry, the seasonal autoregressive operator is
            nonstationary.

        (`errno` 2)
            On entry, the non-seasonal moving average operator is
            non-invertible.

        (`errno` 2)
            On entry, the non-seasonal autoregressive operator is
            nonstationary.

        (`errno` 4)
            Excessive iterations needed to find zeros of determinental
            polynomials.

        (`errno` 5)
            On entry, one or more of the AR operators has a factor in
            common with one or more of the MA operators.

        (`errno` 6)
            The matrix `rcm` could not be computed because one of its
            diagonal elements was found to be non-positive.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, the elements of `v` are nearly identical giving
            near-zero variance.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Consider the univariate multiplicative autoregressive-moving average
    model

        phi(B)Phi(B^s)(W_t-mu) = theta(B)Theta(B^s)epsilon_t

    where W_t, for t = 1,2,...,n, denotes a time series and epsilon_t,
    for t = 1,2,...,n, is a residual series assumed to be normally
    distributed with zero mean and variance sigma^2 (> 0).
    The epsilon_t's are also assumed to be uncorrelated.
    Here mu is the overall mean term, s is the seasonal period and B is
    the backward shift operator such that B^rW_t = W_{t-r}.
    The polynomials in (1) are defined as follows:

        phi(B) = 1-phi_1 B-phi_2 B^2-...-phi_p B^p

    is the non-seasonal autoregressive (AR) operator;

        theta(B) = 1-theta_1 B-theta_2 B^2-...-theta_q B^q

    is the non-seasonal moving average (MA) operator;

        Phi(B^s) = 1-Phi_1 B^s-Phi_2 B^(2s)-...-Phi_P B^(Ps)

    is the seasonal AR operator; and

        Theta(B^s) = 1-Theta_1 B^s-Theta_2 B^(2s)-...-Theta_Q B^(Qs)

    is the seasonal MA operator.
    The model (1) is assumed to be stationary, that is the zeros of
    phi(B) and Phi(B^s) are assumed to lie outside the unit circle.
    The model (1) is also assumed to be invertible, that is the zeros of
    theta(B) and Theta(B^s) are assumed to lie outside the unit circle.
    When both Phi(B^s) and Theta(B^s) are absent from the model, that is
    when P = Q = 0, then the model is said to be non-seasonal.

    The estimated residual autocorrelation coefficient at lag l, r^_l,
    is computed as:

        r^_l = sum_t =
        l+1^n(epsilon^_{t-l}-epsilon_)(epsilon^_t-epsilon_)/sum_t =
        1^n(epsilon^_t-epsilon_)^2, l = 1,2,...

    where epsilon^_t denotes an estimate of the tth residual, epsilon_t,
    and epsilon_= sum_t = 1^n epsilon^_t/n.
    A portmanteau statistic, Q_(m), is calculated from the formula (see
    Box and Ljung (1978)):

        Q_(m) = n(n+2)sum_l = 1^mr^_l^2/(n-l)

    where m denotes the number of residual autocorrelations computed.
    (Advice on the choice of m is given in [Further Comments].) Under
    the hypothesis of model adequacy, Q_(m) has an asymptotic
    chi^2-distribution on m-p-q-P-Q degrees of freedom.
    Let r^^T = (r^_1,r^_2,...,r^_m) then the variance-covariance matrix
    of r^ is given by:

        Var(r^) = [I_m-X(X^TX)^-1X^T]/n.

    The construction of the matrix X is discussed in McLeod (1978).
    (Note that the mean, mu, and the residual variance, sigma^2, play no
    part in calculating Var(r^) and, therefore, are not required as
    input to ``uni_arima_resid``.)

    **Note:** for additive models with fixed parameter values (i.e.,
    fitted by ``multi_varma_estimate``) ``multi_varma_diag`` should be
    used instead of ``uni_arima_resid``.

    References
    ----------
    Box, G E P and Ljung, G M, 1978, `On a measure of lack of fit in
    time series models`, Biometrika (65), 297--303

    McLeod, A I, 1978, `On the distribution of the residual
    autocorrelations in Box--Jenkins models`, J. Roy. Statist. Soc. Ser.
    B (40), 296--302
    """

    fun_name = 'naginterfaces.library.tsa.uni_arima_resid'

    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _rcm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = _b_tsa.uni_arima_resid(
            _n,
            _v,
            _mr,
            _m,
            _par,
            _npar,
            ishow,
            _r,
            _rcm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _rcm_py = _rcm
    _chi_py = _return_scalars.chi
    _idf_py = _return_scalars.idf
    _siglev_py = _return_scalars.siglev
    _return_data = UniArimaResidReturnData(
        _r_py,
        _rcm_py,
        _chi_py,
        _idf_py,
        _siglev_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniMeansReturnData = _collections.namedtuple(
    'UniMeansReturnData',
    ['y', 'mean'],
)

def uni_means(z, m, rs):
    u"""
    Computes quantities needed for range-mean or standard deviation-mean
    plot.

    ``uni_means`` calculates the range (or standard deviation) and the
    mean for groups of successive time series values.
    It is intended for use in the construction of range-mean plots.

    For full information please refer to the NAG Library document for
    g13au

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13auf.html

    Parameters
    ----------
    z : float, array-like, shape (n)
        `z`[t-1] must contain the tth observation Z_t, for t =
        1,2,...,n.

    m : int
        m, the group size.

    rs : str, length 1
        Indicates whether ranges or standard deviations are to be
        calculated.

        `rs` = 'R'

            Ranges are calculated.

        `rs` = 'S'

            Standard deviations are calculated.

    Returns
    -------
    y : float, ndarray, shape (int((n/`m`)))
        `y`[i-1] contains the range or standard deviation, as determined
        by `rs`, of the ith group of observations, for i = 1,2,...,k.

    mean : float, ndarray, shape (int((n/`m`)))
        `mean`[i-1] contains the mean of the ith group of observations,
        for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*, `m` = *<value>* and ngrps =
            *<value>*.

            Constraint: ngrps = int(n/`m`).

        (`errno` 1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 2)
            On entry, `rs` = *<value>*.

            Constraint: `rs` = 'R' or 'S'.

    Notes
    -----
    Let Z_1,Z_2,...,Z_n denote n successive observations in a time
    series.
    The series may be divided into groups of m successive values and for
    each group the range or standard deviation (depending on a
    user-supplied option) and the mean are calculated.
    If n is not a multiple of m then groups of equal size m are found
    starting from the end of the series of observations provided, and
    any remaining observations at the start of the series are ignored.
    The number of groups used, k, is the integer part of n/m.
    If you wish to ensure that no observations are ignored then the
    number of observations, n, should be chosen so that n is divisible
    by m.

    The mean, M_i, the range, R_i, and the standard deviation, S_i, for
    the ith group are defined as

        M_i = 1/msum_j = 1^mZ_{l+m(i-1)+j};;R_i = max_{1 <= j <=
        m}{Z_{l+m(i-1)+j}}-min_{1 <= j <= m}{Z_{l+m(i-1)+j}}

    and

        S_i = sqrt((1/(m-1))sum_j = 1^m(Z_{l+m(i-1)+j}-M_i)^2)

    where l = n-km, the number of observations ignored.

    For seasonal data it is recommended that m should be equal to the
    seasonal period.
    For non-seasonal data the recommended group size is 8.

    A plot of range against mean or of standard deviation against mean
    is useful for finding a transformation of the series which makes the
    variance constant.
    If the plot appears random or the range (or standard deviation)
    seems to be constant irrespective of the mean level then this
    suggests that no transformation of the time series is called for.
    On the other hand an approximate linear relationship between range
    (or standard deviation) and mean would indicate that a log
    transformation is appropriate.
    Further details may be found in either Jenkins (1979) and McLeod
    (1982).

    You have the choice of whether to use the range or the standard
    deviation as a measure of variability.
    If the group size is small they are both equally good but if the
    group size is fairly large (e.g., m = 12 for monthly data) then the
    range may not be as good an estimate of variability as the standard
    deviation.

    References
    ----------
    Jenkins, G M, 1979, `Practical Experiences with Modelling and
    Forecasting Time Series`, GJP Publications, Lancaster

    McLeod, G, 1982, `Box--Jenkins in Practice. 1: Univariate Stochastic
    and Single Output Transfer Function/Noise Analysis`, GJP
    Publications, Lancaster
    """

    fun_name = 'naginterfaces.library.tsa.uni_means'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _ngrps = _b_utils._EngineIntScalarType(
        int((_n_py//_m_py)), {'fun_name': fun_name, 'entity': '_ngrps'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (int((_n_py//_m_py)),),
    )
    _mean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (int((_n_py//_m_py)),),
    )
    _b_tsa.uni_means(
        _n,
        _z,
        _m,
        _ngrps,
        rs,
        _y,
        _mean,
    )
    _y_py = _y
    _mean_py = _mean
    _return_data = UniMeansReturnData(
        _y_py,
        _mean_py,
    )
    return _return_data

UniDickeyFullerUnitReturnData = _collections.namedtuple(
    'UniDickeyFullerUnitReturnData',
    ['ts'],
)

def uni_dickey_fuller_unit(ut_type, p, y):
    u"""
    Computes (augmented) Dickey--Fuller unit root test statistic.

    ``uni_dickey_fuller_unit`` returns the (augmented) Dickey--Fuller
    unit root test.

    For full information please refer to the NAG Library document for
    g13aw

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13awf.html

    Parameters
    ----------
    ut_type : int
        The type of unit test for which the probability is required.

        `ut_type` = 1

            A unit root test will be performed and tau returned.

        `ut_type` = 2

            A unit root test with drift will be performed and tau_mu
            returned.

        `ut_type` = 3

            A unit root test with drift and deterministic time trend
            will be performed and tau_tau returned.

    p : int
        p, the degree of the autoregressive (AR) component of the
        Dickey--Fuller test statistic. When p > 1 the test is usually
        referred to as the augmented Dickey--Fuller test.

    y : float, array-like, shape (n)
        y, the time series.

    Returns
    -------
    ts : float
        The (augmented) Dickey--Fuller test statistic: tau when
        `ut_type` = 1; tau_mu when `ut_type` = 2 or tau_tau when
        `ut_type` = 3.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ut_type` = *<value>*.

            Constraint: `ut_type` = 1, 2 or 3.

        (`errno` 21)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.

        (`errno` 31)
            On entry, n = *<value>*.

            Constraint: if `ut_type` = 1, n > 2*`p` if `ut_type` = 2, n
            > 2*`p`+1 if `ut_type` = 3, n > 2*`p`+2

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 41)
            On entry, the design matrix used in the estimation of beta_1
            is not of full rank, this is usually due to all elements of
            the series being virtually identical. The returned statistic
            is, therefore, not unique and likely to be meaningless.

        (`errno` 42)
            standard error of the parameter estimate is zero, therefore,
            depending on the sign of the parameter estimate, a large
            positive or negative value has been returned.

    Notes
    -----
    If the root of the characteristic equation for a time series is one
    then that series is said to have a unit root.
    Such series are nonstationary. ``uni_dickey_fuller_unit`` returns
    one of three types of (augmented) Dickey--Fuller test statistic:
    tau, tau_mu or tau_tau, used to test for a unit root, a unit root
    with drift or a unit root with drift and a deterministic time trend,
    respectively.

    To test whether a time series, y_t, for t = 1,2,...,n, has a unit
    root, the regression model

        nabla y_t = beta_1 y_{t-1}+sum_i = 1^p-1 delta_inabla
        y_{t-i}+epsilon_t

    is fitted and the test statistic tau constructed as

        tau = (beta^_1)/(sigma_{11})

    where nabla is the difference operator, with nabla y_t =
    y_t-y_{t-1}, and where beta^_1 and sigma_{11} are the least squares
    estimate and associated standard error for beta_1 respectively.

    To test for a unit root with drift the regression model

        nabla y_t = beta_1 y_{t-1}+sum_i = 1^p-1 delta_inabla
        y_{t-i}+alpha+epsilon_t

    is fit and the test statistic tau_mu constructed as

        tau_mu = (beta^_1)/(sigma_{11})

    To test for a unit root with drift and deterministic time trend the
    regression model

        nabla y_t = beta_1 y_{t-1}+sum_i = 1^p-1 delta_inabla
        y_{t-i}+alpha+beta_2 t+epsilon_t

    is fit and the test statistic tau_tau constructed as

        tau_tau = (beta^_1)/(sigma_{11})

    The distributions of the three test statistics; tau, tau_mu and
    tau_tau, are nonstandard.
    An associated probability can be obtained from
    ``stat.prob_dickey_fuller_unit``.

    References
    ----------
    Dickey, A D, 1976, `Estimation and hypothesis testing in
    nonstationary time series`, PhD Thesis, Iowa State University, Ames,
    Iowa

    Dickey, A D and Fuller, W A, 1979, `Distribution of the estimators
    for autoregressive time series with a unit root`, J. Am. Stat.
    Assoc. (74 366), 427--431

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_dickey_fuller_unit_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.uni_dickey_fuller_unit'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_tsa.uni_dickey_fuller_unit(
            ut_type,
            p,
            _n,
            _y,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = UniDickeyFullerUnitReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def multi_filter_arima(y, mr, par, cy=None):
    u"""
    Multivariate time series, filtering (pre-whitening) by an ARIMA
    model.

    ``multi_filter_arima`` filters a time series by an ARIMA model.

    For full information please refer to the NAG Library document for
    g13ba

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13baf.html

    Parameters
    ----------
    y : float, array-like, shape (ny)
        The Q'_y backforecasts, starting with backforecast at time
        1-Q'_y to backforecast at time 0, followed by the time series
        starting at time 1, where Q'_y = `mr`[9]+`mr`[12]*`mr`[13]. If
        there are no backforecasts, either because the ARIMA model for
        the time series is not known, or because it is known but has no
        moving average terms, then the time series starts at the
        beginning of `y`.

    mr : int, array-like, shape (nmr)
        The orders vector for the filtering model, followed by the
        orders vector for the ARIMA model for the time series if the
        latter is known. The orders appear in the standard sequence
        (p,d,q,P,D,Q,s) as given in [the G13 Introduction]. If the ARIMA
        model for the time series is supplied, the function will assume
        that the first Q'_y values of the array `y` are backforecasts.

    par : float, array-like, shape (npar)
        The parameters of the filtering model, followed by the
        parameters of the ARIMA model for the time series, if supplied.
        Within each model the parameters are in the standard order of
        non-seasonal AR and MA followed by seasonal AR and MA.

    cy : None or float, optional
        If the ARIMA model is known (i.e., nmr = 14), `cy` must specify
        the constant term of the ARIMA model for the time series. If
        this model is not known (i.e., nmr = 7), `cy` is not used.

    Returns
    -------
    b : float, ndarray, shape (:)
        The filtered output series. If the ARIMA model for the time
        series was known, and hence Q'_y backforecasts were supplied in
        `y`, then `b` contains Q'_y 'filtered' backforecasts followed by
        the filtered series. Otherwise, the filtered series begins at
        the start of `b` just as the original series began at the start
        of `y`. In either case, if the value of the series at time t is
        held in `y`[t-1], then the filtered value at time t is held in
        `b`[t-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nmr = *<value>*.

            Constraint: nmr = 7 or 14.

        (`errno` 2)
            On entry, the orders vector `mr` is invalid.

        (`errno` 3)
            On entry, npar = *<value>*.

            Constraint: npar must be inconsistent with `mr`.

        (`errno` 4)
            On entry, ny = *<value>* and the minimum size required =
            *<value>*.

            Constraint: ny >= max(1+Q'_y, npar).

        (`errno` 7)
            The orders vector for the filtering model is invalid.

        (`errno` 8)
            The orders vector for the ARIMA model is invalid.

        (`errno` 9)
            The initial values of the filtered series are indeterminate
            for the given models.

    Notes
    -----
    From a given series y_1,y_2,...,y_n, a new series b_1,b_2,...,b_n is
    calculated using a supplied (filtering) ARIMA model.
    This model will be one which has previously been fitted to a series
    x_t with residuals a_t.
    The equations defining b_t in terms of y_t are very similar to those
    by which a_t is obtained from x_t.
    The only dissimilarity is that no constant correction is applied
    after differencing.
    This is because the series y_t is generally distinct from the series
    x_t with which the model is associated, though y_t may be related to
    x_t.
    Whilst it is appropriate to apply the ARIMA model to y_t so as to
    preserve the same relationship between b_t and a_t as exists between
    y_t and x_t, the constant term in the ARIMA model is inappropriate
    for y_t.
    The consequence is that b_t will not necessarily have zero mean.

    The equations are precisely:

        w_t = nabla^d nabla_s^Dy_t,

    the appropriate differencing of y_t; both the seasonal and
    non-seasonal inverted autoregressive operations are then applied,

        u_t = w_t-Phi_1 w_{t-s}-...-Phi_P w_{t-s*P}

        v_t = u_t-phi_1 u_{t-1}-...-phi_p u_{t-p}

    followed by the inverted moving average operations

        z_t = v_t+Theta_1 z_{t-s}+...+Theta_Q z_{t-s*Q}

        b_t = z_t+theta_1 b_{t-1}+...+theta_q b_{t-q}.

    Because the filtered series value b_t depends on present and past
    values y_t,y_{t-1},..., there is a problem arising from ignorance of
    y_0,y_-1,... which particularly affects calculation of the early
    values b_1,b_2,..., causing 'transient errors'.
    The function allows two possibilities.

    (i) The equations (1), (2) and (3) are applied from successively
        later time points so that all terms on their right-hand sides
        are known, with v_t being defined for t = (1+d+s*D+s*P),...,n.
        Equations (4) and (5) are then applied over the same range,
        taking any values on the right-hand side associated with
        previous time points to be zero.

        This procedure may still however result in unacceptably large
        transient errors in early values of b_t.

    (#) The unknown values y_0,y_-1,... are estimated by
        backforecasting. This requires that an ARIMA model distinct from
        that which has been supplied for filtering, should have been
        previously fitted to y_t.

    For efficiency, you are asked to supply both this ARIMA model for
    y_t and a limited number of backforecasts which are prefixed to the
    known values of y_t.
    Within the function further backforecasts of y_t, and the series
    w_t, u_t, v_t in (1), (2) and (3) are then easily calculated, and a
    set of linear equations solved for backforecasts of z_t,b_t for use
    in (4) and (5) in the case that q+Q > 0.

    Even if the best model for y_t is not available, a very approximate
    guess such as

        y_t = c+e_t

    or

        nabla y_t = e_t

    can help to reduce the transients substantially.

    The backforecasts which need to be prefixed to y_t are of length
    Q'_y = q_y+s_y*Q_y, where q_y and Q_y are the non-seasonal and
    seasonal moving average orders and s_y the seasonal period for the
    ARIMA model of y_t.
    Thus you need not carry out the backforecasting exercise if Q'_y =
    0.
    Otherwise, the series y_1,y_2,...,y_n should be reversed to obtain
    y_n,y_{n-1},...,y_1 and ``uni_arima_forcecast`` should be used to
    forecast Q'_y values, y^_0,...,y^_{1-Q'_y}.
    The ARIMA model used is that fitted to y_t (as a forward series)
    except that, if d_y+D_y is odd, the constant should be changed in
    sign (to allow, for example, for the fact that a forward upward
    trend is a reversed downward trend).
    The ARIMA model for y_t supplied to the filtering function must
    however have the appropriate constant for the forward series.

    The series y^_{1-Q'_y},...,y^_0,y_1,...,y_n is then supplied to the
    function, and a corresponding set of values returned for b_t.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_filter_arima'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ny_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _ny = _b_utils._EngineIntScalarType(
        _ny_py, {'fun_name': fun_name, 'entity': '_ny'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _nmr_py = _b_utils._EngineIntArrayType.get_shape(_mr, {'fun_name': fun_name, 'entity': 'mr'}, exp_rank=1)[0]
    _nmr = _b_utils._EngineIntScalarType(
        _nmr_py, {'fun_name': fun_name, 'entity': '_nmr'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    if _nmr_py == 14:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))), {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    elif _nmr_py == 7:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py, {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    else:
        _nb = _b_utils._EngineIntScalarType.c_type(0) # Added by code generator.
        _nb_py = 0 # Added by code generator.
    if _nmr_py == 14:
        _b_dim1 = _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)))
    elif _nmr_py == 7:
        _b_dim1 = _ny_py
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1,)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_b,
    )
    _b_tsa.multi_filter_arima(
        _y,
        _ny,
        _mr,
        _nmr,
        _par,
        _npar,
        cy,
        _b,
        _nb,
    )
    _b_py = _b
    return _b_py

def multi_filter_transf(y, mr, par, cy=None):
    u"""
    Multivariate time series, filtering by a transfer function model.

    ``multi_filter_transf`` filters a time series by a transfer function
    model.

    For full information please refer to the NAG Library document for
    g13bb

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bbf.html

    Parameters
    ----------
    y : float, array-like, shape (ny)
        The Q'_y backforecasts starting with backforecast at time 1-Q'_y
        to backforecast at time 0 followed by the time series starting
        at time 1, where Q'_y = `mr`[5]+`mr`[8]*`mr`[9]. If there are no
        backforecasts either because the ARIMA model for the time series
        is not known or because it is known but has no moving average
        terms, then the time series starts at the beginning of `y`.

    mr : int, array-like, shape (nmr)
        The orders vector for the filtering transfer function model
        followed by the orders vector for the ARIMA model for the time
        series if the latter is known. The transfer function model
        orders appear in the standard form (b,q,p) as given in [the G13
        Introduction]. Note that if the ARIMA model for the time series
        is supplied then the function will assume that the first Q'_y
        values of the array `y` are backforecasts.

    par : float, array-like, shape (npar)
        The parameters of the filtering transfer function model followed
        by the parameters of the ARIMA model for the time series. In the
        transfer function model the parameters are in the standard order
        of MA-like followed by AR-like operator parameters. In the ARIMA
        model the parameters are in the standard order of non-seasonal
        AR and MA followed by seasonal AR and MA.

    cy : None or float, optional
        If the ARIMA model is known (i.e., nmr = 10), `cy` must specify
        the constant term of the ARIMA model for the time series. If
        this model is not known (i.e., nmr = 3) then `cy` is not used.

    Returns
    -------
    b : float, ndarray, shape (:)
        The filtered output series. If the ARIMA model for the time
        series was known, and hence Q'_y backforecasts were supplied in
        `y`, then `b` contains Q'_y 'filtered' backforecasts followed by
        the filtered series. Otherwise, the filtered series begins at
        the start of `b` just as the original series began at the start
        of `y`. In either case, if the value of the series at time t is
        held in `y`[t-1], then the filtered value at time t is held in
        `b`[t-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nb = *<value>* and the minimum size required =
            *<value>*.

            Constraint: if arimas is not None then nb >= max(b, p, q),
            otherwise nb >= ny.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar must be inconsistent with `mr`.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*, `mr`[7] = *<value>* and
            `mr`[8] = *<value>*.

            Constraint: `mr`[6]+`mr`[7]+`mr`[8] != 0.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*, `mr`[7] = *<value>* and
            `mr`[8] = *<value>*.

            Constraint: `mr`[6]+`mr`[7]+`mr`[8] = 0.

        (`errno` 1)
            On entry, `mr`[9] = 1.

            Constraint: `mr`[9] != 1.

        (`errno` 1)
            On entry, i = *<value>*, `mr`[i-1] = *<value>*.

            Constraint: `mr`[i-1] >= 0.

        (`errno` 1)
            On entry, nmr = *<value>*.

            Constraint: nmr = 3 or 10.

        (`errno` 2)
            A supplied model has invalid parameters.

        (`errno` 3)
            The supplied time series is too short.

        (`errno` 4)
            The matrix used to solve for starting values for MA is
            singular.

    Notes
    -----
    From a given series y_1,y_2,...,y_n a new series b_1,b_2,...,b_n is
    calculated using a supplied (filtering) transfer function model
    according to the equation

        b_t = delta_1 b_{t-1}+delta_2 b_{t-2}+...+delta_p
        b_{t-p}+omega_0 y_{t-b}-omega_1 y_{t-b-1}-...-omega_q y_{t-b-q}.

    As in the use of ``multi_filter_arima``, large transient errors may
    arise in the early values of b_t due to ignorance of y_t for t < 0,
    and two possibilities are allowed.

    (i) The equation (1) is applied from t = 1+b+q,...,n so all terms in
        y_t on the right-hand side of (1) are known, the unknown set of
        values b_t for t = b+q,...,b+q+1-p being taken as zero.

    (#) The unknown values of y_t for t <= 0 are estimated by
        backforecasting exactly as for ``multi_filter_arima``.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_filter_transf'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ny_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _ny = _b_utils._EngineIntScalarType(
        _ny_py, {'fun_name': fun_name, 'entity': '_ny'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _nmr_py = _b_utils._EngineIntArrayType.get_shape(_mr, {'fun_name': fun_name, 'entity': 'mr'}, exp_rank=1)[0]
    _nmr = _b_utils._EngineIntScalarType(
        _nmr_py, {'fun_name': fun_name, 'entity': '_nmr'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    if _nmr_py == 3:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py, {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    elif _nmr_py == 10:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))), {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    else:
        _nb = _b_utils._EngineIntScalarType.c_type(0) # Added by code generator.
        _nb_py = 0 # Added by code generator.
    if _nmr_py == 3:
        _b_dim1 = _ny_py
    elif _nmr_py == 10:
        _b_dim1 = _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,)))
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1,)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_b,
    )
    _b_tsa.multi_filter_transf(
        _y,
        _ny,
        _mr,
        _nmr,
        _par,
        _npar,
        cy,
        _b,
        _nb,
    )
    _b_py = _b
    return _b_py

MultiXcorrReturnData = _collections.namedtuple(
    'MultiXcorrReturnData',
    ['s', 'r0', 'r', 'stat'],
)

def multi_xcorr(x, y, nl):
    u"""
    Multivariate time series, cross-correlations.

    ``multi_xcorr`` calculates cross-correlations between two time
    series.

    For full information please refer to the NAG Library document for
    g13bc

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bcf.html

    Parameters
    ----------
    x : float, array-like, shape (nxy)
        The n values of the x series.

    y : float, array-like, shape (nxy)
        The n values of the y series.

    nl : int
        L, the maximum lag for calculating cross-correlations.

    Returns
    -------
    s : float
        The ratio of the standard deviation of the y series to the
        standard deviation of the x series, s_y/s_x.

    r0 : float
        The cross-correlation between the x and y series at lag zero.

    r : float, ndarray, shape (`nl`)
        `r`[l-1] contains the cross-correlations between the x and y
        series at lags L, r_{xy}(l), for l = 1,2,...,L.

    stat : float
        The statistic for testing for absence of cross-correlation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nxy = *<value>*.

            Constraint: nxy > 1.

        (`errno` 1)
            On entry, `nl` = *<value>* and nxy = *<value>*.

            Constraint: 1 <= `nl` < nxy.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` >= 1.

        (`errno` 2)
            One or both of the x and y series have zero variance.

    Notes
    -----
    Given two series x_1,x_2,...,x_n and y_1,y_2,...,y_n the function
    calculates the cross-correlations between x_t and lagged values of
    y_t:

        r_{xy}(l) = sum_t = 1^n-l(x_t-x_)(y_{t+l}-y_)/(ns_x s_y), l =
        0,1,...,L

    where

        x_= sum_t = 1^nx_t/n

        s_x^2 = sum_t = 1^n(x_t-x_)^2/n

    and similarly for y.

    The ratio of standard deviations s_y/s_x is also returned, and a
    portmanteau statistic is calculated:

        `stat` = nsum_l = 1^Lr_{xy}(l)^2.

    Provided n is large, L much less than n, and both x_t,y_t are
    samples of series whose true autocorrelation functions are zero,
    then, under the null hypothesis that the true cross-correlations
    between the series are zero, `stat` has a chi^2-distribution with L
    degrees of freedom.
    Values of `stat` in the upper tail of this distribution provide
    evidence against the null hypothesis.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_xcorr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxy_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nxy = _b_utils._EngineIntScalarType(
        _nxy_py, {'fun_name': fun_name, 'entity': '_nxy'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nl = _b_utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = _b_utils._EngineIntScalarType.to_py(_nl)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    _return_scalars = _b_tsa.multi_xcorr(
        _x,
        _y,
        _nxy,
        _nl,
        _r,
    )
    _s_py = _return_scalars.s
    _r0_py = _return_scalars.r0
    _r_py = _r
    _stat_py = _return_scalars.stat
    _return_data = MultiXcorrReturnData(
        _s_py,
        _r0_py,
        _r_py,
        _stat_py,
    )
    return _return_data

MultiTransfPrelimReturnData = _collections.namedtuple(
    'MultiTransfPrelimReturnData',
    ['wds', 'isf'],
)

def multi_transf_prelim(r0, r, nna, s):
    u"""
    Multivariate time series, preliminary estimation of transfer
    function model.

    ``multi_transf_prelim`` calculates preliminary estimates of the
    parameters of a transfer function model.

    For full information please refer to the NAG Library document for
    g13bd

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bdf.html

    Parameters
    ----------
    r0 : float
        The cross-correlation between the two series at lag 0,
        r_{xy}(0).

    r : float, array-like, shape (nl)
        The cross-correlations between the two series at lags 1 to L,
        r_{xy}(l), for l = 1,2,...,L.

    nna : int, array-like, shape (3)
        The transfer function model orders in the standard form b,q,p
        (i.e., delay time, number of moving-average MA-like followed by
        number of autoregressive AR-like parameters).

    s : float
        The ratio of the standard deviation of the y series to that of
        the x series, s_y/s_x.

    Returns
    -------
    wds : float, ndarray, shape (:)
        The preliminary estimates of the parameters of the transfer
        function model in the order of q+1 MA-like parameters followed
        by the p AR-like parameters. If the estimation of either type of
        parameter fails then these arguments are set to 0.0.

    isf : int, ndarray, shape (2)
        Indicators of the success of the estimation of MA-like and
        AR-like parameters respectively. A value 0 indicates that there
        are no parameters of that type to be estimated. A value of 1 or
        -1 indicates that there are parameters of that type in the model
        and the estimation of that type has been successful or
        unsuccessful respectively. Note that there is always at least
        one MA-like parameter in the model.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nwds = *<value>*, `nna`[1] = *<value>* and
            `nna`[2] = *<value>*.

            Constraint: nwds = `nna`[1]+`nna`[2]+1.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `r`[i-1] = *<value>*.

            Constraint: -1.0 <= `r`[i] <= 1.0.

        (`errno` 1)
            On entry, `r0` = *<value>*.

            Constraint: -1.0 <= `r0` <= 1.0.

        (`errno` 1)
            On entry, nl = *<value>*, `nna`[0] = *<value>*, `nna`[1] =
            *<value>* and `nna`[2] = *<value>*.

            Constraint: nl >= max(`nna`[0]+`nna`[1]+`nna`[2], 1).

        (`errno` 1)
            On entry, i = *<value>* and `nna`[i-1] = *<value>*.

            Constraint: `nna`[i-1] >= 0.

    Notes
    -----
    ``multi_transf_prelim`` calculates estimates of parameters
    delta_1,delta_2,...,delta_p, omega_0,omega_1,...,omega_q in the
    transfer function model

        y_t = delta_1 y_{t-1}+delta_2 y_{t-2}+...+delta_p
        y_{t-p}+omega_0 x_{t-b}-omega_1 x_{t-b-1}-...-omega_q x_{t-b-q}

    given cross-correlations between the series x_t and lagged values of
    y_t:

        r_{xy}(l), l = 0,1,...,L

    and the ratio of standard deviations s_y/s_x, as supplied by
    ``multi_xcorr``.

    It is assumed that the series x_t used to calculate the
    cross-correlations is a sample from a time series with true
    autocorrelations of zero.
    Otherwise the cross-correlations between the series b_t and a_t, as
    defined in the description of ``multi_filter_arima``, should be used
    in place of those between y_t and x_t.

    The estimates are obtained by solving for
    delta_1,delta_2,...,delta_p the equations

        r_{xy}(b+q+j) = delta_1 r_{xy}(b+q+j-1)+...+delta_p
        r_{xy}(b+q+j-p), j = 1,2,...,p

    then calculating

        omega_i = +/-(s_y/s_x)[r_{xy}(b+i)-delta_1
        r_{xy}(b+i-1)-...-delta_p r_{xy}(b+i-p)], i = 0,1,...,q

    where the '+' is used for omega_0 and '-' for omega_i, i > 0.

    Any value of r_{xy}(l) arising in these equations for l < b is taken
    as zero.
    The parameters delta_1,delta_2,...,delta_p are checked as to whether
    they satisfy the stability criterion.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_transf_prelim'

    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nl_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _nl = _b_utils._EngineIntScalarType(
        _nl_py, {'fun_name': fun_name, 'entity': '_nl'},
    ).to_ctypes()
    _nna = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nna'},
        nna,
        copy=False,
    )
    _nwds = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (1,))+_b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (2,))+1, {'fun_name': fun_name, 'entity': '_nwds'},
    ).to_ctypes()
    _wds = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (1,))+_b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (2,))+1,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (2,),
    )
    _b_tsa.multi_transf_prelim(
        r0,
        _r,
        _nl,
        _nna,
        s,
        _nwds,
        _wds,
        _isf,
    )
    _wds_py = _wds
    _isf_py = _isf
    _return_data = MultiTransfPrelimReturnData(
        _wds_py,
        _isf_py,
    )
    return _return_data

MultiInputmodEstimReturnData = _collections.namedtuple(
    'MultiInputmodEstimReturnData',
    ['para', 'xxy', 'zsp', 'itc', 'sd', 'cm', 's', 'd', 'ndf', 'res', 'sttf', 'nsttf'],
)

def multi_inputmod_estim(mr, mt, para, xxy, kfc=1, kef=2, nit=1000, zsp=None, kpriv=0, io_manager=None):
    u"""
    Multivariate time series, estimation of multi-input model.

    ``multi_inputmod_estim`` fits a multi-input model relating one
    output series to the input series with a choice of three different
    estimation criteria: nonlinear least squares, exact likelihood and
    marginal likelihood.
    When no input series are present, ``multi_inputmod_estim`` fits a
    univariate ARIMA model.

    For full information please refer to the NAG Library document for
    g13be

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bef.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function model orders b, p and q of each of the
        input series. The order parameters for input series i are held
        in column i. Row 1 holds the value b_i, row 2 holds the value
        q_i and row 3 holds the value p_i. For a simple input, b_i = q_i
        = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, r_i
        = 2 for a transfer function input for which no allowance is to
        be made for pre-observation period effects, and r_i = 3 for a
        transfer function input for which pre-observation period effects
        will be treated by estimation of appropriate nuisance
        parameters.

        When r_i = 1, any nonzero contents of rows 1, 2, and 3 of column
        i are ignored.

    para : float, array-like, shape (npara)
        Initial values of the multi-input model parameters. These are in
        order, firstly the ARIMA model parameters: p values of phi
        parameters, q values of theta parameters, P values of Phi
        parameters and Q values of Theta parameters. These are followed
        by initial values of the transfer function model parameters
        omega_0,omega_1,...,omega_q_1, delta_1,delta_2,...,delta_p_1 for
        the first of any input series and similarly for each subsequent
        input series. The final component of `para` is the initial value
        of the constant c, whether it is fixed or is to be estimated.

    xxy : float, array-like, shape (nxxy, nser)
        The columns of `xxy` must contain the nxxy original,
        undifferenced values of each of the input series and the output
        series x_t in that order.

    kfc : int, optional
        Must be set to 0 if the constant c is to remain fixed at its
        initial value, and 1 if it is to be estimated.

    kef : int, optional
        Indicates the likelihood option.

        `kef` = 1

            Gives least squares.

        `kef` = 2

            Gives exact likelihood.

        `kef` = 3

            Gives marginal likelihood.

    nit : int, optional
        The maximum required number of iterations.

        `nit` = 0

            No change is made to any of the model parameters in array
            `para` except that the constant c (if `kfc` = 1) and any
            omega relating to simple input series are estimated. (Apart
            from these, estimates are always derived for the nuisance
            parameters relating to any backforecasts and any
            pre-observation period effects for transfer function
            inputs.)

    zsp : None or float, array-like, shape (4), optional
        If `zsp` is not None, then `zsp` must contain the four values
        used to control the strategy of the search procedure.

        `zsp`[0]

            Contains alpha, the value used to constrain the magnitude of
            the search procedure steps.

        `zsp`[1]

            Contains beta, the multiplier which regulates the value of
            alpha.

        `zsp`[2]

            Contains delta, the value of the stationarity and
            invertibility test tolerance factor.

        `zsp`[3]

            Contains gamma, the value of the convergence criterion.

        If `zsp` is None before entry, default values of `zsp` are
        supplied by the function.

        These are 0.01, 10.0, 1000.0 and max(100*machine
        precision,0.0000001), respectively.

    kpriv : int, optional
        Must not be set to 0, if it is required to monitor the course of
        the optimization. The course of the optimization is monitored by
        printing out at each iteration the iteration count (`itc`), the
        residual sum of squares (`s`), the objective function (`d`) and
        a description and value for each of the parameters in the `para`
        array. The descriptions are PHI for phi, THETA for theta, SPHI
        for Phi, STHETA for Theta, OMEGA/SI for omega in a simple input,
        OMEGA for omega in a transfer function input, DELTA for delta
        and CONSTANT for c. In addition SERIES 1, SERIES 2, etc.
        indicate the input series relevant to the OMEGA and DELTA
        parameters.

        `kpriv` must be set to 0 if the print-out of the above
        information is not required.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    para : float, ndarray, shape (npara)
        The latest values of the estimates of these parameters.

    xxy : float, ndarray, shape (nxxy, nser)
        If kzef = 0, `xxy` remains unchanged on exit.

        If kzef != 0, the columns of `xxy` hold the corresponding values
        of the input component series z_t in place of x_t and the output
        noise component n_t in place of y_t, in that order.

    zsp : float, ndarray, shape (4)
        Contains the values, default or otherwise, used by the function.

    itc : int
        The number of iterations carried out.

        `itc` = -1

            Indicates that the only estimates obtained up to this point
            have been for the nuisance parameters relating to
            backforecasts, unless the marginal likelihood option is
            used, in which case estimates have also been obtained for
            simple input coefficients omega and for the constant c (if
            `kfc` = 1). This value of `itc` usually indicates a failure
            in a consequent step of estimating transfer function input
            pre-observation period nuisance parameters.

        `itc` = 0

            Indicates that estimates have been obtained up to this point
            for the constant c (if `kfc` = 1), for simple input
            coefficients omega and for the nuisance parameters relating
            to the backforecasts and to transfer function input
            pre-observation period effects.

    sd : float, ndarray, shape (npara)
        The npara values of the standard deviations corresponding to
        each of the parameters in `para`. When the constant is fixed its
        standard deviation is returned as zero. When the values of
        `para` are valid, the values of `sd` are usually also valid.
        However, if an exit value of `errno` = 3, 8 or 10, then the
        contents of `sd` will be indeterminate.

    cm : float, ndarray, shape (npara, npara)
        The first npara rows and columns of `cm` contain the correlation
        coefficients relating to each pair of parameters in `para`. All
        coefficients relating to the constant will be zero if the
        constant is fixed. The contents of `cm` will be indeterminate
        under the same conditions as `sd`.

    s : float
        The residual sum of squares, S, at the latest set of valid
        parameter estimates.

    d : float
        The objective function, D, at the latest set of valid parameter
        estimates.

    ndf : int
        The number of degrees of freedom associated with S.

    res : float, ndarray, shape (nxxy)
        The values of the residuals relating to the differenced values
        of the output series. The remainder of the first nxxy terms in
        the array will be zero.

    sttf : float, ndarray, shape (`nsttf`)
        The `nsttf` values of the state set array.

    nsttf : int
        The number of values in the state set array `sttf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nser = 1 and there are no parameters in the model.

        (`errno` 1)
            On entry, nser = *<value>*.

            Constraint: nser >= 1.

        (`errno` 1)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 1)
            On entry, `kef` = *<value>*.

            Constraint: `kef` = 1, 2 or 3.

        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 2)
            On entry, `ndf` = *<value>*.

            Constraint: `ndf` > 0.

        (`errno` 2)
            On entry, npara = *<value>* and `kfc` = *<value>*.

            Constraint: npara, `kfc`, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

        (`errno` 2)
            The orders vector `mr` is invalid.

        (`errno` 3)
            On entry, or during execution, one or more sets of delta
            parameters do not satisfy the stationarity or invertibility
            conditions.

        (`errno` 4)
            On entry, `zsp`[3] = *<value>*.

            Constraint: 0.0 <= `zsp`[3] < 1.0.

        (`errno` 4)
            On entry, `zsp`[2] = *<value>*.

            Constraint: `zsp`[2] >= 1.0.

        (`errno` 4)
            On entry, `zsp`[1] = *<value>*.

            Constraint: `zsp`[1] > 1.0.

        (`errno` 4)
            On entry, `zsp`[0] = *<value>*.

            Constraint: `zsp`[0] > 0.0.

        (`errno` 8)
            Unable to calculate the latest parameter estimates.

        (`errno` 9)
            Failure in inversion of second derivative matrix.

        (`errno` 10)
            On entry, or during execution, one or more sets of ARIMA
            parameters do not satisfy the stationarity or invertibility
            conditions.

        (`errno` 11)
            On entry, isttf too small.

            Constraint: isttf >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncg.

        (`errno` 12)
            The routine has failed to converge after *<value>*
            iterations.

        (`errno` 13)
            On entry, `nit` = *<value>*.

            isttf is too small and the process failed to converge after
            `nit` iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The Multi-input Model
    ~~~~~~~~~~~~~~~~~~~~~
    The output series y_t, for t = 1,2,...,n, is assumed to be the sum
    of (unobserved) components z_{i,t} which are due respectively to the
    inputs x_{i,t}, for i = 1,2,...,m.

    Thus y_t = z_{1,t}+...+z_{m,t}+n_t where n_t is the error, or output
    noise component.

    A typical component z_t may be either

    (a) a simple regression component, z_t = omega x_t (here x_t is
        called a simple input), or

    (#) a transfer function model component which allows for the effect
        of lagged values of the variable, related to x_t by

            z_t = delta_1 z_{t-1}+delta_2 z_{t-2}+...+delta_p
            z_{t-p}+omega_0 x_{t-b}-omega_1 x_{t-b-1}-...-omega_q
            x_{t-b-q}.

    The noise n_t is assumed to follow a (possibly seasonal) ARIMA
    model, i.e., may be represented in terms of an uncorrelated series,
    a_t, by the hierarchy of equations

    (i) nabla^d nabla_s^Dn_t = c+w_t

    (#) w_t = Phi_1 w_{t-s}+Phi_2 w_{t-2*s}+...+Phi_P
        w_{t-P*s}+e_t-Theta_1 e_{t-s}-Theta_2 e_{t-2*s}-...-Theta_Q
        e_{t-Q*s}

    (#) e_t = phi_1 e_{t-1}+phi_2 e_{t-2}+...+phi_p e_{t-p}+a_t-theta_1
        a_{t-1}-theta_2 a_{t-2}-...-theta_q a_{t-q}

    as outlined in [Notes for uni_arima_estim].

    **Note:** the orders p,q appearing in each of the transfer function
    models and the ARIMA model are not necessarily the same; nabla^d
    nabla_s^Dn_t is the result of applying non-seasonal differencing of
    order d and seasonal differencing of seasonality s and order D to
    the series n_t: the differenced series is then of length N =
    n-d-s*D; the constant term parameter c may optionally be held fixed
    at its initial value (usually, but not necessarily zero) rather than
    being estimated.

    For the purpose of defining an estimation criterion it is assumed
    that the series a_t is a sequence of independent Normal variates
    having mean 0 and variance sigma_a^2.
    An allowance has to be made for the effects of unobserved data prior
    to the observation period.
    For the noise component an allowance is always made using a form of
    backforecasting.

    For each transfer function input, you have to decide what values are
    to be assumed for the pre-period terms z_0,z_-1,...,z_{1-p} and
    x_0,x_-1,...,x_{1-b-q} which are in theory necessary to re-create
    the component series z_1,z_2,...,z_n, during the estimation
    procedure.

    The first choice is to assume that all these values are zero.
    In this case, in order to avoid undesirable transient distortion of
    the early values z_1,z_2,..., you are advised first to correct the
    input series x_t by subtracting from all the terms a suitable
    constant to make the early values x_1,x_2,..., close to zero.
    The series mean x_ is one possibility, but for a series with strong
    trend the constant might be simply x_1.

    The second choice is to treat the unknown pre-period terms as
    nuisance parameters and estimate them along with the other
    parameters.
    This choice should be used with caution.
    For example, if p = 1 and b = q = 0, it is equivalent to fitting to
    the data a decaying geometric curve of the form A delta^t, for t =
    1,2,...,, along with the other inputs, this being the form of the
    transient.
    If the output y_t contains a strong trend of this form, which is not
    otherwise represented in the model, it will have a tendency to
    influence the estimate of delta away from the value appropriate to
    the transfer function model.

    In most applications the first choice should be adequate, with the
    option possibly being used as a refinement at the end of the
    modelling process.
    The number of nuisance parameters is then max(p, b+q), with a
    corresponding loss of degrees of freedom in the residuals.
    If you align the input x_t with the output by using in its place the
    shifted series x_{t-b}, then setting b = 0 in the transfer function
    model, there is some improvement in efficiency.
    On some occasions when the model contains two or more inputs, each
    with estimation of pre-period nuisance parameters, these parameters
    may be co-linear and lead to failure of the function.
    The option must then be 'switched off' for one or more inputs.

    The Estimation Criterion
    ~~~~~~~~~~~~~~~~~~~~~~~~
    This is a measure of how well a proposed set of parameters in the
    transfer function and noise ARIMA models matches the data.
    The estimation function searches for parameter values which minimize
    this criterion.
    For a proposed set of parameter values it is derived by calculating

    (i) the components z_{1,t},z_{2,t},...,z_{m,t} as the responses to
        the input series x_{1,t},x_{2,t}...,x_{m,t} using the equations
        (a) or (b) above,

    (#) the discrepancy between the output and the sum of these
        components, as the noise

            n_t = y_t-(z_{1,t}+z_{2,t}+...+z_{m,t}),

    (#) the residual series a_t from n_t by reversing the recursive
        equations (i), (ii) and (iii) above.

    This last step again requires treatment of the effect of unknown
    pre-period values of n_t and other terms in the equations
    regenerating a_t. This is identical to the treatment given in [Notes
    for uni_arima_estim], and leads to a criterion which is a sum of
    squares function S, of the residuals a_t. It may be shown that the
    finite algorithm presented there is equivalent to taking the
    infinite set of past values n_0,n_-1,n_-2,..., as (linear) nuisance
    parameters. There is no loss of degrees of freedom however, because
    the sum of squares function S may be expressed as including the
    corresponding set of past residuals; see page 273 of Box and Jenkins
    (1976), who prove that

        S = sum_- infinity^na_t^2.

    The function D = S is the first of the three possible criteria, and
    is quite adequate for moderate to long series with no seasonal
    parameters.
    The second is the exact likelihood criterion which considers the
    past set n_0,n_-1,n_-2 not as simple nuisance parameters, but as
    unobserved random variables with known distribution.
    Calculation of the likelihood of the observed set n_1,n_2,...,n_n
    requires theoretical integration over the range of the past set.
    Fortunately this yields a criterion of the form D = M*S (whose
    minimization is equivalent to maximizing the exact likelihood of the
    data), where S is exactly as before, and the multiplier M is a
    function calculated from the ARIMA model parameters.
    The value of M is always >= 1, and M tends to 1 for any fixed
    parameter set as the sample size n tends to infinity.
    There is a moderate computational overhead in using this option, but
    its use avoids appreciable bias in the ARIMA model parameters and
    yields a better conditioned estimation problem.

    The third criterion of marginal likelihood treats the coefficients
    of the simple inputs in a manner analogous to that given to the past
    set n_0,n_-1,n_-2,....
    These coefficients, together with the constant term c used to
    represent the mean of w_t, are in effect treated as random variables
    with highly dispersed distributions.
    This leads to the criterion D = M*S again, but with a different
    value of M which now depends on the simple input series values x_t.
    In the presence of a moderate to large number of simple inputs, the
    marginal likelihood criterion can counteract bias in the ARIMA model
    parameters which is caused by estimation of the simple inputs.
    This is particularly important in relatively short series.

    ``multi_inputmod_estim`` can be used with no input series present,
    to estimate a univariate ARIMA model for the output alone.
    The marginal likelihood criterion is then distinct from exact
    likelihood only if a constant term is being estimated in the model,
    because this is treated as an implicit simple input.

    The Estimation Procedure
    ~~~~~~~~~~~~~~~~~~~~~~~~
    This is the minimization of the estimation criterion or objective
    function D (for deviance).
    The function uses an extension of the algorithm of Marquardt (1963).
    The step size in the minimization is inversely related to a
    parameter alpha, which is increased or decreased by a factor beta at
    successive iterations, depending on the progress of the
    minimization.
    Convergence is deemed to have occurred if the fractional reduction
    of D in successive iterations is less than a value gamma, while
    alpha < 1.

    Certain model parameters (in fact all excluding the omegas) are
    subject to stability constraints which are checked throughout to
    within a specified tolerance multiple delta of machine accuracy.
    Using the least squares criterion, the minimization may halt
    prematurely when some parameters 'stick' at a constraint boundary.
    This can happen particularly with short seasonal series (with a
    small number of whole seasons).
    It will not happen using the exact likelihood criterion, although
    convergence to a point on the boundary may sometimes be rather slow,
    because the criterion function may be very flat in such a region.
    There is also a smaller risk of a premature halt at a constraint
    boundary when marginal likelihood is used.

    A positive, or zero number of iterations can be specified.
    In either case, the value D of the objective function at iteration
    zero is presented at the initial parameter values, except for
    estimation of any pre-period terms for the input series,
    backforecasts for the noise series, and the coefficients of any
    simple inputs, and the constant term (unless this is held fixed).

    At any later iteration, the value of D is computed after
    re-estimation of the backforecasts to their optimal values,
    corresponding to the model parameters presented at that iteration.
    This is not true for any pre-period terms for the input series
    which, although they are updated from the previous iteration, may
    not be precisely optimal for the parameter values presented, unless
    convergence of those parameters has occurred.
    However, in the case of marginal likelihood being specified, the
    coefficients of the simple inputs and the constant term are also
    re-estimated together with the backforecasts at each iteration, to
    values which are optimal for the other parameter values presented.

    Further Results
    ~~~~~~~~~~~~~~~
    The residual variance is taken as erv = S/df, where df = N-(total
    number of parameters estimated), is the residual degrees of freedom.
    The pre-period nuisance parameters for the input series are included
    in the reduction of df, as is the constant if it is estimated.

    The covariance matrix of the vector of model parameter estimates is
    given by

        erv*H^-1

    where H is the linearized least squares matrix taken from the final
    iteration of the algorithm of Marquardt.
    From this expression are derived the vector of standard deviations,
    and the correlation matrix of parameter estimates.
    These are approximations which are only valid asymptotically, and
    must be treated with great caution when the parameter estimates are
    close to their constraint boundaries.

    The residual series a_t is available upon completion of the
    iterations over the range t = 1+d+s*D,...,n corresponding to the
    differenced noise series w_t.

    Because of the algorithm used for backforecasting, these are only
    true residuals for t >= 1+q+s*Q-p-s*P-d-s*D, provided this is
    positive.
    Estimation of pre-period terms for the inputs will also tend to
    reduce the magnitude of the early residuals, sometimes severely.

    The model component series z_{1,t},...,z_{m,t} and n_t may
    optionally be returned in place of the supplied series values, in
    order to assess the effects of the various inputs on the output.

    Forecasting Information
    ~~~~~~~~~~~~~~~~~~~~~~~
    For the purpose of constructing forecasts of the output series at
    future time points t = n+1,n+2,... using
    ``multi_inputmod_forecast_state``, it is not necessary to use the
    whole set of observations y_t and x_{1,t},x_{2,t},...,x_{m,t}, for t
    = 1,2,...,m.
    It is sufficient to retain a limited set of quantities constituting
    the 'state set' as follows: for each series which appears with
    lagged subscripts in equations (a), (b), (i), (ii) and (iii) above,
    include the values at times n+1-k for k = 1 up to the maximum lag
    associated with that series in the equations.
    Note that (i) implicitly includes past values of n_t and
    intermediate differences of n_t such as nabla^(d-1)nabla_s^D.

    If later observations of the series become available, it is possible
    to update the state set (without re-estimating the model) using
    ``multi_inputmod_update``.
    If time series data is supplied with a previously estimated model,
    it is possible to construct the state set (and forecasts) using
    ``multi_inputmod_forecast``.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Marquardt, D W, 1963, `An algorithm for least squares estimation of
    nonlinear parameters`, J. Soc. Indust. Appl. Math. (11), 431

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.multi_inputmod_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_estim'

    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _xxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxy'},
        xxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_mt, 'mt'), (_xxy, 'xxy')),
    )
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _nxxy_py = _b_utils._EngineFloat64ArrayType.get_shape(_xxy, {'fun_name': fun_name, 'entity': 'xxy'}, exp_rank=2)[0]
    _nxxy = _b_utils._EngineIntScalarType(
        _nxxy_py, {'fun_name': fun_name, 'entity': '_nxxy'},
    ).to_ctypes()
    if zsp is None:
        _kzsp = 0
        _kzsp_py = _kzsp
    else:
        _kzsp = 1
        _kzsp_py = _kzsp
    if zsp is None:
        _zsp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (4,),
        )
    else:
        _zsp = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'zsp'},
            zsp,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _sd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npara_py,),
    )
    _cm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npara_py, _npara_py),
        sorder=_m_sorder,
    )
    _kzef = 1
    _kzef_py = _kzef
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nxxy_py,),
    )
    _vl_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))
    _vl_d_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))
    _vu_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))
    _vl_s_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _vu_d_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))
    _vl_ncg_py = 0
    for _vl_i_py in range(1, _nser_py+1):
        if _b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (3, _vl_i_py-1)) > 1:
            _vl_ncg_py = _vl_ncg_py+_b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (0, _vl_i_py-1))+_b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (1, _vl_i_py-1))+_b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (2, _vl_i_py-1))
    _vl_ncg = _b_utils._EngineIntScalarType(
        _vl_ncg_py, {'fun_name': fun_name, 'entity': '_vl_ncg_py'},
    ).to_ctypes()
    _isttf = _b_utils._EngineIntScalarType(
        (_vu_p_py*_vl_s_py)+_vl_d_py+(_vu_d_py*_vl_s_py)+_vl_q_py+max(_vl_p_py, _vu_q_py*_vl_s_py)+_vl_ncg_py, {'fun_name': fun_name, 'entity': '_isttf'},
    ).to_ctypes()
    _sttf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_vu_p_py*_vl_s_py)+_vl_d_py+(_vu_d_py*_vl_s_py)+_vl_q_py+max(_vl_p_py, _vu_q_py*_vl_s_py)+_vl_ncg_py,),
    )
    _return_scalars = _b_tsa.multi_inputmod_estim(
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        kfc,
        _nxxy,
        _xxy,
        kef,
        nit,
        _kzsp,
        _zsp,
        _sd,
        _cm,
        _kzef,
        _res,
        _sttf,
        _isttf,
        kpriv,
        io_manager,
    )
    _para_py = _para
    _xxy_py = _xxy
    _zsp_py = _zsp
    _itc_py = _return_scalars.itc
    _sd_py = _sd
    _cm_py = _cm
    _s_py = _return_scalars.s
    _d_py = _return_scalars.d
    _ndf_py = _return_scalars.ndf
    _res_py = _res
    _nsttf_py = _return_scalars.nsttf
    _sttf_py = _sttf[:_nsttf_py]
    _return_data = MultiInputmodEstimReturnData(
        _para_py,
        _xxy_py,
        _zsp_py,
        _itc_py,
        _sd_py,
        _cm_py,
        _s_py,
        _d_py,
        _ndf_py,
        _res_py,
        _sttf_py,
        _nsttf_py,
    )
    return _return_data

MultiInputmodUpdateReturnData = _collections.namedtuple(
    'MultiInputmodUpdateReturnData',
    ['sttf', 'xxyn', 'res'],
)

def multi_inputmod_update(sttf, mr, mt, para, xxyn, kzef):
    u"""
    Multivariate time series, update state set for forecasting from
    multi-input model.

    ``multi_inputmod_update`` accepts a series of new observations of an
    output time series and any associated input time series, for which a
    multi-input model is already fully specified, and updates the 'state
    set' information for use in constructing further forecasts.

    The previous specification of the multi-input model will normally
    have been obtained by using ``multi_inputmod_estim`` to estimate the
    relevant transfer function and ARIMA parameters.
    The supplied state set will originally have been produced by
    ``multi_inputmod_estim`` (or possibly ``multi_inputmod_forecast``),
    but may since have been updated by ``multi_inputmod_update``.

    For full information please refer to the NAG Library document for
    g13bg

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bgf.html

    Parameters
    ----------
    sttf : float, array-like, shape (nsttf)
        The nsttf values in the state set before updating as returned by
        ``multi_inputmod_estim`` or ``multi_inputmod_forecast``, or a
        previous call to ``multi_inputmod_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing, and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function model orders b, p and q of each of the
        input series. The data for input series i are held in column i.
        Row 1 holds the value b_i, row 2 holds the value q_i and row 3
        holds the value p_i. For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input and
        r_i = 2 or 3 for a transfer function input.

        When r_i = 1 any nonzero contents of rows 1, 2 and 3 of column i
        are ignored.

        The choice of r_i = 2 or r_i = 3 is an option for use in model
        estimation and does not affect the operation of
        ``multi_inputmod_update``.

    para : float, array-like, shape (npara)
        Estimates of the multi-input model parameters as returned by
        ``multi_inputmod_estim``. These are in order, firstly the ARIMA
        model parameters: p values of phi parameters, q values of theta
        parameters, P values of Phi parameters and Q values of Theta
        parameters. These are followed by the transfer function model
        parameter values omega_0,omega_1,...,omega_q_1,
        delta_1,delta_2,...,delta_p_1 for the first of any input series
        and similarly for each subsequent input series. The final
        component of `para` is the value of the constant c.

    xxyn : float, array-like, shape (nnv, nser)
        The nnv new observation sets being used to update the state set.
        Column i-1 contains the values of input series i, for i =
        1,2,...,nser-1. Column nser-1 contains the values of the output
        series. Consecutive rows correspond to increasing time sequence.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxyn` on exit, and must be set to
        0 if `xxyn` is to remain unchanged on exit.

    Returns
    -------
    sttf : float, ndarray, shape (nsttf)
        The state set values after updating.

    xxyn : float, ndarray, shape (nnv, nser)
        If `kzef` = 0, `xxyn` remains unchanged.

        If `kzef` != 0, the columns of `xxyn` hold the corresponding
        values of the input component series z_t and the output noise
        component n_t in that order.

    res : float, ndarray, shape (nnv)
        The values of the residual series a_t corresponding to the new
        observations of the output series.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nsttf = *<value>*.

            Constraint: nsttf, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, npara = *<value>*.

            Constraint: npara, `mr` and `mt` must be consistent.

        (`errno` 5)
            On entry, the orders vector `mr` is invalid.

        (`errno` 5)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

    Notes
    -----
    The multi-input model is specified in [Notes for
    multi_inputmod_estim].
    The form of these equations required to update the state set is as
    follows:

        z_t = delta_1 z_{t-1}+delta_2 z_{t-2}+...+delta_p
        z_{t-p}+omega_0 x_{t-b}-omega_1 x_{t-b-1}-...-omega_q x_{t-b-q}

    the transfer models which generate input component values z_{i,t}
    from one or more inputs x_{i,t},

        n_t = y_t-z_{1,t}-z_{2,t}-...-z_{m,t}

    which generates the output noise component from the output y_t and
    the input components, and

        [table omitted]

    the ARIMA model for the output noise which generates the residuals
    a_t.

    The state set (as also given in [Notes for multi_inputmod_estim]) is
    the collection of terms

        z_{n+1-k},x_{n+1-k},n_{n+1-k},w_{n+1-k},e_{n+1-k} and a_{n+1-k}

    for k = 1 up to the maximum lag associated with each of these series
    respectively, in the above model equations. n is the latest time
    point of the series from which the state set has been generated.

    The function accepts further values of the series y_t,
    x_{1,t},x_{2,t},...,x_{m,t}, for t = n+1,...,n+l, and applies the
    above model equations over this time range, to generate new values
    of the various model components, noise series and residuals.
    The state set is reconstructed, corresponding to the latest time
    point n+l, the earlier values being discarded.

    The set of residuals corresponding to the new observations may be of
    use in checking that the new observations conform to the previously
    fitted model.
    The components of the new observations of the output series which
    are due to the various inputs, and the noise component, are also
    optionally returned.

    The parameters of the model are not changed in this function.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_update'

    _sttf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sttf'},
        sttf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nsttf_py = _b_utils._EngineFloat64ArrayType.get_shape(_sttf, {'fun_name': fun_name, 'entity': 'sttf'}, exp_rank=1)[0]
    _nsttf = _b_utils._EngineIntScalarType(
        _nsttf_py, {'fun_name': fun_name, 'entity': '_nsttf'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _xxyn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxyn'},
        xxyn,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnv_py = _b_utils._EngineFloat64ArrayType.get_shape(_xxyn, {'fun_name': fun_name, 'entity': 'xxyn'}, exp_rank=2)[0]
    _nnv = _b_utils._EngineIntScalarType(
        _nnv_py, {'fun_name': fun_name, 'entity': '_nnv'},
    ).to_ctypes()
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nnv_py,),
    )
    _b_tsa.multi_inputmod_update(
        _sttf,
        _nsttf,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _nnv,
        _xxyn,
        kzef,
        _res,
    )
    _sttf_py = _sttf
    _xxyn_py = _xxyn
    _res_py = _res
    _return_data = MultiInputmodUpdateReturnData(
        _sttf_py,
        _xxyn_py,
        _res_py,
    )
    return _return_data

MultiInputmodForecastStateReturnData = _collections.namedtuple(
    'MultiInputmodForecastStateReturnData',
    ['xxyn', 'mrx', 'fva', 'fsd'],
)

def multi_inputmod_forecast_state(sttf, mr, mt, para, xxyn, mrx, parx, rmsxy, kzef):
    u"""
    Multivariate time series, forecasting from state set of multi-input
    model.

    ``multi_inputmod_forecast_state`` produces forecasts of a time
    series (the output series) which depends on one or more other
    (input) series via a multi-input model which will usually have been
    fitted using ``multi_inputmod_estim``.
    The future values of the input series must be supplied.
    The original observations are not required.
    ``multi_inputmod_forecast_state`` uses as input either the original
    state set obtained from ``multi_inputmod_estim``, or the state set
    updated by a series of new observations from
    ``multi_inputmod_update``.
    Standard errors of the forecasts are produced.
    If future values of some of the input series have been obtained as
    forecasts using ARIMA models for those series, this may be allowed
    for in the calculation of the standard errors.

    For full information please refer to the NAG Library document for
    g13bh

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bhf.html

    Parameters
    ----------
    sttf : float, array-like, shape (nsttf)
        The nsttf values in the state set as returned by
        ``multi_inputmod_estim`` or ``multi_inputmod_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q give respectively the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing, and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function orders b, p and q of each of the input
        series. The data for input series i are held in column i. Row 1
        holds the value b_i, row 2 holds the value q_i and row 3 holds
        the value p_i. For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, r_i
        = 2 or 3 for a transfer function input.

        When r_i = 1, any nonzero contents of rows 1, 2 and 3 of column
        i are ignored.

        The choice of r_i = 2 or r_i = 3 is an option for use in model
        estimation and does not affect the operation of
        ``multi_inputmod_forecast_state``.

    para : float, array-like, shape (npara)
        Estimates of the multi-input model parameters as returned by
        ``multi_inputmod_estim``. These are in order, firstly the ARIMA
        model parameters: p values of phi parameters, q values of theta
        parameters, P values of Phi parameters and Q values of Theta
        parameters. These are followed by the transfer function model
        parameter values omega_0,omega_1,...,omega_q_1,
        delta_1,delta_2,...,delta_p_1 for the first of any input series
        and similar sets of values for any subsequent input series. The
        final component of `para` is the constant c.

    xxyn : float, array-like, shape (nfv, nser)
        The supplied nfv values for each of the input series required to
        produce the nfv output series forecasts. Column i contains the
        values for input series i. Column nser need not be supplied.

    mrx : int, array-like, shape (7, nser)
        The orders array for each of the input series ARIMA models.
        Thus, column i contains values of p, d, q, P, D, Q, s for input
        series i. In the case of those inputs for which no ARIMA model
        is available, the corresponding orders should be set to 0. (The
        model for any input series only affects the standard errors of
        the forecast values.)

    parx : float, array-like, shape (:, nser)
        Values of the parameters (phi, theta, Phi and Theta) for each of
        the input series ARIMA models. Thus column i contains
        `mrx`[0,i-1] values of phi parameters, `mrx`[2,i-1] values of
        theta parameters, `mrx`[3,i-1] values of Phi parameters and
        `mrx`[5,i-1] values of Theta parameters -- in that order.

        Values in the columns relating to those input series for which
        no ARIMA model is available are ignored. (The model for any
        input series only affects the standard errors of the forecast
        values.)

    rmsxy : float, array-like, shape (nser)
        The estimated residual variances for each input series ARIMA
        model followed by that for the output noise ARIMA model. In the
        case of those inputs for which no ARIMA model is available, or
        when its effects are to be excluded in the calculation of
        forecast standard errors, the corresponding entry of `rmsxy`
        should be set to 0.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxyn` on exit, and must be set to
        0 if `xxyn` is to remain unchanged on exit, apart from the
        appearance of the forecast values in column nser.

    Returns
    -------
    xxyn : float, ndarray, shape (nfv, nser)
        If `kzef` = 0, then column nser of `xxyn` contains the output
        series forecast values (as does `fva`), but `xxyn` is otherwise
        unchanged.

        If `kzef` != 0, then the columns of `xxyn` hold the
        corresponding values of the forecast components z_t for each of
        the input series and the values of the output noise component
        n_t in that order.

    mrx : int, ndarray, shape (7, nser)
        Unchanged, apart from column nser which is used for workspace.

    fva : float, ndarray, shape (nfv)
        The required forecast values for the output series.

    fsd : float, ndarray, shape (nfv)
        The standard errors for each of the forecast values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nsttf = *<value>* and the expected value =
            *<value>*.

            Constraint: nsttf, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, npara = *<value>* and the expected value =
            *<value>*.

            Constraint: npara, `mr` and `mt` must be consistent.

        (`errno` 6)
            The orders vector `mr` is invalid.

        (`errno` 6)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The forecasts of the output series y_t are calculated, for t =
    n+1,...,n+L, where n is the latest time point of the observations
    used to produce the state set and L is the maximum lead time of the
    forecasts.

    First the new input series values x_t are used to form the input
    components z_t, for t = n+1,...,n+L, using the transfer function
    models:

    [table omitted]

    The output noise component n_t is then forecast by setting a_t = 0,
    for t = n+1,...,n+L, and using the ARIMA model equations:

    [table omitted]

    This last step of 'integration' reverses the process of
    differencing.
    Finally the output forecasts are calculated as

        y_t = z_{1,t}+z_{2,t}+...+z_{m,t}+n_t.

    The forecast error variance of y_{t+l} (i.e., at lead time l) is
    S_l^2, which is the sum of parts which arise from the various input
    series, and the output noise component.
    That part due to the output noise is

        sn_l^2 = V_n*(psi_0^2+psi_1^2+...+psi_l-1^2),

    where V_n is the estimated residual variance of the output noise
    ARIMA model, and psi_0,psi_1,... are the 'psi-weights' of this model
    as defined in Box and Jenkins (1976).
    They are calculated by applying the equations (b), (c) and (d)
    above, for t = 0,1,...,L, but with artificial values for the various
    series and with the constant c set to 0.
    Thus all values of a_t, e_t, w_t and n_t are taken as zero, for t <
    0; a_t is taken to be 1, for t = 0 and 0, for t > 0.
    The resulting values of n_t, for t = 0,1,...,L, are precisely
    psi_0,psi_1,...,psi_L as required.

    Further contributions to S_l^2 come only from those input series,
    for which future values are forecasts which have been obtained by
    applying input series ARIMA models.
    For such a series the contribution is

        sz_l^2 = V_x*(nu_0^2+nu_1^2+...+nu_l-1^2),

    where V_x is the estimated residual variance of the input series
    ARIMA model.
    The coefficients nu_0,nu_1,... are calculated by applying the
    transfer function model equation (a) above, for t = 0,1,...,L, but
    again with artificial values of the series.
    Thus all values of z_t and x_t, for t < 0, are taken to be zero, and
    x_0,x_1,... are taken to be the psi-weight sequence psi_0,psi_1,...
    for the **input** **series** ARIMA model.
    The resulting values of z_t, for t = 0,1,...,L, are precisely
    nu_0,nu_1,...,nu_L as required.

    In adding such contributions sz_l^2 to sn_l^2 to make up the total
    forecast error variance S_l^2, it is assumed that the various input
    series with which these contributions are associated are
    statistically independent of each other.

    When using the function in practice an ARIMA model is required for
    all the input series.
    In the case of those inputs for which no such ARIMA model is
    available (or its effects are to be excluded), the corresponding
    orders and parameters and the estimated residual variance should be
    set to zero.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.multi_inputmod_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_forecast_state'

    _sttf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sttf'},
        sttf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nsttf_py = _b_utils._EngineFloat64ArrayType.get_shape(_sttf, {'fun_name': fun_name, 'entity': 'sttf'}, exp_rank=1)[0]
    _nsttf = _b_utils._EngineIntScalarType(
        _nsttf_py, {'fun_name': fun_name, 'entity': '_nsttf'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=True,
    )
    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _xxyn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxyn'},
        xxyn,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nfv_py = _b_utils._EngineFloat64ArrayType.get_shape(_xxyn, {'fun_name': fun_name, 'entity': 'xxyn'}, exp_rank=2)[0]
    _nfv = _b_utils._EngineIntScalarType(
        _nfv_py, {'fun_name': fun_name, 'entity': '_nfv'},
    ).to_ctypes()
    _mrx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mrx'},
        mrx,
        copy=True,
    )
    _parx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parx'},
        parx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rmsxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rmsxy'},
        rmsxy,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _b_tsa.multi_inputmod_forecast_state(
        _sttf,
        _nsttf,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _nfv,
        _xxyn,
        _mrx,
        _parx,
        _rmsxy,
        kzef,
        _fva,
        _fsd,
    )
    _xxyn_py = _xxyn
    _mrx_py = _mrx
    _fva_py = _fva
    _fsd_py = _fsd
    _return_data = MultiInputmodForecastStateReturnData(
        _xxyn_py,
        _mrx_py,
        _fva_py,
        _fsd_py,
    )
    return _return_data

MultiInputmodForecastReturnData = _collections.namedtuple(
    'MultiInputmodForecastReturnData',
    ['para', 'xxy', 'rmsxy', 'mrx', 'fva', 'fsd', 'sttf', 'nsttf'],
)

def multi_inputmod_forecast(mr, mt, para, kfc, nev, nfv, xxy, kzef, rmsxy, mrx, parx, isttf, io_manager=None):
    u"""
    Multivariate time series, state set and forecasts from fully
    specified multi-input model.

    ``multi_inputmod_forecast`` produces forecasts of a time series (the
    output series) which depends on one or more other (input) series via
    a previously estimated multi-input model for which the state set
    information is not available.
    The future values of the input series must be supplied.
    In contrast with ``multi_inputmod_forecast_state`` the original past
    values of the input and output series are required.
    Standard errors of the forecasts are produced.
    If future values of some of the input series have been obtained as
    forecasts using ARIMA models for those series, this may be allowed
    for in the calculation of the standard errors.

    For full information please refer to the NAG Library document for
    g13bj

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13bjf.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function model orders b, p and q of each of the
        input series. The data for input series i is held in column i.
        Row 1 holds the value b_i, row 2 holds the value q_i and row 3
        holds the value p_i.

        For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, and
        r_i = 2 or 3 for a transfer function input.

        The choice r_i = 3 leads to estimation of the pre-period input
        effects as nuisance parameters, and r_i = 2 suppresses this
        estimation.

        This choice may affect the returned forecasts and the state set.

        When r_i = 1, any nonzero contents of rows 1, 2 and 3 of column
        i are ignored.

    para : float, array-like, shape (npara)
        Estimates of the multi-input model parameters. These are in
        order, firstly the ARIMA model parameters: p values of phi
        parameters, q values of theta parameters, P values of Phi
        parameters, Q values of Theta parameters.

        These are followed by the transfer function model parameter
        values omega_0,omega_1,...,omega_q_1, delta_1,...,delta_p_1 for
        the first of any input series and similarly for each subsequent
        input series.

        The final component of `para` is the value of the constant c.

    kfc : int
        Must be set to 1 if the constant was estimated when the model
        was fitted, and 0 if it was held at a fixed value. This only
        affects the degrees of freedom used in calculating the estimated
        residual variance.

    nev : int
        The number of original (undifferenced) values in each of the
        input and output time series.

    nfv : int
        The number of forecast values of the output series required.

    xxy : float, array-like, shape ((`nev`+`nfv`), nser)
        The columns of `xxy` must contain in the first `nev` places, the
        past values of each of the input and output series, in that
        order. In the next `nfv` places, the columns relating to the
        input series (i.e., columns 1 to nser-1) contain the future
        values of the input series which are necessary for construction
        of the forecasts of the output series y.

    kzef : int
        Must be set to 0 if the relevant `nfv` values of the forecasts
        (`fva`) are to be held in the output series column (nser) of
        `xxy` (which is otherwise unchanged) on exit, and must not be
        set to 0 if the values of the input component series z_t and the
        values of the output noise component n_t are to overwrite the
        contents of `xxy` on exit.

    rmsxy : float, array-like, shape (nser)
        The first (nser-1) elements of `rmsxy` must contain the
        estimated residual variance of the input series ARIMA models. In
        the case of those inputs for which no ARIMA model is available
        or its effects are to be excluded in the calculation of forecast
        standard errors, the corresponding entry of `rmsxy` should be
        set to 0.

    mrx : int, array-like, shape (7, nser)
        The orders array for each of the input series ARIMA models.
        Thus, column i contains values of p, d, q, P, D, Q, s for input
        series i. In the case of those inputs for which no ARIMA model
        is available, the corresponding orders should be set to 0.

    parx : float, array-like, shape (:, nser)
        Values of the parameters (phi, theta, Phi, and Theta) for each
        of the input series ARIMA models.

        Thus column i contains `mrx`[0,i-1] values of phi, `mrx`[2,i-1]
        values of theta, `mrx`[3,i-1] values of Phi and `mrx`[5,i-1]
        values of Theta, in that order.

        Values in the columns relating to those input series for which
        no ARIMA model is available are ignored.

    isttf : int
        The dimension of the array `sttf`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    para : float, ndarray, shape (npara)
        The parameter values may be updated using an additional
        iteration in the estimation process.

    xxy : float, ndarray, shape ((`nev`+`nfv`), nser)
        If `kzef` = 0 then `xxy` is unchanged except that the relevant
        `nfv` values in the column relating to the output series (column
        nser) contain the forecast values (`fva`), but if `kzef` != 0
        then the columns of `xxy` contain the corresponding values of
        the input component series z_t and the values of the output
        noise component n_t, in that order.

    rmsxy : float, ndarray, shape (nser)
        `rmsxy`[nser-1] contains the estimated residual variance of the
        output noise ARIMA model which is calculated from the supplied
        series. Otherwise `rmsxy` is unchanged.

    mrx : int, ndarray, shape (7, nser)
        Unchanged, except for column nser which is used as workspace.

    fva : float, ndarray, shape (`nfv`)
        The required forecast values for the output series. (Note that
        these are also output in column nser of `xxy` if `kzef` = 0.)

    fsd : float, ndarray, shape (`nfv`)
        The standard errors for each of the forecast values.

    sttf : float, ndarray, shape (`isttf`)
        The `nsttf` values of the state set based on the first `nev`
        sets of (past) values of the input and output series.

    nsttf : int
        The number of values in the state set array `sttf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nfv` = *<value>*.

            Constraint: `nfv` > 0.

        (`errno` 1)
            On entry, ldxxy = *<value>*, `nev` = *<value>* and `nfv` =
            *<value>*.

            Constraint: ldxxy >= `nev`+`nfv`.

        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 2)
            The orders vector `mr` is invalid.

        (`errno` 2)
            npara is inconsistent with `mr` and `mt`.

        (`errno` 2)
            Insufficient degrees of freedom to solve the problem.

        (`errno` 2)
            On entry, npara = *<value>* and the expected value =
            *<value>*.

            Constraint: npara, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

        (`errno` 3)
            One or more sets of delta parameters do not satisfy the
            stationarity or invertibility conditions.

        (`errno` 8)
            Unable to calculate the latest parameter estimates.

        (`errno` 9)
            Failure in inversion of second derivative matrix.

        (`errno` 10)
            One or more sets of ARIMA parameters do not satisfy the
            stationarity or invertibility conditions.

        (`errno` 11)
            On entry, `isttf` = *<value>*.

            Constraint: `isttf` >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncf.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``multi_inputmod_forecast`` has two stages.
    The first stage is essentially the same as a call to the model
    estimation function ``multi_inputmod_estim``, with zero iterations.
    In particular, all the parameters remain unchanged in the supplied
    input series transfer function models and output noise series ARIMA
    model.
    The internal nuisance parameters associated with the pre-observation
    period effects of the input series are estimated where requested,
    and so are any backforecasts of the output noise series.
    The output components z_t and n_t, and residuals a_t are calculated
    exactly as in [Notes for multi_inputmod_estim], and the state set
    for forecasting is constituted.

    The second stage is essentially the same as a call to the
    forecasting function ``multi_inputmod_forecast_state``.
    The same information is required, and the same information is
    returned.

    Use of ``multi_inputmod_forecast`` should be confined to situations
    in which the state set for forecasting is unknown.
    Forecasting from the original data is relatively expensive because
    it requires recalculation of the state set.
    ``multi_inputmod_forecast`` returns the state set for use in
    producing further forecasts using ``multi_inputmod_forecast_state``,
    or for updating the state set using ``multi_inputmod_update``.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_forecast'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=True,
    )
    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _xxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxy'},
        xxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rmsxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rmsxy'},
        rmsxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mrx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mrx'},
        mrx,
        copy=True,
    )
    _parx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parx'},
        parx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nfv = _b_utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = _b_utils._EngineIntScalarType.to_py(_nfv)
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _isttf = _b_utils._EngineIntScalarType(
        isttf, {'fun_name': fun_name, 'entity': 'isttf'},
    ).to_ctypes()
    _isttf_py = _b_utils._EngineIntScalarType.to_py(_isttf)
    _sttf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_isttf_py,),
    )
    _return_scalar = _b_tsa.multi_inputmod_forecast(
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        kfc,
        nev,
        _nfv,
        _xxy,
        kzef,
        _rmsxy,
        _mrx,
        _parx,
        _fva,
        _fsd,
        _sttf,
        _isttf,
        io_manager,
    )
    _para_py = _para
    _xxy_py = _xxy
    _rmsxy_py = _rmsxy
    _mrx_py = _mrx
    _fva_py = _fva
    _fsd_py = _fsd
    _sttf_py = _sttf
    _nsttf_py = _return_scalar
    _return_data = MultiInputmodForecastReturnData(
        _para_py,
        _xxy_py,
        _rmsxy_py,
        _mrx_py,
        _fva_py,
        _fsd_py,
        _sttf_py,
        _nsttf_py,
    )
    return _return_data

UniSpectrumLagReturnData = _collections.namedtuple(
    'UniSpectrumLagReturnData',
    ['c', 'xg', 'ng', 'stats'],
)

def uni_spectrum_lag(nx, mtx, px, iw, mw, ic, c, kc, l, lg, xg):
    u"""
    Univariate time series, smoothed sample spectrum using rectangular,
    Bartlett, Tukey or Parzen lag window.

    ``uni_spectrum_lag`` calculates the smoothed sample spectrum of a
    univariate time series using one of four lag windows -- rectangular,
    Bartlett, Tukey or Parzen window.

    For full information please refer to the NAG Library document for
    g13ca

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13caf.html

    Parameters
    ----------
    nx : int
        n, the length of the time series.

    mtx : int
        If covariances are to be calculated by the function (`ic` = 0),
        `mtx` must specify whether the data are to be initially mean or
        trend corrected.

        `mtx` = 0

            For no correction.

        `mtx` = 1

            For mean correction.

        `mtx` = 2

            For trend correction.

    px : float
        If covariances are to be calculated by the function (`ic` = 0),
        `px` must specify the proportion of the data (totalled over both
        ends) to be initially tapered by the split cosine bell taper.

        If covariances are supplied (`ic` != 0), `px` must specify the
        proportion of data tapered before the supplied covariances were
        calculated and after any mean or trend correction. `px` is
        required for the calculation of output statistics.

        A value of 0.0 implies no tapering.

    iw : int
        The choice of lag window.

        `iw` = 1

            Rectangular.

        `iw` = 2

            Bartlett.

        `iw` = 3

            Tukey.

        `iw` = 4

            Parzen.

    mw : int
        M, the 'cut-off' point of the lag window. Windowed covariances
        at lag M or greater are zero.

    ic : int
        Indicates whether covariances are to be calculated in the
        function or supplied in the call to the function.

        `ic` = 0

            Covariances are to be calculated.

        `ic` != 0

            Covariances are to be supplied.

    c : float, array-like, shape (nc)
        If `ic` != 0, `c` must contain the nc covariances for lags from
        0 to (nc-1), otherwise `c` need not be set.

    kc : int
        If `ic` = 0, `kc` must specify the order of the fast Fourier
        transform (FFT) used to calculate the covariances.

        If `ic` != 0, that is covariances are supplied, `kc` is not
        used.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        Therefore, it is also the order of the FFT used to construct the
        sample spectrum from the covariances.

    lg : int
        Indicates whether unlogged or logged spectral estimates and
        confidence limits are required.

        `lg` = 0

            Unlogged.

        `lg` != 0

            Logged.

    xg : float, array-like, shape (nxg)
        If the covariances are to be calculated, then `xg` must contain
        the `nx` data points. If covariances are supplied, `xg` may
        contain any values.

    Returns
    -------
    c : float, ndarray, shape (nc)
        If `ic` = 0, `c` will contain the nc calculated covariances.

        If `ic` != 0, the contents of `c` will be unchanged.

    xg : float, ndarray, shape (nxg)
        Contains the `ng` spectral estimates, f^(omega_i), for i =
        0,1,...,[L/2] in `xg`[0] to `xg`[`ng`-1] respectively (logged if
        `lg` = 1). The elements `xg`[i-1], for i = `ng`+1,...,nxg
        contain 0.0.

    ng : int
        The number of spectral estimates, [L/2]+1, in `xg`.

    stats : float, ndarray, shape (4)
        Four associated statistics. These are the degrees of freedom in
        `stats`[0], the lower and upper 95% confidence limit factors in
        `stats`[1] and `stats`[2] respectively (logged if `lg` = 1), and
        the bandwidth in `stats`[3].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nc = *<value>* and `nx` = *<value>*.

            Constraint: nc <= `nx`.

        (`errno` 1)
            On entry, nc = *<value>* and `mw` = *<value>*.

            Constraint: nc >= `mw`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nx` = *<value>*.

            Constraint: `mw` <= `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `iw` = *<value>*.

            Constraint: `iw` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` <= 1.0.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 0.0.

        (`errno` 1)
            On entry, nxg = *<value>* and `l` = *<value>*.

            Constraint: if `ic` != 0, nxg >= `l`.

        (`errno` 1)
            On entry, nxg = *<value>*, `kc` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` = 0, nxg >= max(`kc`, `l`).

        (`errno` 1)
            On entry, `ic` = 0 and `mtx` > 2: `mtx` = *<value>*.

        (`errno` 1)
            On entry, `ic` = 0 and `mtx` < 0: `mtx` = *<value>*.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>*, `nx` = *<value>* and nc =
            *<value>*.

            Constraint: if `ic` = 0, `kc` >= (`nx`+nc).

        (`errno` 3)
            On entry, `l` = *<value>* and `mw` = *<value>*.

            Constraint: `l` >= 2*`mw`-1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            One or more spectral estimates are negative.

            Unlogged spectral estimates are returned in `xg`, and the
            degrees of freedom, unloged confidence limit factors and
            bandwidth in `stats`.

        (`errno` 5)
            The calculation of confidence limit factors has failed.

            Spectral estimates (logged if requested) are returned in
            `xg`, and degrees of freedom and bandwidth in `stats`.

    Notes
    -----
    The smoothed sample spectrum is defined as

        f^(omega) = 1/2 pi(C_0+2sum_k = 1^M-1w_k C_k cos(omega k)),

    where M is the window width, and is calculated for frequency values

        omega_i = 2 pi i/L, i = 0,1,...,[L/2],

    where [] denotes the integer part.

    The autocovariances C_k may be supplied by you, or constructed from
    a time series x_1,x_2,...,x_n, as

        C_k = 1/nsum_t = 1^n-kx_t x_{t+k},

    the fast Fourier transform (FFT) being used to carry out the
    convolution in this formula.

    The time series may be mean or trend corrected (by classical least
    squares), and tapered before calculation of the covariances, the
    tapering factors being those of the split cosine bell:

        [table omitted]

    where T = [np/2] and p is the tapering proportion.

    The smoothing window is defined by

        w_k = W(k/M), k <= M-1,

    which for the various windows is defined over 0 <= alpha < 1 by

    rectangular:

        W(alpha) = 1

    Bartlett:

        W(alpha) = 1-alpha

    Tukey:

        W(alpha) = 1/2(1+cos(pialpha))

    Parzen:

        [table omitted]

    The sampling distribution of f^(omega) is approximately that of a
    scaled chi_d^2 variate, whose degrees of freedom d is provided by
    the function, together with multiplying limits mu, ml from which
    approximate 95% confidence intervals for the true spectrum f(omega)
    may be constructed as [ml* f^(omega),mu* f^(omega)].
    Alternatively, log f^(omega) may be returned, with additive limits.

    The bandwidth b of the corresponding smoothing window in the
    frequency domain is also provided.
    Spectrum estimates separated by (angular) frequencies much greater
    than b may be assumed to be independent.

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.uni_spectrum_lag'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nc_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]
    _nc = _b_utils._EngineIntScalarType(
        _nc_py, {'fun_name': fun_name, 'entity': '_nc'},
    ).to_ctypes()
    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxg_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _nxg = _b_utils._EngineIntScalarType(
        _nxg_py, {'fun_name': fun_name, 'entity': '_nxg'},
    ).to_ctypes()
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalar = _b_tsa.uni_spectrum_lag(
            nx,
            mtx,
            px,
            iw,
            mw,
            ic,
            _nc,
            _c,
            kc,
            l,
            lg,
            _nxg,
            _xg,
            _stats,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _c_py = _c
    _xg_py = _xg
    _ng_py = _return_scalar
    _stats_py = _stats
    _return_data = UniSpectrumLagReturnData(
        _c_py,
        _xg_py,
        _ng_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniSpectrumDaniellReturnData = _collections.namedtuple(
    'UniSpectrumDaniellReturnData',
    ['xg', 'ng', 'stats'],
)

def uni_spectrum_daniell(nx, mtx, px, mw, pw, l, lg, xg):
    u"""
    Univariate time series, smoothed sample spectrum using spectral
    smoothing by the trapezium frequency (Daniell) window.

    ``uni_spectrum_daniell`` calculates the smoothed sample spectrum of
    a univariate time series using spectral smoothing by the trapezium
    frequency (Daniell) window.

    For full information please refer to the NAG Library document for
    g13cb

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13cbf.html

    Parameters
    ----------
    nx : int
        n, the length of the time series.

    mtx : int
        Whether the data are to be initially mean or trend corrected.

        `mtx` = 0

            For no correction.

        `mtx` = 1

            For mean correction.

        `mtx` = 2

            For trend correction.

    px : float
        The proportion of the data (totalled over both ends) to be
        initially tapered by the split cosine bell taper. (A value of
        0.0 implies no tapering.)

    mw : int
        The value of M which determines the frequency width of the
        smoothing window as 2*pi/M. A value of n implies no smoothing is
        to be carried out.

    pw : float
        p, the shape parameter of the trapezium frequency window.

        A value of 0.0 gives a triangular window, and a value of 1.0 a
        rectangular window.

        If `mw` = `nx` (i.e., no smoothing is carried out), `pw` is not
        used.

    l : int
        L, the frequency division of smoothed spectral estimates as
        2*pi/L.

    lg : int
        Indicates whether unlogged or logged spectral estimates and
        confidence limits are required.

        `lg` = 0

            For unlogged.

        `lg` != 0

            For logged.

    xg : float, array-like, shape (kc)
        The n data points.

    Returns
    -------
    xg : float, ndarray, shape (kc)
        Contains the `ng` spectral estimates f^(omega_i), for i =
        0,1,...,[L/2], in `xg`[0] to `xg`[`ng`-1] (logged if `lg` != 0).
        The elements `xg`[i-1], for i = `ng`+1,...,kc, contain 0.0.

    ng : int
        The number of spectral estimates, [L/2]+1, in `xg`.

    stats : float, ndarray, shape (4)
        Four associated statistics. These are the degrees of freedom in
        `stats`[0], the lower and upper 95% confidence limit factors in
        `stats`[1] and `stats`[2] respectively (logged if `lg` != 0),
        and the bandwidth in `stats`[3].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `px` = *<value>*, `mw` = *<value>* and `nx` =
            *<value>*.

            Constraint: if `pw` > 1.0, `mw` = `nx`.

        (`errno` 1)
            On entry, `px` = *<value>*, `mw` = *<value>* and `nx` =
            *<value>*.

            Constraint: if `pw` < 0.0, `mw` = `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nx` = *<value>*.

            Constraint: `mw` <= `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` <= 1.0.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 0.0.

        (`errno` 1)
            On entry, `mtx` = *<value>*.

            Constraint: `mtx` <= 2.

        (`errno` 1)
            On entry, `mtx` = *<value>*.

            Constraint: `mtx` >= 0.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, kc = *<value>* and `l` = *<value>*.

            Constraint: kc must be a multiple of `l`.

        (`errno` 2)
            On entry, kc = *<value>* and `nx` = *<value>*.

            Constraint: kc >= 2*`nx`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            One or more spectral estimates are negative.

            Unlogged spectral estimates are returned in `xg`, and the
            degrees of freedom, unloged confidence limit factors and
            bandwidth in `stats`.

        (`errno` 5)
            The calculation of confidence limit factors has failed.

            Spectral estimates (logged if requested) are returned in
            `xg`, and degrees of freedom and bandwidth in `stats`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The supplied time series may be mean or trend corrected (by least
    squares), and tapered, the tapering factors being those of the split
    cosine bell:

        [table omitted]

    where T = [np/2] and p is the tapering proportion.

    The unsmoothed sample spectrum

        f*(omega) = 1/2 pi|sum_t = 1^nx_t exp(i omega t)|^2

    is then calculated for frequency values

        omega_k = 2 pi k/K, k = 0,1,...,[K/2],

    where [ ] denotes the integer part.

    The smoothed spectrum is returned as a subset of these frequencies
    for which k is a multiple of a chosen value r, i.e.,

        omega_{rl} = nu_l = 2 pi l/L, l = 0,1,...,[L/2],

    where K = r*L.
    You will normally fix L first, then choose r so that K is
    sufficiently large to provide an adequate representation for the
    unsmoothed spectrum, i.e., K >= 2*n.
    It is possible to take L = K, i.e., r = 1.

    The smoothing is defined by a trapezium window whose shape is
    supplied by the function

        [table omitted]

    the proportion p being supplied by you.

    The width of the window is fixed as 2 pi/M by you supplying M.
    A set of averaging weights are constructed:

        W_k = g*W(omega_k M/pi), 0 <= omega_k <= pi/M,

    where g is a normalizing constant, and the smoothed spectrum
    obtained is

        f^(nu_l) = sum_{|omega_k| < pi/M}W_k f*(nu_l+omega_k).

    If no smoothing is required M should be set to n, in which case the
    values returned are f^(nu_l) = f*(nu_l).
    Otherwise, in order that the smoothing approximates well to an
    integration, it is essential that K≫M, and preferable, but not
    essential, that K be a multiple of M.
    A choice of L > M would normally be required to supply an adequate
    description of the smoothed spectrum.
    Typical choices of L ~= n and K ~= 4n should be adequate for usual
    smoothing situations when M < n/5.

    The sampling distribution of f^(omega) is approximately that of a
    scaled chi_d^2 variate, whose degrees of freedom d is provided by
    the function, together with multiplying limits mu, ml from which
    approximate 95% confidence intervals for the true spectrum f(omega)
    may be constructed as [ml* f^(omega)mu* f^(omega)].
    Alternatively, log f^(omega) may be returned, with additive limits.

    The bandwidth b of the corresponding smoothing window in the
    frequency domain is also provided.
    Spectrum estimates separated by (angular) frequencies much greater
    than b may be assumed to be independent.

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.uni_spectrum_daniell'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _kc_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _kc = _b_utils._EngineIntScalarType(
        _kc_py, {'fun_name': fun_name, 'entity': '_kc'},
    ).to_ctypes()
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalar = _b_tsa.uni_spectrum_daniell(
            nx,
            mtx,
            px,
            mw,
            pw,
            l,
            _kc,
            lg,
            _xg,
            _stats,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _xg_py = _xg
    _ng_py = _return_scalar
    _stats_py = _stats
    _return_data = UniSpectrumDaniellReturnData(
        _xg_py,
        _ng_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiSpectrumLagReturnData = _collections.namedtuple(
    'MultiSpectrumLagReturnData',
    ['cxy', 'cyx', 'xg', 'yg', 'ng'],
)

def multi_spectrum_lag(nxy, mtxy, pxy, iw, mw, ish, ic, cxy, cyx, kc, l, xg=None, yg=None):
    u"""
    Multivariate time series, smoothed sample cross spectrum using
    rectangular, Bartlett, Tukey or Parzen lag window.

    ``multi_spectrum_lag`` calculates the smoothed sample cross spectrum
    of a bivariate time series using one of four lag windows:
    rectangular, Bartlett, Tukey or Parzen.

    For full information please refer to the NAG Library document for
    g13cc

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ccf.html

    Parameters
    ----------
    nxy : int
        n, the length of the time series x and y.

    mtxy : int
        If cross-covariances are to be calculated by the function (`ic`
        = 0), `mtxy` must specify whether the data is to be initially
        mean or trend corrected.

        `mtxy` = 0

            For no correction.

        `mtxy` = 1

            For mean correction.

        `mtxy` = 2

            For trend correction.

        If cross-covariances are supplied (`ic` != 0), `mtxy` is not
        used.

    pxy : float
        If cross-covariances are to be calculated by the function (`ic`
        = 0), `pxy` must specify the proportion of the data (totalled
        over both ends) to be initially tapered by the split cosine bell
        taper. A value of 0.0 implies no tapering.

        If cross-covariances are supplied (`ic` != 0), `pxy` is not
        used.

    iw : int
        The choice of lag window.

        `iw` = 1

            Rectangular.

        `iw` = 2

            Bartlett.

        `iw` = 3

            Tukey.

        `iw` = 4

            Parzen.

    mw : int
        M, the 'cut-off' point of the lag window, relative to any
        alignment shift that has been applied. Windowed
        cross-covariances at lags (-`mw`+`ish`) or less, and at lags
        (`mw`+`ish`) or greater are zero.

    ish : int
        S, the alignment shift between the x and y series. If x leads y,
        the shift is positive.

    ic : int
        Indicates whether cross-covariances are to be calculated in the
        function or supplied in the call to the function.

        `ic` = 0

            Cross-covariances are to be calculated.

        `ic` != 0

            Cross-covariances are to be supplied.

    cxy : float, array-like, shape (nc)
        If `ic` != 0, `cxy` must contain the nc cross-covariances
        between values in the y series and earlier values in time in the
        x series, for lags from 0 to (nc-1).

        If `ic` = 0, `cxy` need not be set.

    cyx : float, array-like, shape (nc)
        If `ic` != 0, `cyx` must contain the nc cross-covariances
        between values in the y series and later values in time in the x
        series, for lags from 0 to (nc-1).

        If `ic` = 0, `cyx` need not be set.

    kc : int
        If `ic` = 0, `kc` must specify the order of the fast Fourier
        transform (FFT) used to calculate the cross-covariances.

        If `ic` != 0, that is if covariances are supplied, `kc` is not
        used.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        Therefore, it is also the order of the FFT used to construct the
        sample spectrum from the cross-covariances.

    xg : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ic`=0: max(`kc`, `l`); otherwise: `l`.

        If the cross-covariances are to be calculated, then `xg` must
        contain the `nxy` data points of the x series. If covariances
        are supplied, `xg` need not be set.

    yg : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ic`=0: max(`kc`, `l`); otherwise: `l`.

        If cross-covariances are to be calculated, `yg` must contain the
        `nxy` data points of the y series. If covariances are supplied,
        `yg` need not be set.

    Returns
    -------
    cxy : float, ndarray, shape (nc)
        If `ic` = 0, `cxy` will contain the nc calculated
        cross-covariances.

        If `ic` != 0, the contents of `cxy` will be unchanged.

    cyx : float, ndarray, shape (nc)
        If `ic` = 0, `cyx` will contain the nc calculated
        cross-covariances.

        If `ic` != 0, the contents of `cyx` will be unchanged.

    xg : float, ndarray, shape (:)
        Contains the real parts of the `ng` complex spectral estimates
        in elements `xg`[0] to `xg`[`ng`-1], and `xg`[`ng`] to
        `xg`[nxyg-1] contain 0.0. The y series leads the x series.

    yg : float, ndarray, shape (:)
        Contains the imaginary parts of the `ng` complex spectral
        estimates in elements `yg`[0] to `yg`[`ng`-1], and `yg`[`ng`] to
        `yg`[nxyg-1] contain 0.0. The y series leads the x series.

    ng : int
        The number, [`l`/2]+1, of complex spectral estimates, whose
        separate parts are held in `xg` and `yg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ic` = *<value>*, nxyg = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` != 0, nxyg >= `l`.

        (`errno` 1)
            On entry, nxyg = *<value>*, `kc` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` = 0, nxyg >= max(`kc`, `l`).

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: if `ic` = 0, `pxy` <= 1.0.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: if `ic` = 0, `pxy` >= 0.0.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: if `ic` = 0 then `mtxy` != 0, 1 or 2.

        (`errno` 1)
            On entry, nc = *<value>* and `nxy` = *<value>*.

            Constraint: nc <= `nxy`.

        (`errno` 1)
            On entry, nc = *<value>*, `mw` = *<value>* and `ish` =
            *<value>*.

            Constraint: nc >= `mw`+abs(`ish`).

        (`errno` 1)
            On entry, `mw` = *<value>*, `ish` = *<value>* and `nxy` =
            *<value>*.

            Constraint: `mw`+abs(`ish`) <= `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `ish` = *<value>* and `mw` = *<value>*.

            Constraint: abs(`ish`) <= `mw`.

        (`errno` 1)
            On entry, `iw` = *<value>*.

            Constraint: `iw` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>*, `nxy` = *<value>* and nc =
            *<value>*.

            Constraint: if `ic` = 0, `kc` >= `nxy`+nc.

        (`errno` 3)
            On entry, `l` = *<value>* and `mw` = *<value>*.

            Constraint: `l` >= 2*`mw`-1.

    Notes
    -----
    The smoothed sample cross spectrum is a complex valued function of
    frequency omega, f_{xy}(omega) = cf(omega)+iqf(omega), defined by
    its real part or co-spectrum

        cf(omega) = 1/2 pisum_k = -M+1^M-1w_k C_{xy}(k+S)cos(omega k)

    and imaginary part or quadrature spectrum

        qf(omega) = 1/2 pisum_k = -M+1^M-1w_k C_{xy}(k+S)sin(omega k)

    where w_k = w_{-k}, for k = 0,1,...,M-1, is the smoothing lag window
    as defined in the description of ``uni_spectrum_lag``.
    The alignment shift S is recommended to be chosen as the lag k at
    which the cross-covariances c_{xy}(k) peak, so as to minimize bias.

    The results are calculated for frequency values

        omega_j = 2 pi j/L, j = 0,1,...,[L/2],

    where [] denotes the integer part.

    The cross-covariances c_{xy}(k) may be supplied by you, or
    constructed from supplied series x_1,x_2,...,x_n; y_1,y_2,...,y_n as

        c_{xy}(k) = sum_t = 1^n-kx_t y_{t+k}/n, k >= 0

        c_{xy}(k) = sum_t = 1-k^nx_t y_{t+k}/n = c_{yx}(-k), k < 0

    this convolution being carried out using the finite Fourier
    transform.

    The supplied series may be mean and trend corrected and tapered
    before calculation of the cross-covariances, in exactly the manner
    described in ``uni_spectrum_lag`` for univariate spectrum
    estimation.
    The results are corrected for any bias due to tapering.

    The bandwidth associated with the estimates is not returned.
    It will normally already have been calculated in previous calls of
    ``uni_spectrum_lag`` for estimating the univariate spectra of y_t
    and x_t.

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_spectrum_lag'

    _cxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cxy'},
        cxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nc_py = _b_utils._EngineFloat64ArrayType.get_shape(_cxy, {'fun_name': fun_name, 'entity': 'cxy'}, exp_rank=1)[0]
    _nc = _b_utils._EngineIntScalarType(
        _nc_py, {'fun_name': fun_name, 'entity': '_nc'},
    ).to_ctypes()
    _cyx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cyx'},
        cyx,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxyg_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1, allow_none=True)[0]
    _nxyg = _b_utils._EngineIntScalarType(
        _nxyg_py, {'fun_name': fun_name, 'entity': '_nxyg'},
    ).to_ctypes()
    _ic = _b_utils._EngineIntScalarType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
    ).to_ctypes()
    _ic_py = _b_utils._EngineIntScalarType.to_py(_ic)
    _kc = _b_utils._EngineIntScalarType(
        kc, {'fun_name': fun_name, 'entity': 'kc'},
    ).to_ctypes()
    _kc_py = _b_utils._EngineIntScalarType.to_py(_kc)
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    if xg is None:
        if _ic_py == 0:
            _xg_dim1 = max(_kc_py, _l_py)
        else:
            _xg_dim1 = _l_py
        exp_shape_xg = (_xg_dim1,)
        _xg = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_xg,
        )
    if yg is None:
        if _ic_py == 0:
            _yg_dim1 = max(_kc_py, _l_py)
        else:
            _yg_dim1 = _l_py
        exp_shape_yg = (_yg_dim1,)
        _yg = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_yg,
        )
    else:
        _yg = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'yg'},
            yg,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _return_scalar = _b_tsa.multi_spectrum_lag(
        nxy,
        mtxy,
        pxy,
        iw,
        mw,
        ish,
        _ic,
        _nc,
        _cxy,
        _cyx,
        _kc,
        _l,
        _nxyg,
        _xg,
        _yg,
    )
    _cxy_py = _cxy
    _cyx_py = _cyx
    _xg_py = _xg
    _yg_py = _yg
    _ng_py = _return_scalar
    _return_data = MultiSpectrumLagReturnData(
        _cxy_py,
        _cyx_py,
        _xg_py,
        _yg_py,
        _ng_py,
    )
    return _return_data

MultiSpectrumDaniellReturnData = _collections.namedtuple(
    'MultiSpectrumDaniellReturnData',
    ['xg', 'yg', 'ng'],
)

def multi_spectrum_daniell(nxy, mtxy, pxy, mw, ish, pw, l, xg, yg):
    u"""
    Multivariate time series, smoothed sample cross spectrum using
    spectral smoothing by the trapezium frequency (Daniell) window.

    ``multi_spectrum_daniell`` calculates the smoothed sample cross
    spectrum of a bivariate time series using spectral smoothing by the
    trapezium frequency (Daniell) window.

    For full information please refer to the NAG Library document for
    g13cd

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13cdf.html

    Parameters
    ----------
    nxy : int
        n, the length of the time series x and y.

    mtxy : int
        Whether the data is to be initially mean or trend corrected.

        `mtxy` = 0

            For no correction.

        `mtxy` = 1

            For mean correction.

        `mtxy` = 2

            For trend correction.

    pxy : float
        The proportion of the data (totalled over both ends) to be
        initially tapered by the split cosine bell taper.

        A value of 0.0 implies no tapering.

    mw : int
        M, the frequency width of the smoothing window as 2*pi/M.

        A value of n implies that no smoothing is to be carried out.

    ish : int
        S, the alignment shift between the x and y series. If x leads y,
        the shift is positive.

    pw : float
        p, the shape parameter of the trapezium frequency window.

        A value of 0.0 gives a triangular window, and a value of 1.0 a
        rectangular window.

        If `mw` = `nxy` (i.e., no smoothing is carried out) then `pw` is
        not used.

    l : int
        L, the frequency division of smoothed cross spectral estimates
        as 2*pi/L.

    xg : float, array-like, shape (kc)
        The `nxy` data points of the x series.

    yg : float, array-like, shape (kc)
        The `nxy` data points of the y series.

    Returns
    -------
    xg : float, ndarray, shape (kc)
        The real parts of the `ng` cross spectral estimates in elements
        `xg`[0] to `xg`[`ng`-1], and `xg`[`ng`] to `xg`[kc-1] contain
        0.0. The y series leads the x series.

    yg : float, ndarray, shape (kc)
        The imaginary parts of the `ng` cross spectral estimates in
        elements `yg`[0] to `yg`[`ng`-1], and `yg`[`ng`] to `yg`[kc-1]
        contain 0.0. The y series leads the x series.

    ng : int
        The number of spectral estimates, [L/2]+1, whose separate parts
        are held in `xg` and `yg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ish` = *<value>* and `l` = *<value>*.

            Constraint: abs(`ish`) < `l`.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `pxy` = *<value>*, `mw` = *<value>* and `nxy` =
            *<value>*.

            Constraint: if `pw` > 1.0, `mw` = `nxy`.

        (`errno` 1)
            On entry, `pxy` = *<value>*, `mw` = *<value>* and `nxy` =
            *<value>*.

            Constraint: if `pw` < 0.0, `mw` = `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nxy` = *<value>*.

            Constraint: `mw` <= `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: `pxy` <= 1.0.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: `pxy` >= 0.0.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: `mtxy` <= 2.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: `mtxy` >= 0.

        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` >= 1.

        (`errno` 2)
            On entry, kc = *<value>* and `l` = *<value>*.

            Constraint: kc must be a multiple of `l`.

        (`errno` 2)
            On entry, kc = *<value>* and `nxy` = *<value>*.

            Constraint: kc >= 2*`nxy`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The supplied time series may be mean and trend corrected and tapered
    as in the description of ``uni_spectrum_daniell`` before calculation
    of the unsmoothed sample cross-spectrum

        f_xy^*(omega) = 1/(2 pi n){sum_t = 1^ny_t exp(i omega t)}*{sum_t
        = 1^nx_t exp(-i omega t)}

    for frequency values omega_j = 2 pi j/K, 0 <= omega_j <= pi.

    A correction is made for bias due to any tapering.

    As in the description of ``uni_spectrum_daniell`` for univariate
    frequency window smoothing, the smoothed spectrum is returned as a
    subset of these frequencies,

        nu_l = 2 pi l/L, l = 0,1,...,[L/2]

    where [ ] denotes the integer part.

    Its real part or co-spectrum cf(nu_l), and imaginary part or
    quadrature spectrum qf(nu_l) are defined by

        f_{xy}(nu_l) = cf(nu_l)+iqf(nu_l) = sum_{|omega_k| < pi/M}w~_k
        f_xy^*(nu_l+omega_k)

    where the weights w~_k are similar to the weights w_k defined for
    ``uni_spectrum_daniell``, but allow for an implicit alignment shift
    S between the series:

        w~_k = w_k exp(-2 pi iSk/L).

    It is recommended that S is chosen as the lag k at which the
    cross-covariances c_{xy}(k) peak, so as to minimize bias.

    If no smoothing is required, the integer M, which determines the
    frequency window width 2 pi/M, should be set to n.

    The bandwidth of the estimates will normally have been calculated in
    a previous call of ``uni_spectrum_daniell`` for estimating the
    univariate spectra of y_t and x_t.

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_spectrum_daniell'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _kc_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _kc = _b_utils._EngineIntScalarType(
        _kc_py, {'fun_name': fun_name, 'entity': '_kc'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_tsa.multi_spectrum_daniell(
        nxy,
        mtxy,
        pxy,
        mw,
        ish,
        pw,
        l,
        _kc,
        _xg,
        _yg,
    )
    _xg_py = _xg
    _yg_py = _yg
    _ng_py = _return_scalar
    _return_data = MultiSpectrumDaniellReturnData(
        _xg_py,
        _yg_py,
        _ng_py,
    )
    return _return_data

MultiSpectrumBivarReturnData = _collections.namedtuple(
    'MultiSpectrumBivarReturnData',
    ['ca', 'calw', 'caup', 't', 'sc', 'sclw', 'scup'],
)

def multi_spectrum_bivar(xg, yg, xyrg, xyig, stats):
    u"""
    Multivariate time series, cross amplitude spectrum, squared
    coherency, bounds, univariate and bivariate (cross) spectra.

    For a bivariate time series, ``multi_spectrum_bivar`` calculates the
    cross amplitude spectrum and squared coherency, together with lower
    and upper bounds from the univariate and bivariate (cross) spectra.

    For full information please refer to the NAG Library document for
    g13ce

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13cef.html

    Parameters
    ----------
    xg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, array-like, shape (ng)
        The real parts, cf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, array-like, shape (ng)
        The imaginary parts, qf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        **Note:** the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    stats : float, array-like, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the
        degrees of freedom, `stats`[1] and `stats`[2] contain the lower
        and upper bound multiplying factors respectively and `stats`[3]
        contains the bandwidth.

    Returns
    -------
    ca : float, ndarray, shape (ng)
        The ng cross amplitude spectral estimates A^(omega) at each
        frequency of omega.

    calw : float, ndarray, shape (ng)
        The ng lower bounds for the ng cross amplitude spectral
        estimates.

    caup : float, ndarray, shape (ng)
        The ng upper bounds for the ng cross amplitude spectral
        estimates.

    t : float
        The critical value for the significance of the squared
        coherency, T.

    sc : float, ndarray, shape (ng)
        The ng squared coherency estimates, W^(omega) at each frequency
        omega.

    sclw : float, ndarray, shape (ng)
        The ng lower bounds for the ng squared coherency estimates.

    scup : float, ndarray, shape (ng)
        The ng upper bounds for the ng squared coherency estimates.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stats`[2] = *<value>*.

            Constraint: `stats`[2] >= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] <= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] > 0.0.

        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

        (`errno` 4)
            A univariate spectral estimate is zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Estimates of the cross amplitude spectrum A(omega) and squared
    coherency W(omega) are calculated for each frequency omega as

        A(omega) = |f_{xy}(omega)| = sqrt(cf(omega)^2+qf(omega)^2)
        and;;W(omega) =
        (|f_{xy}(omega)|^2)/(f_{xx}(omega)f_{yy}(omega)),

    where

        cf(omega) and qf(omega) are the co-spectrum and quadrature
        spectrum estimates between the series, i.e., the real and
        imaginary parts of the cross spectrum f_{xy}(omega) as obtained
        using ``multi_spectrum_lag`` or ``multi_spectrum_daniell``;

        f_{xx}(omega) and f_{yy}(omega) are the univariate spectrum
        estimates for the two series as obtained using
        ``uni_spectrum_lag`` or ``uni_spectrum_daniell``.

    The same type and amount of smoothing should be used for these
    estimates, and this is specified by the degrees of freedom and
    bandwidth values which are passed from the calls of
    ``uni_spectrum_lag`` or ``uni_spectrum_daniell``.

    Upper and lower 95% confidence limits for the cross amplitude are
    given approximately by

        A(omega)[1+/-(1.96/sqrt(d))sqrt(W(omega)^-1+1)],

    except that a negative lower limit is reset to 0.0, in which case
    the approximation is rather poor.
    You are, therefore, particularly recommended to compare the
    coherency estimate W(omega) with the critical value T derived from
    the upper 5% point of the F-distribution on (2,d-2) degrees of
    freedom:

        T = 2F/(d-2+2F),

    where d is the degrees of freedom associated with the univariate
    spectrum estimates.
    The value of T is returned by the function.

    The hypothesis that the series are unrelated at frequency omega,
    i.e., that both the true cross amplitude and coherency are zero, may
    be rejected at the 5% level if W(omega) > T.
    Tests at two frequencies separated by more than the bandwidth may be
    taken to be independent.

    The confidence limits on A(omega) are strictly appropriate only at
    frequencies for which the coherency is significant.
    The same applies to the confidence limits on W(omega) which are
    however calculated at all frequencies using the approximation that
    arctanh(sqrt(W(l))) is Normal with variance 1/d.

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_spectrum_bivar'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyrg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyrg'},
        xyrg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyig'},
        xyig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _stats = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'stats'},
        stats,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ca = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _calw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _caup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _sclw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _scup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    try:
        _return_scalar = _b_tsa.multi_spectrum_bivar(
            _xg,
            _yg,
            _xyrg,
            _xyig,
            _ng,
            _stats,
            _ca,
            _calw,
            _caup,
            _sc,
            _sclw,
            _scup,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ca_py = _ca
    _calw_py = _calw
    _caup_py = _caup
    _t_py = _return_scalar
    _sc_py = _sc
    _sclw_py = _sclw
    _scup_py = _scup
    _return_data = MultiSpectrumBivarReturnData(
        _ca_py,
        _calw_py,
        _caup_py,
        _t_py,
        _sc_py,
        _sclw_py,
        _scup_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiGainBivarReturnData = _collections.namedtuple(
    'MultiGainBivarReturnData',
    ['gn', 'gnlw', 'gnup', 'ph', 'phlw', 'phup'],
)

def multi_gain_bivar(xg, yg, xyrg, xyig, stats):
    u"""
    Multivariate time series, gain, phase, bounds, univariate and
    bivariate (cross) spectra.

    For a bivariate time series, ``multi_gain_bivar`` calculates the
    gain and phase together with lower and upper bounds from the
    univariate and bivariate spectra.

    For full information please refer to the NAG Library document for
    g13cf

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13cff.html

    Parameters
    ----------
    xg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, array-like, shape (ng)
        The real parts, cf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, array-like, shape (ng)
        The imaginary parts, qf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        **Note:** the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    stats : float, array-like, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the
        degrees of freedom, `stats`[1] and `stats`[2] contain the lower
        and upper bound multiplying factors respectively and `stats`[3]
        holds the bandwidth.

    Returns
    -------
    gn : float, ndarray, shape (ng)
        The ng gain estimates, G^(omega), at each frequency omega.

    gnlw : float, ndarray, shape (ng)
        The ng lower bounds for the ng gain estimates.

    gnup : float, ndarray, shape (ng)
        The ng upper bounds for the ng gain estimates.

    ph : float, ndarray, shape (ng)
        The ng phase estimates, phi^(omega), at each frequency omega.

    phlw : float, ndarray, shape (ng)
        The ng lower bounds for the ng phase estimates.

    phup : float, ndarray, shape (ng)
        The ng upper bounds for the ng phase estimates.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

        (`errno` 4)
            A univariate spectral estimate is zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Estimates of the gain G(omega) and phase phi(omega) of the
    dependency of series y on series x at frequency omega are given by

        [table omitted]

    The quantities used in these definitions are obtained as in [Notes
    for multi_spectrum_bivar].

    Confidence limits are returned for both gain and phase, but should
    again be taken as very approximate when the coherency W(omega), as
    calculated by ``multi_spectrum_bivar``, is not significant.
    These are based on the assumption that both (G^(omega)/G(omega))-1
    and phi^(omega) are Normal with variance

        1/d(1/W(omega)-1).

    Although the estimate of phi(omega) is always given in the range
    [0,2 pi), no attempt is made to restrict its confidence limits to
    this range.

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_gain_bivar'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyrg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyrg'},
        xyrg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyig'},
        xyig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _stats = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'stats'},
        stats,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _gn = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _gnlw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _gnup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _ph = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _phlw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _phup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    try:
        _b_tsa.multi_gain_bivar(
            _xg,
            _yg,
            _xyrg,
            _xyig,
            _ng,
            _stats,
            _gn,
            _gnlw,
            _gnup,
            _ph,
            _phlw,
            _phup,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _gn_py = _gn
    _gnlw_py = _gnlw
    _gnup_py = _gnup
    _ph_py = _ph
    _phlw_py = _phlw
    _phup_py = _phup
    _return_data = MultiGainBivarReturnData(
        _gn_py,
        _gnlw_py,
        _gnup_py,
        _ph_py,
        _phlw_py,
        _phup_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiNoiseBivarReturnData = _collections.namedtuple(
    'MultiNoiseBivarReturnData',
    ['er', 'erlw', 'erup', 'rf', 'rfse'],
)

def multi_noise_bivar(xg, yg, xyrg, xyig, stats, l, n):
    u"""
    Multivariate time series, noise spectrum, bounds, impulse response
    function and its standard error.

    For a bivariate time series, ``multi_noise_bivar`` calculates the
    noise spectrum together with multiplying factors for the bounds and
    the impulse response function and its standard error, from the
    univariate and bivariate spectra.

    For full information please refer to the NAG Library document for
    g13cg

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13cgf.html

    Parameters
    ----------
    xg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, array-like, shape (ng)
        The real parts, cf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, array-like, shape (ng)
        The imaginary parts, qf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        **Note:** the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    stats : float, array-like, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the degree
        of freedom, `stats`[1] and `stats`[2] contain the lower and
        upper bound multiplying factors respectively and `stats`[3]
        contains the bandwidth.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        It is also the order of the FFT used to calculate the impulse
        response function. `l` must relate to the parameter ng by the
        relationship.

    n : int
        The number of points in each of the time series x and y. `n`
        should have the same value as nxy in the call of
        ``multi_spectrum_lag`` or ``multi_spectrum_daniell`` which
        calculated the smoothed sample cross spectrum. `n` is used in
        calculating the impulse response function standard error
        (`rfse`).

    Returns
    -------
    er : float, ndarray, shape (ng)
        The ng estimates of the noise spectrum, f^_{y | x}(omega) at
        each frequency.

    erlw : float
        The noise spectrum lower limit multiplying factor.

    erup : float
        The noise spectrum upper limit multiplying factor.

    rf : float, ndarray, shape (`l`)
        The impulse response function. Causal responses are stored in
        ascending frequency in `rf`[0] to `rf`[ng-1] and anticipatory
        responses are stored in descending frequency in `rf`[ng] to
        `rf`[`l`-1].

    rfse : float
        The impulse response function standard error.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `stats`[2] = *<value>*.

            Constraint: `stats`[2] >= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] <= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] > 0.0.

        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 1.

        (`errno` 6)
            On entry, ng = *<value>* and `l` = *<value>*.

            Constraint: ng = `l`/2+1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

            For this frequency the noise spectrum is set to zero, and
            the contribution to the impulse response function and its
            standard error is set to zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

            For this frequency the noise spectrum is set to zero, and
            the contributions to the impulse response function and its
            standard error are set to zero.

        (`errno` 4)
            A univariate spectral estimate is zero.

            For this frequency the noise spectrum is set to zero and the
            contributions to the impulse response function and its
            standard error are set to zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

            For this frequency the squared coherency is reset to 1.0
            with the consequence that the noise spectrum is zero and the
            contribution to the impulse response function at this
            frequency is zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    An estimate of the noise spectrum in the dependence of series y on
    series x at frequency omega is given by

        f_{y | x}(omega) = f_{yy}(omega)(1-W(omega)),

    where W(omega) is the squared coherency described in
    ``multi_spectrum_bivar`` and f_{yy}(omega) is the univariate
    spectrum estimate for series y.
    Confidence limits on the true spectrum are obtained using
    multipliers as described for ``uni_spectrum_lag``, but based on
    (d-2) degrees of freedom.

    If the dependence of y_t on x_t can be assumed to be represented in
    the time domain by the one sided relationship

        y_t = v_0 x_t+v_1 x_{t-1}+...+n_t,

    where the noise n_t is independent of x_t, then it is the spectrum
    of this noise which is estimated by f_{y | x}(omega).

    Estimates of the impulse response function v_0,v_1,v_2,... may also
    be obtained as

        v_k = 1/piint_0^pi Re(exp(ik omega)f_{xy}(omega)/f_{xx}(omega)),

    where Re indicates the real part of the expression.
    For this purpose it is essential that the univariate spectrum for x,
    f_{xx}(omega), and the cross spectrum, f_{xy}(omega), be supplied to
    this function for a frequency range

        omega_l = [2 pi l/L], 0 <= l <= [L/2],

    where [] denotes the integer part, the integral being approximated
    by a finite Fourier transform.

    An approximate standard error is calculated for the estimates v_k.
    Significant values of v_k in the locations described as anticipatory
    responses in the argument array `rf` indicate that feedback exists
    from y_t to x_t.
    This will bias the estimates of v_k in any causal dependence of y_t
    on x_t,x_{t-1},....

    References
    ----------
    Bloomfield, P, 1976, `Fourier Analysis of Time Series: An
    Introduction`, Wiley

    Jenkins, G M and Watts, D G, 1968, `Spectral Analysis and its
    Applications`, Holden--Day
    """

    fun_name = 'naginterfaces.library.tsa.multi_noise_bivar'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyrg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyrg'},
        xyrg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyig'},
        xyig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _stats = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'stats'},
        stats,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _er = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    _rf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l_py,),
    )
    try:
        _return_scalars = _b_tsa.multi_noise_bivar(
            _xg,
            _yg,
            _xyrg,
            _xyig,
            _ng,
            _stats,
            _l,
            n,
            _er,
            _rf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _er_py = _er
    _erlw_py = _return_scalars.erlw
    _erup_py = _return_scalars.erup
    _rf_py = _rf
    _rfse_py = _return_scalars.rfse
    _return_data = MultiNoiseBivarReturnData(
        _er_py,
        _erlw_py,
        _erup_py,
        _rf_py,
        _rfse_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiAutocorrPartReturnData = _collections.namedtuple(
    'MultiAutocorrPartReturnData',
    ['p', 'v0', 'v', 'd', 'db', 'w', 'wb', 'nvp'],
)

def multi_autocorr_part(c0, c, nk):
    u"""
    Multivariate time series, multiple squared partial autocorrelations.

    ``multi_autocorr_part`` calculates the multivariate partial
    autocorrelation function of a multivariate time series.

    For full information please refer to the NAG Library document for
    g13db

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dbf.html

    Parameters
    ----------
    c0 : float, array-like, shape (ns, ns)
        Contains the zero lag cross-covariances between the ns series as
        returned by ``multi_corrmat_cross``. (`c0` is assumed to be
        symmetric, upper triangle only is used.)

    c : float, array-like, shape (ns, ns, nl)
        Contains the cross-covariances at lags 1 to nl. `c`[i-1,j-1,k-1]
        must contain the cross-covariance, c_{ijk}, of series i and
        series j at lag k. Series j leads series i.

    nk : int
        The number of lags to which partial auto-correlations are to be
        calculated.

    Returns
    -------
    p : float, ndarray, shape (`nvp`)
        The multiple squared partial autocorrelations from lags 1 to
        `nvp`; that is, `p`[l-1] contains p_l^2, for l = 1,2,...,`nvp`.
        For lags `nvp`+1 to `nk` the elements of `p` are set to zero.

    v0 : float
        The lag zero prediction error variance (equal to the determinant
        of `c0`).

    v : float, ndarray, shape (`nvp`)
        The prediction error variance ratios from lags 1 to `nvp`; that
        is, `v`[l-1] contains v_l, for l = 1,2,...,`nvp`. For lags
        `nvp`+1 to `nk` the elements of `v` are set to zero.

    d : float, ndarray, shape (ns, ns, `nk`)
        The prediction error variance matrices at lags 1 to `nvp`.

        Element (i,j,k) of `d` contains the prediction error covariance
        of series i and series j at lag k, for k = 1,2,...,`nvp`.

        Series j leads series i; that is, the (i,j)th element of D_k.

        For lags `nvp`+1 to `nk` the elements of `d` are set to zero.

    db : float, ndarray, shape (ns, ns)
        The backward prediction error variance matrix at lag `nvp`.

        `db`[i-1,j-1] contains the backward prediction error covariance
        of series i and series j; that is, the (i,j)th element of the
        G_k, where k = `nvp`.

    w : float, ndarray, shape (ns, ns, `nk`)
        The prediction coefficient matrices at lags 1 to `nvp`.

        `w`[i-1,j-1,l-1] contains the jth prediction coefficient of
        series i at lag l; that is, the (i,j)th element of Phi_{kl},
        where k = `nvp`, for l = 1,2,...,`nvp`.

        For lags `nvp`+1 to `nk` the elements of `w` are set to zero.

    wb : float, ndarray, shape (ns, ns, `nk`)
        The backward prediction coefficient matrices at lags 1 to `nvp`.

        `wb`[i-1,j-1,l-1] contains the jth backward prediction
        coefficient of series i at lag l; that is, the (i,j)th element
        of Psi_{kl}, where k = `nvp`, for l = 1,2,...,`nvp`.

        For lags `nvp`+1 to `nk` the elements of `wb` are set to zero.

    nvp : int
        The maximum lag, L, for which calculation of `p`, `v`, `d`,
        `db`, `w` and `wb` was successful. If the function completes
        successfully `nvp` will equal `nk`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nk` = *<value>* and nl = *<value>*.

            Constraint: `nk` <= nl.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` >= 1.

        (`errno` 1)
            On entry, nl = *<value>*.

            Constraint: nl >= 1.

        (`errno` 1)
            On entry, ns = *<value>* and ldc0 = *<value>*.

            Constraint: ns <= ldc0.

        (`errno` 1)
            On entry, ns = *<value>*.

            Constraint: ns >= 1.

        (`errno` 2)
            `c0` is not positive definite.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            For `nvp` = *<value>*, at lag k = `nvp`+1, D_k was found not
            to be positive definite.

    Notes
    -----
    The input is a set of lagged autocovariance matrices
    C_0,C_1,C_2,...,C_m.
    These will generally be sample values such as are obtained from a
    multivariate time series using ``multi_corrmat_cross``.

    The main calculation is the recursive determination of the
    coefficients in the finite lag (forward) prediction equation

        x_t = Phi_{l,1}x_{t-1}+...+Phi_{l,l}x_{t-l}+e_{l,t}

    and the associated backward prediction equation

        x_{t-l-1} = Psi_{l,1}x_{t-l}+...+Psi_{l,l}x_{t-1}+f_{l,t}

    together with the covariance matrices D_l of e_{l,t} and G_l of
    f_{l,t}.

    The recursive cycle, by which the order of the prediction equation
    is extended from l to l+1, is to calculate

        M_{l+1} = C_{l+1}^T-Phi_{l,1}C_l^T-...-Phi_{l,l} C_1^T

    then Phi_{l+1,l+1} = M_{l+1}D_l^-1,  Psi_{l+1,l+1} = M^T_{l+1}G_l^-1

    from which

        Phi_{l+1,j} = Phi_{l,j}-Phi_{l+1,l+1}Psi_{l,l+1-j}, j =
        1,2,...,l

    and

        Psi_{l+1,j} = Psi_{l,j}-Psi_{l+1,l+1}Phi_{l,l+1-j}, j =
        1,2,...,l.

    Finally, D_{l+1} = D_l-M_{l+1}Phi^T_{l+1,l+1} and G_{l+1} =
    G_l-M^T_{l+1}Psi^T_{l+1,l+1}.

    (Here T denotes the transpose of a matrix.)

    The cycle is initialized by taking (for l = 0)

        D_0 = G_0 = C_0.

    In the step from l = 0 to 1, the above equations contain redundant
    terms and simplify.
    Thus (1) becomes M_1 = C^T_1 and neither (2) and (3) are needed.

    Quantities useful in assessing the effectiveness of the prediction
    equation are generalized variance ratios

        v_l = det(D_l)/det(C_0), l = 1,2,...

    and multiple squared partial autocorrelations

        p_l^2 = 1-v_l/v_{l-1}.

    References
    ----------
    Akaike, H, 1971, `Autoregressive model fitting for control`, Ann.
    Inst. Statist. Math. (23), 163--180

    Whittle, P, 1963, `On the fitting of multivariate autoregressions
    and the approximate canonical factorization of a spectral density
    matrix`, Biometrika (50), 129--134
    """

    fun_name = 'naginterfaces.library.tsa.multi_autocorr_part'

    _c0 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c0'},
        c0,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c0, 'c0'), (_c, 'c')),
    )
    _ns_py = _b_utils._EngineFloat64ArrayType.get_shape(_c0, {'fun_name': fun_name, 'entity': 'c0'}, exp_rank=2)[0]
    _ns = _b_utils._EngineIntScalarType(
        _ns_py, {'fun_name': fun_name, 'entity': '_ns'},
    ).to_ctypes()
    _nl_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=3)[2]
    _nl = _b_utils._EngineIntScalarType(
        _nl_py, {'fun_name': fun_name, 'entity': '_nl'},
    ).to_ctypes()
    _nk = _b_utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = _b_utils._EngineIntScalarType.to_py(_nk)
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nk_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nk_py,),
    )
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py, _nk_py),
        sorder=_m_sorder,
    )
    _db = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py),
        sorder=_m_sorder,
    )
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py, _nk_py),
        sorder=_m_sorder,
    )
    _wb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py, _nk_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = _b_tsa.multi_autocorr_part(
            _c0,
            _c,
            _ns,
            _nl,
            _nk,
            _p,
            _v,
            _d,
            _db,
            _w,
            _wb,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _nvp_py = _return_scalars.nvp
    _p_py = _p[:_nvp_py]
    _v0_py = _return_scalars.v0
    _v_py = _v[:_nvp_py]
    _d_py = _d[:, :, :_nvp_py]
    _db_py = _db
    _w_py = _w[:, :, :_nvp_py]
    _wb_py = _wb[:, :, :_nvp_py]
    _return_data = MultiAutocorrPartReturnData(
        _p_py,
        _v0_py,
        _v_py,
        _d_py,
        _db_py,
        _w_py,
        _wb_py,
        _nvp_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiVarmaEstimateReturnData = _collections.namedtuple(
    'MultiVarmaEstimateReturnData',
    ['par', 'qq', 'niter', 'rlogl', 'v', 'g', 'cm'],
)

def multi_varma_estimate(ip, iq, mean, par, qq, w, parhld, exact, iprint, cgetol, ishow, maxcal=None, io_manager=None):
    u"""
    Multivariate time series, estimation of VARMA model.

    ``multi_varma_estimate`` fits a vector autoregressive moving average
    (VARMA) model to an observed vector of time series using the method
    of Maximum Likelihood (ML).
    Standard errors of parameter estimates are computed along with their
    appropriate correlation matrix.
    The function also calculates estimates of the residual series.

    For full information please refer to the NAG Library document for
    g13dd

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ddf.html

    Parameters
    ----------
    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    mean : bool
        `mean` = True, if components of mu have been estimated and
        `mean` = False, if all elements of mu are to be taken as zero.

    par : float, array-like, shape (npar)
        Initial parameter estimates read in row by row in the order
        phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q,mu.

        Thus,

            if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal
            to an initial estimate of the (i,j)th element of phi_l, for
            j = 1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,p;

            if `iq` > 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
            equal to an initial estimate of the (i,j)th element of
            theta_l, l = 1,2,...,q and i,j = 1,2,...,k;

            if `mean` = True, `par`[(p+q)*k*k+i-1] should be set equal
            to an initial estimate of the ith component of mu (mu(i)).
            (If you set `par`[(p+q)*k*k+i-1] to 0.0 then
            ``multi_varma_estimate`` will calculate the mean of the ith
            series and use this as an initial estimate of mu(i).)

        The first p*k*k elements of `par` must satisfy the stationarity
        condition and the next q*k*k elements of `par` must satisfy the
        invertibility condition.

        If in doubt set all elements of `par` to 0.0.

    qq : float, array-like, shape (k, k)
        `qq`[i-1,j-1] must be set equal to an initial estimate of the
        (i,j)th element of Sigma. The lower triangle only is needed.
        `qq` must be positive definite. It is strongly recommended that
        on entry the elements of `qq` are of the same order of magnitude
        as at the solution point. If you set `qq`[i-1,j-1] = 0.0, for j
        = 1,2,...,i, for i = 1,2,...,k, then ``multi_varma_estimate``
        will calculate the covariance matrix between the k time series
        and use this as an initial estimate of Sigma.

    w : float, array-like, shape (k, n)
        `w`[i-1,t-1] must be set equal to the ith component of W_t, for
        t = 1,2,...,n, for i = 1,2,...,k.

    parhld : bool, array-like, shape (npar)
        `parhld`[i-1] must be set to True if `par`[i-1] is to be held
        constant at its input value and False if `par`[i-1] is a free
        parameter, for i = 1,2,...,npar.

        If in doubt try setting all elements of `parhld` to False.

    exact : bool
        Must be set equal to True if you wish ``multi_varma_estimate``
        to compute exact maximum likelihood estimates. `exact` must be
        set equal to False if only conditional likelihood estimates are
        required.

    iprint : int
        The frequency with which the automatic monitoring function is to
        be called.

        `iprint` > 0

            The ML search procedure is monitored once every `iprint`
            iterations and just before exit from the search function.

        `iprint` = 0

            The search function is monitored once at the final point.

        `iprint` < 0

            The search function is not monitored at all.

    cgetol : float
        The accuracy to which the solution in `par` and `qq` is
        required.

        If `cgetol` is set to 10^(-l) and on exit no exception or
        warning is raised or `errno` >= 6, then all the elements in
        `par` and `qq` should be accurate to approximately l decimal
        places.

        For most practical purposes the value 10^-4 should suffice.

        You should be wary of setting `cgetol` too small since the
        convergence criteria may then have become too strict for the
        machine to handle.

        If `cgetol` has been set to a value which is less than the
        machine precision, epsilon, then ``multi_varma_estimate`` will
        use the value 10.0*sqrt(epsilon) instead.

    ishow : int
        Specifies which of the following two quantities are to be
        printed.

        (i) table of maximum likelihood estimates and their standard
            errors (as returned in the output arrays `par`, `qq` and
            `cm`);

        (#) table of residual series (as returned in the output array
            `v`).

        `ishow` = 0

            None of the above are printed.

        `ishow` = 1

            (i) only is printed.

        `ishow` = 2

            (i) and (ii) are printed.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 40*npar*(npar+5).

        The maximum number of likelihood evaluations to be permitted by
        the search procedure.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    par : float, ndarray, shape (npar)
        If no exception or warning is raised or `errno` >= 4 then all
        the elements of `par` will be overwritten by the latest
        estimates of the corresponding ARMA parameters.

    qq : float, ndarray, shape (k, k)
        If no exception or warning is raised or `errno` >= 4 then
        `qq`[i-1,j-1] will contain the latest estimate of the (i,j)th
        element of Sigma. The lower triangle only is returned.

    niter : int
        If no exception or warning is raised or `errno` >= 4 then
        `niter` contains the number of iterations performed by the
        search function.

    rlogl : float
        If no exception or warning is raised or `errno` >= 4 then
        `rlogl` contains the value of the log-likelihood function
        corresponding to the final point held in `par` and `qq`.

    v : float, ndarray, shape (k, n)
        If no exception or warning is raised or `errno` >= 4 then
        `v`[i-1,t-1] will contain an estimate of the ith component of
        epsilon_t, for t = 1,2,...,n, for i = 1,2,...,k, corresponding
        to the final point held in `par` and `qq`.

    g : float, ndarray, shape (npar)
        If no exception or warning is raised or `errno` >= 4 then
        `g`[i-1] will contain the estimated first derivative of the
        log-likelihood function with respect to the ith element in the
        array `par`. If the gradient cannot be computed then all the
        elements of `g` are returned as zero.

    cm : float, ndarray, shape (npar, npar)
        If no exception or warning is raised or `errno` >= 4 then
        `cm`[i-1,j-1] will contain an estimate of the correlation
        coefficient between the ith and jth elements in the `par` array
        for 1 <= i <= npar, 1 <= j <= npar. If i = j, then `cm`[i-1,j-1]
        will contain the estimated standard error of `par`[i-1]. If the
        lth component of `par` has been held constant, i.e.,
        `parhld`[l-1] was set to True, then the lth row and column of
        `cm` will be set to zero. If the second derivative matrix cannot
        be computed then all the elements of `cm` are returned as zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar >= 0.

        (`errno` 1)
            On entry, n = *<value>*, k = *<value>* and npar = *<value>*.

            Constraint: n*k > npar+k*(k+1)/2.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = 0 and `iq` = 0.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = *<value>*.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `ishow` = *<value>*.

            Constraint: 0 <= `ishow` <= 2.

        (`errno` 2)
            The initial estimate of Sigma is not positive definite.

        (`errno` 2)
            The initial AR parameter estimates are outside the
            stationarity region.

        (`errno` 2)
            The initial MA parameter estimates are outside the
            invertibility region.

        (`errno` 2)
            The starting point is too close to the boundary of the
            admissibility region.

        (`errno` 3)
            The function cannot compute a sufficiently accurate estimate
            of the gradient vector at the user-supplied starting point.

        (`errno` 4)
            There have been `maxcal` log-likelihood evaluations made in
            the function.

            See the function document for further details.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The conditions for a solution have not all been met, but a
            point at which the log-likelihood took a larger value could
            not be found.

            See the function document for further details.

        (`errno` 6)
            The ML solution is so close to the boundary of either the
            stationarity region or the invertibility region that
            ``multi_varma_estimate`` cannot evaluate the Hessian matrix.
            The elements of `cm` are set to zero, as are the elements of
            `g`. All other output quantities are correct.

        (`errno` 7)
            An estimate of the second derivative matrix and the gradient
            vector at the solution point was computed. Either the
            Hessian matrix was found to be too ill-conditioned to be
            evaluated accurately or the gradient vector could not be
            computed to an acceptable degree of accuracy. The elements
            of `cm` are set to zero, as are the elements of `g`. All
            other output quantities are correct.

        (`errno` 8)
            The second-derivative matrix at the solution point is not
            positive definite. The elements of `cm` are set to zero. All
            other output quantities are correct.

    Notes
    -----
    Let W_t = (w_{1t},w_{2t},...,w_{kt})^T, for t = 1,2,...,n, denote a
    vector of k time series which is assumed to follow a multivariate
    ARMA model of the form

        [table omitted]

    where epsilon_t = (epsilon_{1t},epsilon_{2t},...,epsilon_{kt})^T,
    for t = 1,2,...,n, is a vector of k residual series assumed to be
    Normally distributed with zero mean and positive definite covariance
    matrix Sigma.
    The components of epsilon_t are assumed to be uncorrelated at
    non-simultaneous lags.
    The phi_i and theta_j are k*k matrices of parameters. {phi_i}, for i
    = 1,2,...,p, are called the autoregressive (AR) parameter matrices,
    and {theta_i}, for i = 1,2,...,q, the moving average (MA) parameter
    matrices.
    The parameters in the model are thus the p (k*k) phi-matrices, the q
    (k*k) theta-matrices, the mean vector, mu, and the residual error
    covariance matrix Sigma.
    Let

        A(phi) = [[table omitted]]_{pk*pk} and B(theta) = [[table
        omitted]]_{qk*qk}

    where I denotes the k*k identity matrix.

    The ARMA model (1) is said to be stationary if the eigenvalues of
    A(phi) lie inside the unit circle.
    Similarly, the ARMA model (1) is said to be invertible if the
    eigenvalues of B(theta) lie inside the unit circle.

    The method of computing the exact likelihood function (using a
    Kalman filter algorithm) is discussed in Shea (1987).
    A quasi-Newton algorithm (see Gill and Murray (1972)) is then used
    to search for the maximum of the log-likelihood function.
    Stationarity and invertibility are enforced on the model using the
    reparameterisation discussed in Ansley and Kohn (1986).
    Conditional on the maximum likelihood estimates being equal to their
    true values the estimates of the residual series are uncorrelated
    with zero mean and constant variance Sigma.

    You have the option of setting an argument (`exact` to False) so
    that ``multi_varma_estimate`` calculates conditional maximum
    likelihood estimates (conditional on W_0 = W_-1 = ... = W_{1-p} =
    epsilon_0 = epsilon_-1 = ... = epsilon_{1-q} = 0).
    This may be useful if the exact maximum likelihood estimates are
    close to the boundary of the invertibility region.

    You also have the option (see [Parameters]) of requesting
    ``multi_varma_estimate`` to constrain elements of the phi and theta
    matrices and mu vector to have pre-specified values.

    References
    ----------
    Ansley, C F and Kohn, R, 1986, `A note on reparameterising a vector
    autoregressive moving average model to enforce stationarity`, J.
    Statist. Comput. Simulation (24), 99--106

    Gill, P E and Murray, W, 1972, `Quasi-Newton methods for
    unconstrained optimization`, J. Inst. Math. Appl. (9), 91--108

    Shea, B L, 1987, `Estimation of multivariate time series`, J. Time
    Ser. Anal. (8), 95--110
    """

    fun_name = 'naginterfaces.library.tsa.multi_varma_estimate'

    _qq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qq'},
        qq,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_qq, 'qq'), (_w, 'w')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_qq, {'fun_name': fun_name, 'entity': 'qq'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _parhld = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parhld'},
        parhld,
        copy=False,
    )
    _maxcal = 40*_npar_py*(_npar_py+5) if maxcal is None else maxcal
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _n_py),
        sorder=_m_sorder,
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _cm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = _b_tsa.multi_varma_estimate(
            _k,
            _n,
            ip,
            iq,
            mean,
            _par,
            _npar,
            _qq,
            _w,
            _parhld,
            exact,
            iprint,
            cgetol,
            _maxcal,
            ishow,
            _v,
            _g,
            _cm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _par_py = _par
    _qq_py = _qq
    _niter_py = _return_scalars.niter
    _rlogl_py = _return_scalars.rlogl
    _v_py = _v
    _g_py = _g
    _cm_py = _cm
    _return_data = MultiVarmaEstimateReturnData(
        _par_py,
        _qq_py,
        _niter_py,
        _rlogl_py,
        _v_py,
        _g_py,
        _cm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiVarmaForecastReturnData = _collections.namedtuple(
    'MultiVarmaForecastReturnData',
    ['qq', 'predz', 'sefz', 'ref'],
)

def multi_varma_forecast(z, tr, dord, delta, ip, iq, mean, par, qq, v, lmax, lref):
    u"""
    Multivariate time series, forecasts and their standard errors.

    ``multi_varma_forecast`` computes forecasts of a multivariate time
    series.
    It is assumed that a vector ARMA model has already been fitted to
    the appropriately differenced/transformed time series using
    ``multi_varma_estimate``.
    The standard deviations of the forecast errors are also returned.
    A reference vector is set up so that, should future series values
    become available, the forecasts and their standard errors may be
    updated by calling ``multi_varma_update``.

    For full information please refer to the NAG Library document for
    g13dj

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13djf.html

    Parameters
    ----------
    z : float, array-like, shape (k, n)
        `z`[i,t] must contain, z_{it}, the ith component of Z_t, for t =
        1,2,...,n, for i = 1,2,...,k.

    tr : str, length 1, array-like, shape (k)
        `tr`[i-1] indicates whether the ith time series is to be
        transformed, for i = 1,2,...,k.

        `tr`[i-1] = 'N'

            No transformation is used.

        `tr`[i-1] = 'L'

            A log transformation is used.

        `tr`[i-1] = 'S'

            A square root transformation is used.

    dord : int, array-like, shape (k)
        `dord`[i-1] must specify, d_i, the order of differencing
        required for the ith series.

    delta : float, array-like, shape (k, d)
        If `dord`[i-1] > 0, then `delta`[i-1,j-1] must be set equal to
        delta_{ij}, for i = 1,2,...,k, for j = 1,2,...,d_i.

        If d = 0, `delta` is not referenced.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    mean : str, length 1
        `mean` = 'M', if components of mu have been estimated and `mean`
        = 'Z', if all elements of mu are to be taken as zero.

    par : float, array-like, shape (lpar)
        Must contain the parameter estimates read in row by row in the
        order phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q, mu.

        Thus,

            if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal
            to an estimate of the (i,j)th element of phi_l, for j =
            1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,p;

            if `iq` > 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
            equal to an estimate of the (i,j)th element of theta_l, for
            j = 1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,q;

            if `mean` = 'M', `par`[(p+q)*k*k+i-1] must be set equal to
            an estimate of the ith component of mu, for i = 1,2,...,k.

    qq : float, array-like, shape (k, k)
        `qq`[i-1,j-1] must contain an estimate of the (i,j)th element of
        Sigma. The lower triangle only is needed.

    v : float, array-like, shape (k, n-d)
        `v`[i-1,t-1] must contain an estimate of the ith component of
        epsilon_{t+d}, for t = 1,2,...,n-d, for i = 1,2,...,k.

        If q = 0, `v` is not used.

    lmax : int
        The number, l_max, of forecasts required.

    lref : int
        The dimension of the array `ref`.

    Returns
    -------
    qq : float, ndarray, shape (k, k)
        If `errno` != 1, then the upper triangle is set equal to the
        lower triangle.

    predz : float, ndarray, shape (k, `lmax`)
        `predz`[i-1,l-1] contains the forecast of z_{i,n+l}, for l =
        1,2,...,l_max, for i = 1,2,...,k.

    sefz : float, ndarray, shape (k, `lmax`)
        `sefz`[i-1,l-1] contains an estimate of the standard error of
        the forecast of z_{i,n+l}, for l = 1,2,...,l_max, for i =
        1,2,...,k.

    ref : float, ndarray, shape (`lref`)
        The reference vector which may be used to update forecasts using
        ``multi_varma_update``. The first (`lmax`-1)*k*k elements
        contain the psi weight matrices, psi_1,psi_2,...,psi_{l_max-1}.
        The next k*`lmax` elements contain the forecasts of the
        transformed series Z^_n+1^*,Z^_n+2^*,...,Z^_n+l_max^* and the
        next k*`lmax` contain the variances of the forecasts of the
        transformed variables. The last k elements are used to store the
        transformations for the series.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lref` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `lref` >= (`lmax`-1)*k*k+2*k*`lmax`+k.

        (`errno` 1)
            On entry, i = *<value>*, `dord`[i-1] = *<value>* and
            n-max(`ip`, `iq`) = *<value>*.

            Constraint: 0 <= `dord`[i-1] < n-max(`ip`, `iq`).

        (`errno` 1)
            On entry, `lmax` = *<value>*.

            Constraint: `lmax` >= 1.

        (`errno` 1)
            On entry, number of observations = *<value>* and number of
            parameters = *<value>*.

            Constraint: number of observations >= number of parameters.

        (`errno` 1)
            On entry, lpar is too small: lpar = *<value>* and the
            minimum size required = *<value>*.

        (`errno` 1)
            On entry, `mean` is an invalid character.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `tr`[i-1] is invalid.

            Constraint: `tr`[i-1] = 'N', 'L' or 'S'.

        (`errno` 3)
            On entry, one (or more) of the transformations requested is
            invalid.

        (`errno` 4)
            On entry, the MA parameter matrices are outside the
            invertibility region.

        (`errno` 4)
            On entry, the AR parameter matrices are outside the
            stationarity region.

        (`errno` 4)
            On entry, the covariance matrix `qq` is not positive
            definite.

        (`errno` 5)
            An excessive number of iterations were needed to evaluate
            the eigenvalues of the matrices used to test for
            stationarity and invertibility.

        (`errno` 6)
            The covariance matrix may be nearly non-positive definite.

        (`errno` 7)
            The forecasts will overflow if computed.

    Notes
    -----
    Let the vector Z_t = (z_{1t},z_{2t},...,z_{kt})^T, for t =
    1,2,...,n, denote a k-dimensional time series for which forecasts of
    Z_{n+1},Z_{n+2},...,Z_{n+l_max} are required.
    Let W_t = (w_{1t},w_{2t},...,w_{kt})^T be defined as follows:

        w_{it} = delta_i(B)z_it^*, i = 1,2,...,k,

    where delta_i(B) is the differencing operator applied to the ith
    series and where z_it^* is equal to either z_{it}, sqrt(z_{it}) or
    loge(z_{it}) depending on whether or not a transformation was
    required to stabilize the variance before fitting the model.

    If the order of differencing required for the ith series is d_i,
    then the differencing operator for the ith series is defined by
    delta_i(B) = 1-delta_{i1}B-delta_{i2}B^2-...-delta_{id_i}B^(d_i)
    where B is the backward shift operator; that is, BZ_t = Z_{t-1}.
    The differencing parameters delta_{ij}, for j = 1,2,...,d_i, for i =
    1,2,...,k, must be supplied by you.
    If the ith series does not require differencing, then d_i = 0.

    W_t is assumed to follow a multivariate ARMA model of the form:

        W_t-mu =
        phi_1(W_{t-1}-mu)+phi_2(W_{t-2}-mu)+...+phi_p(W_{t-p}-mu)+epsilon_t-theta_1epsilon_{t-1}-...-theta_qepsilon_{t-q},

    where epsilon_t = (epsilon_{1t},epsilon_{2t},...,epsilon_{kt})^T,
    for t = 1,2,...,n, is a vector of k residual series assumed to be
    Normally distributed with zero mean and positive definite covariance
    matrix Sigma.
    The components of epsilon_t are assumed to be uncorrelated at
    non-simultaneous lags.
    The phi_i and theta_j are k*k matrices of parameters.
    The matrices phi_i, for i = 1,2,...,p, are the autoregressive (AR)
    parameter matrices, and the matrices theta_i, for i = 1,2,...,q, the
    moving average (MA) parameter matrices.
    The parameters in the model are thus the p (k*k) phi-matrices, the q
    (k*k) theta-matrices, the mean vector mu and the residual error
    covariance matrix Sigma.
    The ARMA model (1) must be both stationary and invertible; see
    ``uni_arma_roots`` for a method of checking these conditions.

    The ARMA model (1) may be rewritten as

        phi(B)(delta(B)Z_t^*-mu) = theta(B)epsilon_t,

    where phi(B) and theta(B) are the autoregressive and moving average
    polynomials and delta(B) denotes the k*k diagonal matrix whose ith
    diagonal elements is delta_i(B) and Z_t^* =
    (z_1t^*,z_2t^*...z_kt^*)^T.

    This may be rewritten as

        phi(B)delta(B)Z_t^* = phi(B)mu+theta(B)epsilon_t

    or

        Z_t^* = tau+psi(B)epsilon_t =
        tau+epsilon_t+psi_1epsilon_{t-1}+psi_2epsilon_{t-2}+...

    where psi(B) = delta^-1(B)phi^-1(B)theta(B) and tau = delta^-1(B)mu
    is a vector of length k.

    Forecasts are computed using a multivariate version of the procedure
    described in Box and Jenkins (1976).
    If Z^_n^*(l) denotes the forecast of Z_n+l^*, then Z^_n^*(l) is
    taken to be that linear function of Z_n^*,Z_n-1^*,... which
    minimizes the elements of E{e_n(l)e'_n(l)} where e_n(l) =
    Z_n+l^*-Z^_n^*(l) is the forecast error. Z^_n^*(l) is referred to as
    the linear minimum mean square error forecast of Z_n+l^*.

    The linear predictor which minimizes the mean square error may be
    expressed as

        Z^_n^*(l) =
        tau+psi_lepsilon_n+psi_{l+1}epsilon_{n-1}+psi_{l+2}epsilon_{n-2}+....

    The forecast error at t for lead l is then

        e_n(l) = Z_n+l^*-Z^_n^*(l) =
        epsilon_{n+l}+psi_1epsilon_{n+l-1}+psi_2epsilon_{n+l-2}+...+psi_{l-1}epsilon_{n+1}.

    Let d = max(d_i), for i = 1,2,...,k.
    Unless q = 0 the function requires estimates of epsilon_t, for t =
    d+1,...,n, which are obtainable from ``multi_varma_estimate``.
    The terms epsilon_t are assumed to be zero, for t = n+1,...,n+l_max.
    You may use ``multi_varma_update`` to update these l_max forecasts
    should further observations, Z_{n+1},Z_{n+2},..., become available.
    Note that when l_max or more further observations are available then
    ``multi_varma_forecast`` must be used to produce new forecasts for
    Z_{n+l_max+1},Z_{n+l_max+2},..., should they be required.

    When a transformation has been used the forecasts and their standard
    errors are suitably modified to give results in terms of the
    original series, Z_t; see Granger and Newbold (1976).

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Granger, C W J and Newbold, P, 1976, `Forecasting transformed
    series`, J. Roy. Statist. Soc. Ser. B (38), 189--203

    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.tsa.multi_varma_forecast'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _delta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'delta'},
        delta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _qq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qq'},
        qq,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_delta, 'delta'), (_qq, 'qq'), (_v, 'v')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _tr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tr'},
        tr,
        copy=False,
    )
    _dord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dord'},
        dord,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lpar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _lpar = _b_utils._EngineIntScalarType(
        _lpar_py, {'fun_name': fun_name, 'entity': '_lpar'},
    ).to_ctypes()
    _lmax = _b_utils._EngineIntScalarType(
        lmax, {'fun_name': fun_name, 'entity': 'lmax'},
    ).to_ctypes()
    _lmax_py = _b_utils._EngineIntScalarType.to_py(_lmax)
    _predz = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _lmax_py),
        sorder=_m_sorder,
    )
    _sefz = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _lmax_py),
        sorder=_m_sorder,
    )
    _lref = _b_utils._EngineIntScalarType(
        lref, {'fun_name': fun_name, 'entity': 'lref'},
    ).to_ctypes()
    _lref_py = _b_utils._EngineIntScalarType.to_py(_lref)
    _ref = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lref_py,),
    )
    _b_tsa.multi_varma_forecast(
        _k,
        _n,
        _z,
        _tr,
        _dord,
        _delta,
        ip,
        iq,
        mean,
        _par,
        _lpar,
        _qq,
        _v,
        _lmax,
        _predz,
        _sefz,
        _ref,
        _lref,
    )
    _qq_py = _qq
    _predz_py = _predz
    _sefz_py = _sefz
    _ref_py = _ref
    _return_data = MultiVarmaForecastReturnData(
        _qq_py,
        _predz_py,
        _sefz_py,
        _ref_py,
    )
    return _return_data

MultiVarmaUpdateReturnData = _collections.namedtuple(
    'MultiVarmaUpdateReturnData',
    ['mlast', 'ref', 'v', 'predz', 'sefz'],
)

def multi_varma_update(mlast, z, ref, predz, sefz):
    u"""
    Multivariate time series, updates forecasts and their standard
    errors.

    ``multi_varma_update`` accepts a sequence of new observations in a
    multivariate time series and updates both the forecasts and the
    standard deviations of the forecast errors.
    A call to ``multi_varma_forecast`` must be made prior to calling
    this function in order to calculate the elements of a reference
    vector together with a set of forecasts and their standard errors.
    On a successful exit from ``multi_varma_update`` the reference
    vector is updated so that should future series values become
    available these forecasts may be updated by recalling
    ``multi_varma_update``.

    For full information please refer to the NAG Library document for
    g13dk

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dkf.html

    Parameters
    ----------
    mlast : int
        On the first call to ``multi_varma_update``, since calling
        ``multi_varma_forecast``, `mlast` must be set to 0 to indicate
        that no new observations have yet been used to update the
        forecasts; on subsequent calls `mlast` must contain the value of
        `mlast` as output on the previous call to
        ``multi_varma_update``.

    z : float, array-like, shape (k, m)
        `z`[i-1,j-1] must contain the value of z_{i,n+`mlast`+j}, for j
        = 1,2,...,m, for i = 1,2,...,k, and where n is the number of
        observations in the time series in the last call made to
        ``multi_varma_forecast``.

    ref : float, array-like, shape (lref)
        Must contain the first (lmax-1)*k*k+2*k*lmax+k elements of the
        reference vector as returned on a successful exit from
        ``multi_varma_forecast`` (or a previous call to
        ``multi_varma_update``).

    predz : float, array-like, shape (k, lmax)
        Nonupdated values are kept intact.

    sefz : float, array-like, shape (k, lmax)
        Nonupdated values are kept intact.

    Returns
    -------
    mlast : int
        Is incremented by m to indicate that `mlast`+m observations have
        now been used to update the forecasts since the last call to
        ``multi_varma_forecast``.

        `mlast` must not be changed between calls to
        ``multi_varma_update``, unless a call to
        ``multi_varma_forecast`` has been made between the calls in
        which case `mlast` should be reset to 0.

    ref : float, ndarray, shape (lref)
        The elements of `ref` are updated. The first (lmax-1)*k*k
        elements store the psi weights psi_1,psi_2,...,psi_{l_max-1}.
        The next k*lmax elements contain the forecasts of the
        transformed series and the next k*lmax elements contain the
        variances of the forecasts of the transformed variables; see
        ``multi_varma_forecast``. The last k elements are not updated.

    v : float, ndarray, shape (k, m)
        `v`[i-1,j-1] contains an estimate of the ith component of
        epsilon_{n+`mlast`+j}, for j = 1,2,...,m, for i = 1,2,...,k.

    predz : float, ndarray, shape (k, lmax)
        `predz`[i-1,j-1] contains the updated forecast of z_{i,n+j}, for
        j = `mlast`+m+1,...,l_max, for i = 1,2,...,k.

        The columns of `predz` corresponding to the new observations
        since the last call to either ``multi_varma_forecast`` or
        ``multi_varma_update`` are set equal to the corresponding
        columns of `z`.

    sefz : float, ndarray, shape (k, lmax)
        `sefz`[i-1,j-1] contains an estimate of the standard error of
        the corresponding element of `predz`, for j =
        `mlast`+m+1,...,l_max, for i = 1,2,...,k.

        The columns of `sefz` corresponding to the new observations
        since the last call to either ``multi_varma_forecast`` or
        ``multi_varma_update`` are set equal to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lref = *<value>* and the minimum size required =
            *<value>*.

            Constraint: lref >= (lmax-1)*k*k+2*k*lmax+k.

        (`errno` 1)
            On entry, `mlast` = *<value>*.

            Constraint: `mlast` >= 0.

        (`errno` 1)
            On entry, m = *<value>*, lmax = *<value>* and `mlast` =
            *<value>*.

            Constraint: m < lmax-`mlast`.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 1)
            On entry, lmax = *<value>*.

            Constraint: lmax >= 2.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, some of the elements of the array `ref` have been
            corrupted.

        (`errno` 3)
            On entry, one (or more) of the transformations requested is
            invalid.

        (`errno` 4)
            The updated forecasts will overflow if computed.

    Notes
    -----
    Let Z_t = (z_{1t},z_{2t},...,z_{kt})^T, for t = 1,2,...,n, denote a
    k-dimensional time series for which forecasts of
    Z^_{n+1},Z^_{n+2},...,Z^_{n+l_max} have been computed using
    ``multi_varma_forecast``.
    Given m further observations Z_{n+1},Z_{n+2},...,Z_{n+m}, where m <
    l_max, ``multi_varma_update`` updates the forecasts of
    Z_{n+m+1},Z_{n+m+2},...,Z_{n+l_max} and their corresponding standard
    errors.

    ``multi_varma_update`` uses a multivariate version of the procedure
    described in Box and Jenkins (1976).
    The forecasts are updated using the psi weights, computed in
    ``multi_varma_forecast``.
    If Z_t^* denotes the transformed value of Z_t and Z^_t^*(l) denotes
    the forecast of Z_t+l^* from time t with a lead of l (that is the
    forecast of Z_t+l^* given observations Z_t^*,Z_t-1^*,...), then

        Z^_t+1^*(l) =
        tau+psi_lepsilon_{t+1}+psi_{l+1}epsilon_t+psi_{l+2}epsilon_{t-1}+...

    and

        Z^_t^*(l+1) = tau+psi_{l+1}epsilon_t+psi_{l+2}epsilon_{t-1}+...

    where tau is a constant vector of length k involving the
    differencing parameters and the mean vector mu.
    By subtraction we obtain

        Z^_t+1^*(l) = Z^_t^*(l+1)+psi_lepsilon_{t+1}.

    Estimates of the residuals corresponding to the new observations are
    also computed as epsilon_{n+l} = Z_n+l^*-Z^_n^*(l), for l =
    1,2,...,m.
    These may be of use in checking that the new observations conform to
    the previously fitted model.

    On a successful exit, the reference array is updated so that
    ``multi_varma_update`` may be called again should future series
    values become available, see [Further Comments].

    When a transformation has been used the forecasts and their standard
    errors are suitably modified to give results in terms of the
    original series Z_t; see Granger and Newbold (1976).

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Granger, C W J and Newbold, P, 1976, `Forecasting transformed
    series`, J. Roy. Statist. Soc. Ser. B (38), 189--203

    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.tsa.multi_varma_update'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _predz = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'predz'},
        predz,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sefz = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sefz'},
        sefz,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_predz, 'predz'), (_sefz, 'sefz')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _lmax_py = _b_utils._EngineFloat64ArrayType.get_shape(_predz, {'fun_name': fun_name, 'entity': 'predz'}, exp_rank=2)[1]
    _lmax = _b_utils._EngineIntScalarType(
        _lmax_py, {'fun_name': fun_name, 'entity': '_lmax'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _ref = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ref'},
        ref,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lref_py = _b_utils._EngineFloat64ArrayType.get_shape(_ref, {'fun_name': fun_name, 'entity': 'ref'}, exp_rank=1)[0]
    _lref = _b_utils._EngineIntScalarType(
        _lref_py, {'fun_name': fun_name, 'entity': '_lref'},
    ).to_ctypes()
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _m_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_tsa.multi_varma_update(
        _k,
        _lmax,
        _m,
        mlast,
        _z,
        _ref,
        _lref,
        _v,
        _predz,
        _sefz,
    )
    _mlast_py = _return_scalar
    _ref_py = _ref
    _predz_py = _predz
    _sefz_py = _sefz
    _v_py = _v
    _return_data = MultiVarmaUpdateReturnData(
        _mlast_py,
        _ref_py,
        _v_py,
        _predz_py,
        _sefz_py,
    )
    return _return_data

MultiDiffReturnData = _collections.namedtuple(
    'MultiDiffReturnData',
    ['w', 'nd'],
)

def multi_diff(z, tr, dord, delta):
    u"""
    Multivariate time series, differences and/or transforms.

    ``multi_diff`` differences and/or transforms a multivariate time
    series. It is intended to be used prior to ``multi_varma_estimate``
    to fit a vector autoregressive moving average (VARMA) model to the
    differenced/transformed series.

    For full information please refer to the NAG Library document for
    g13dl

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dlf.html

    Parameters
    ----------
    z : float, array-like, shape (k, n)
        `z`[i,t] must contain, z_{it}, the ith component of Z_t, for t =
        1,2,...,n, for i = 1,2,...,k.

    tr : str, length 1, array-like, shape (k)
        `tr`[i-1] indicates whether the ith time series is to be
        transformed, for i = 1,2,...,k.

        `tr`[i-1] = 'N'

            No transformation is used.

        `tr`[i-1] = 'L'

            A log transformation is used.

        `tr`[i-1] = 'S'

            A square root transformation is used.

    dord : int, array-like, shape (k)
        The order of differencing for each series, d_1,d_2,...,d_k.

    delta : float, array-like, shape (k, d)
        If `dord`[i-1] > 0, then `delta`[i-1,j-1] must be set equal to
        delta_{ij}, for i = 1,2,...,k, for j = 1,2,...,d_i.

        If d = 0, `delta` is not referenced.

    Returns
    -------
    w : float, ndarray, shape (k, `nd`)
        `w`[i-1,t-1] contains the value of w_{i,t+d}, for t =
        1,2,...,n-d, for i = 1,2,...,k.

    nd : int
        The number of differenced values, n-d, in the series, where d =
        max(`dord`[i-1]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `dord`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: `dord` < n.

        (`errno` 2)
            On entry, i = *<value>* and `dord`[i-1] = *<value>*.

            Constraint: `dord`[i-1] >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `tr`[i-1] is invalid.

            Constraint: `tr`[i-1] = 'N', 'L' or 'S'.

        (`errno` 4)
            On entry, one (or more) of the transformations requested is
            invalid.

    Notes
    -----
    For certain time series it may first be necessary to difference the
    original data to obtain a stationary series before calculating
    autocorrelations, etc.
    This function also allows you to apply either a square root or a log
    transformation to the original time series to stabilize the variance
    if required.

    If the order of differencing required for the ith series is d_i,
    then the differencing operator is defined by delta_i(B) =
    1-delta_{i1}B-delta_{i2}B^2-...-delta_{id_i}B^(d_i), where B is the
    backward shift operator; that is, BZ_t = Z_{t-1}.
    Let d denote the maximum of the orders of differencing, d_i, over
    the k series.
    The function computes values of the differenced/transformed series
    W_t = (w_{1t},w_{2t},...,w_{kt})^T, for t = d+1,...,n, as follows:

        w_{it} = delta_i(B)z_it^*, i = 1,2,...,k

    where z_it^* are the transformed values of the original
    k-dimensional time series Z_t = (z_{1t},z_{2t},...,z_{kt})^T.

    The differencing parameters delta_{ij}, for i = 1,2,...,k and j =
    1,2,...,d_i, must be supplied by you.
    If the ith series does not require differencing, then d_i = 0.

    References
    ----------
    Box, G E P and Jenkins, G M, 1976, `Time Series Analysis:
    Forecasting and Control`, (Revised Edition), Holden--Day

    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.tsa.multi_diff'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _delta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'delta'},
        delta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_delta, 'delta')),
    )
    _tr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tr'},
        tr,
        copy=False,
    )
    _k_py = _b_utils._EngineCharArrayType.get_shape(_tr, {'fun_name': fun_name, 'entity': 'tr'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _dord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dord'},
        dord,
        copy=False,
    )
    _vl_d_py = max(_dord[0:_k_py-1+1])
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _n_py-_vl_d_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_tsa.multi_diff(
        _k,
        _n,
        _z,
        _tr,
        _dord,
        _delta,
        _w,
    )
    _nd_py = _return_scalar
    _w_py = _w[:, :_nd_py]
    _return_data = MultiDiffReturnData(
        _w_py,
        _nd_py,
    )
    return _return_data

MultiCorrmatCrossReturnData = _collections.namedtuple(
    'MultiCorrmatCrossReturnData',
    ['wmean', 'r0', 'r'],
)

def multi_corrmat_cross(matrix, m, w):
    u"""
    Multivariate time series, sample cross-correlation or
    cross-covariance matrices.

    ``multi_corrmat_cross`` calculates the sample cross-correlation (or
    cross-covariance) matrices of a multivariate time series.

    For full information please refer to the NAG Library document for
    g13dm

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dmf.html

    Parameters
    ----------
    matrix : str, length 1
        Indicates whether the cross-covariance or cross-correlation
        matrices are to be computed.

        `matrix` = 'V'

            The cross-covariance matrices are computed.

        `matrix` = 'R'

            The cross-correlation matrices are computed.

    m : int
        m, the number of cross-correlation (or cross-covariance)
        matrices to be computed. If in doubt set `m` = 10. However it
        should be noted that `m` is usually taken to be at most n/4.

    w : float, array-like, shape (k, n)
        `w`[i-1,t-1] must contain the observation w_{it}, for t =
        1,2,...,n, for i = 1,2,...,k.

    Returns
    -------
    wmean : float, ndarray, shape (k)
        The means, w__i, for i = 1,2,...,k.

    r0 : float, ndarray, shape (k, k)
        If i != j, then `r0`[i-1,j-1] contains an estimate of the
        (i,j)th element of the cross-correlation (or cross-covariance)
        matrix at lag zero, R^_{ij}(0); if i = j, then if `matrix` =
        'V', `r0`[i-1,i-1] contains the variance of the ith series,
        C^_{ii}(0), and if `matrix` = 'R', `r0`[i-1,i-1] contains the
        standard deviation of the ith series, sqrt(C^_{ii}(0)).

        If `errno` = 2 and `matrix` = 'R', then on exit all the elements
        in `r0` whose computation involves the zero variance are set to
        zero.

    r : float, ndarray, shape (k, k, `m`)
        `r`[i-1,j-1,l-1] contains an estimate of the (i,j)th element of
        the cross-correlation (or cross-covariance) at lag l,
        R^_{ij}(l), for j = 1,2,...,k, for i = 1,2,...,k, for l =
        1,2,...,m.

        If `errno` = 2 and `matrix` = 'R', then on exit all the elements
        in `r` whose computation involves the zero variance are set to
        zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= 1 and `m` < n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 1)
            On entry, `matrix` = *<value>*.

            Constraint: `matrix` = 'V' or 'R'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, at least one of the series is such that all its
            elements are practically identical giving zero (or near
            zero) variance.

    Notes
    -----
    Let W_t = (w_{1t},w_{2t},...,w_{kt})^T, for t = 1,2,...,n, denote n
    observations of a vector of k time series.
    The sample cross-covariance matrix at lag l is defined to be the k*k
    matrix C^(l), whose (i,j)th element is given by

        C^_{ij}(l) = 1/nsum_t = l+1^n(w_{i(t-l)}-w__i)(w_{jt}-w__j), l =
        0,1,2,...,m, i = 1,2,...,k and j = 1,2,...,k,

    where w__i and w__j denote the sample means for the ith and jth
    series respectively.
    The sample cross-correlation matrix at lag l is defined to be the
    k*k matrix R^(l), whose (i,j)th element is given by

        R^_{ij}(l) = C^_{ij}(l)/sqrt(C^_{ii}(0)C^_{jj}(0)), l =
        0,1,2,...,m, i = 1,2,...,k and j = 1,2,...,k.

    The number of lags, m, is usually taken to be at most n/4.

    If W_t follows a vector moving average model of order q, then it can
    be shown that the theoretical cross-correlation matrices (R(l)) are
    zero beyond lag q.
    In order to help spot a possible cut-off point, the elements of
    R^(l) are usually compared to their approximate standard error of
    1/sqrt(n).
    For further details see, for example, Wei (1990).

    The function uses a single pass through the data to compute the
    means and the cross-covariance matrix at lag zero.
    The cross-covariance matrices at further lags are then computed on a
    second pass through the data.

    References
    ----------
    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    fun_name = 'naginterfaces.library.tsa.multi_corrmat_cross'

    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_w, 'w'),),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _wmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _r0 = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py),
        sorder=_m_sorder,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    try:
        _b_tsa.multi_corrmat_cross(
            matrix,
            _k,
            _n,
            _m,
            _w,
            _wmean,
            _r0,
            _r,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _wmean_py = _wmean
    _r0_py = _r0
    _r_py = _r
    _return_data = MultiCorrmatCrossReturnData(
        _wmean_py,
        _r0_py,
        _r_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiCorrmatPartlagReturnData = _collections.namedtuple(
    'MultiCorrmatPartlagReturnData',
    ['maxlag', 'parlag', 'x', 'pvalue'],
)

def multi_corrmat_partlag(n, r0, r):
    u"""
    Multivariate time series, sample partial lag correlation matrices,
    chi^2 statistics and significance levels.

    ``multi_corrmat_partlag`` calculates the sample partial lag
    correlation matrices of a multivariate time series.
    A set of chi^2-statistics and their significance levels are also
    returned.
    A call to ``multi_corrmat_cross`` is usually made prior to calling
    this function in order to calculate the sample cross-correlation
    matrices.

    For full information please refer to the NAG Library document for
    g13dn

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dnf.html

    Parameters
    ----------
    n : int
        n, the number of observations in each series.

    r0 : float, array-like, shape (k, k)
        If i != j, then `r0`[i-1,j-1] must contain the (i,j)th element
        of the sample cross-correlation matrix at lag zero, R^_{ij}(0).
        If i = j, then `r0`[i-1,i-1] must contain the standard deviation
        of the ith series.

    r : float, array-like, shape (k, k, m)
        `r`[i-1,j-1,l-1] must contain the (i,j)th element of the sample
        cross-correlation at lag l, R^_{ij}(l), for j = 1,2,...,k, for i
        = 1,2,...,k, for l = 1,2,...,m, where series j leads series i
        (see [Further Comments]).

    Returns
    -------
    maxlag : int
        The maximum lag up to which partial lag correlation matrices
        (along with chi^2-statistics and their significance levels) have
        been successfully computed. On a successful exit `maxlag` will
        equal m. If `errno` = 2 on exit, `maxlag` will be less than m.

    parlag : float, ndarray, shape (k, k, m)
        `parlag`[i-1,j-1,l-1] contains the (i,j)th element of the sample
        partial lag correlation matrix at lag l, P^_{ij}(l), for j =
        1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,`maxlag`.

    x : float, ndarray, shape (m)
        `x`[l-1] contains the chi^2-statistic at lag l, for l =
        1,2,...,`maxlag`.

    pvalue : float, ndarray, shape (m)
        `pvalue`[l-1] contains the significance level of the
        corresponding chi^2-statistic in `x`, for l = 1,2,...,`maxlag`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>* and `n` = *<value>*.

            Constraint: m < `n`.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The recursive equations used to compute the partial lag
            correlation matrices are ill-conditioned (they have been
            computed up to lag *<value>*).

    Notes
    -----
    Let W_t = (w_{1t},w_{2t},...,w_{kt})^T, for t = 1,2,...,n, denote n
    observations of a vector of k time series.
    The partial lag correlation matrix at lag l, P(l), is defined to be
    the correlation matrix between W_t and W_{t+l}, after removing the
    linear dependence on each of the intervening vectors
    W_{t+1},W_{t+2},...,W_{t+l-1}.
    It is the correlation matrix between the residual vectors resulting
    from the regression of W_{t+l} on the carriers W_{t+l-1},...,W_{t+1}
    and the regression of W_t on the same set of carriers; see Heyse and
    Wei (1985).

    P(l) has the following properties.

    (i) If W_t follows a vector autoregressive model of order p, then
        P(l) = 0 for l > p;

    (#) When k = 1, P(l) reduces to the univariate partial
        autocorrelation at lag l;

    (#) Each element of P(l) is a properly normalized correlation
        coefficient;

    (#) When l = 1, P(l) is equal to the cross-correlation matrix at lag
        1 (a natural property which also holds for the univariate
        partial autocorrelation function).

    Sample estimates of the partial lag correlation matrices may be
    obtained using the recursive algorithm described in Wei (1990).
    They are calculated up to lag m, which is usually taken to be at
    most n/4.
    Only the sample cross-correlation matrices ( R^(l), for l =
    0,1,...,m) and the standard deviations of the series are required as
    input to ``multi_corrmat_partlag``.
    These may be computed by ``multi_corrmat_cross``.
    Under the hypothesis that W_t follows an autoregressive model of
    order s-1, the elements of the sample partial lag matrix P^(s),
    denoted by P^_{ij}(s), are asymptotically Normally distributed with
    mean zero and variance 1/n.
    In addition the statistic

        X(s) = nsum_i = 1^ksum_j = 1^kP^_{ij}(s)^2

    has an asymptotic chi^2-distribution with k^2 degrees of freedom.
    These quantities, X(l), are useful as a diagnostic aid for
    determining whether the series follows an autoregressive model and,
    if so, of what order.

    References
    ----------
    Heyse, J F and Wei, W W S, 1985, `The partial lag autocorrelation
    function`, Technical Report No. 32, Department of Statistics, Temple
    University, Philadelphia

    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.tsa.multi_corrmat_partlag'

    _r0 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r0'},
        r0,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_r0, 'r0'), (_r, 'r')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_r0, {'fun_name': fun_name, 'entity': 'r0'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=3)[2]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _parlag = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _pvalue = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _return_scalar = _b_tsa.multi_corrmat_partlag(
            _k,
            n,
            _m,
            _r0,
            _r,
            _parlag,
            _x,
            _pvalue,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _maxlag_py = _return_scalar
    _parlag_py = _parlag
    _x_py = _x
    _pvalue_py = _pvalue
    _return_data = MultiCorrmatPartlagReturnData(
        _maxlag_py,
        _parlag_py,
        _x_py,
        _pvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiRegmatPartialReturnData = _collections.namedtuple(
    'MultiRegmatPartialReturnData',
    ['maxlag', 'parlag', 'se', 'qq', 'x', 'pvalue', 'loglhd'],
)

def multi_regmat_partial(z, m):
    u"""
    Multivariate time series, partial autoregression matrices.

    ``multi_regmat_partial`` calculates the sample partial
    autoregression matrices of a multivariate time series.
    A set of likelihood ratio statistics and their significance levels
    are also returned.
    These quantities are useful for determining whether the series
    follows an autoregressive model and, if so, of what order.

    For full information please refer to the NAG Library document for
    g13dp

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dpf.html

    Parameters
    ----------
    z : float, array-like, shape (k, n)
        `z`[i-1,t-1] must contain the observation w_{it}, for t =
        1,2,...,n, for i = 1,2,...,k.

    m : int
        m, the number of partial autoregression matrices to be computed.
        If in doubt set `m` = 10.

    Returns
    -------
    maxlag : int
        The maximum lag up to which partial autoregression matrices
        (along with their likelihood ratio statistics and their
        significance levels) have been successfully computed. On a
        successful exit `maxlag` will equal `m`. If `errno` = 2 on exit
        then `maxlag` will be less than `m`.

    parlag : float, ndarray, shape (k, k, `m`)
        `parlag`[i-1,j-1,l-1] contains an estimate of the (i,j)th
        element of the partial autoregression matrix at lag l, P^_l(ij),
        for l = 1,2,...,`maxlag`, i = 1,2,...,k and j = 1,2,...,k.

    se : float, ndarray, shape (k, k, `m`)
        `se`[i-1,j-1,l-1] contains an estimate of the standard error of
        the corresponding element in the array `parlag`.

    qq : float, ndarray, shape (k, k, `m`)
        `qq`[i-1,j-1,l-1] contains an estimate of the (i,j)th element of
        the corresponding variance-covariance matrix Sigma^_l, for j =
        1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,`maxlag`.

    x : float, ndarray, shape (`m`)
        `x`[l-1] contains X_l, the likelihood ratio statistic at lag l,
        for l = 1,2,...,`maxlag`.

    pvalue : float, ndarray, shape (`m`)
        `pvalue`[l-1] contains the significance level of the statistic
        in the corresponding element of `x`.

    loglhd : float, ndarray, shape (`m`)
        `loglhd`[l-1] contains an estimate of the maximum of the
        log-likelihood function when an AR(l-1) model has been fitted to
        the series, for l = 1,2,...,`maxlag`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*, `m` = *<value>* and n = *<value>*.

            Constraint: n-`m`-(k*`m`+1) >= k.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 4.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The recursive equations used to compute the partial
            autoregression matrices are ill-conditioned. They have been
            computed up to lag *<value>*.

    Notes
    -----
    Let W_t = (w_{1t},w_{2t},...,w_{kt})^T, for t = 1,2,...,n, denote a
    vector of k time series.
    The partial autoregression matrix at lag l, P_l, is defined to be
    the last matrix coefficient when a vector autoregressive model of
    order l is fitted to the series. P_l has the property that if W_t
    follows a vector autoregressive model of order p then P_l = 0 for l
    > p.

    Sample estimates of the partial autoregression matrices may be
    obtained by fitting autoregressive models of successively higher
    orders by multivariate least squares; see Tiao and Box (1981) and
    Wei (1990).
    These models are fitted using a QR algorithm based on the functions
    ``correg.linregm_obs_edit`` and ``correg.linregm_var_del``.
    They are calculated up to lag m, which is usually taken to be at
    most n/4.

    The function also returns the asymptotic standard errors of the
    elements of P^_l and an estimate of the residual variance-covariance
    matrix Sigma^_l, for l = 1,2,...,m.
    If S_l denotes the residual sum of squares and cross-products matrix
    after fitting an AR(l) model to the series then under the null
    hypothesis H_0:P_l = 0 the test statistic

        X_l = -((n-m-1)-1/2-lk)log(|S_l|/|S_{l-1}|)

    is asymptotically distributed as chi^2 with k^2 degrees of freedom.
    X_l provides a useful diagnostic aid in determining the order of an
    autoregressive model. (Note that Sigma^_l = S_l/(n-l).) The function
    also returns an estimate of the maximum of the log-likelihood
    function for each AR model that has been fitted.

    References
    ----------
    Tiao, G C and Box, G E P, 1981, `Modelling multiple time series with
    applications`, J. Am. Stat. Assoc. (76), 802--816

    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.tsa.multi_regmat_partial'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'),),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _parlag = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _qq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _pvalue = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _loglhd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _return_scalar = _b_tsa.multi_regmat_partial(
            _k,
            _n,
            _z,
            _m,
            _parlag,
            _se,
            _qq,
            _x,
            _pvalue,
            _loglhd,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _maxlag_py = _return_scalar
    _parlag_py = _parlag
    _se_py = _se
    _qq_py = _qq
    _x_py = _x
    _pvalue_py = _pvalue
    _loglhd_py = _loglhd
    _return_data = MultiRegmatPartialReturnData(
        _maxlag_py,
        _parlag_py,
        _se_py,
        _qq_py,
        _x_py,
        _pvalue_py,
        _loglhd_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MultiVarmaDiagReturnData = _collections.namedtuple(
    'MultiVarmaDiagReturnData',
    ['qq', 'r0', 'r', 'rcm', 'chi', 'idf', 'siglev'],
)

def multi_varma_diag(v, ip, iq, m, par, parhld, qq, ishow, io_manager=None):
    u"""
    Multivariate time series, diagnostic checking of residuals,
    following ``multi_varma_estimate``.

    ``multi_varma_diag`` is a diagnostic checking function suitable for
    use after fitting a vector ARMA model to a multivariate time series
    using ``multi_varma_estimate``.
    The residual cross-correlation matrices are returned along with an
    estimate of their asymptotic standard errors and correlations.
    Also, ``multi_varma_diag`` calculates the modified Li--McLeod
    portmanteau statistic and its significance level for testing model
    adequacy.

    For full information please refer to the NAG Library document for
    g13ds

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dsf.html

    Parameters
    ----------
    v : float, array-like, shape (k, n)
        `v`[i-1,t-1] must contain an estimate of the ith component of
        epsilon_t, for t = 1,2,...,n, for i = 1,2,...,k.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    m : int
        The value of m, the number of residual cross-correlation
        matrices to be computed. See [Further Comments] for advice on
        the choice of `m`.

    par : float, array-like, shape ((`ip`+`iq`)*k*k)
        The parameter estimates read in row by row in the order
        phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q.

        Thus,

            if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal
            to an estimate of the (i,j)th element of phi_l, for i =
            1,2,...,k, for l = 1,2,...,p;

            if `iq` >= 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
            equal to an estimate of the (i,j)th element of theta_l, for
            i = 1,2,...,k, for l = 1,2,...,q.

        The first p*k*k elements of `par` must satisfy the stationarity
        condition and the next q*k*k elements of `par` must satisfy the
        invertibility condition.

    parhld : bool, array-like, shape ((`ip`+`iq`)*k*k)
        `parhld`[i-1] must be set to True if `par`[i-1] has been held
        constant at a pre-specified value and False if `par`[i-1] is a
        free parameter, for i = 1,2,...,(p+q)*k*k.

    qq : float, array-like, shape (k, k)
        `qq`[i-1,j-1] is an efficient estimate of the (i,j)th element of
        Sigma. The lower triangle only is needed.

    ishow : int
        Must be nonzero if the residual cross-correlation matrices
        {r^_{ij}(l)} and their standard errors {se(r^_{ij}(l))}, the
        modified portmanteau statistic with its significance and a
        summary table are to be printed. The summary table indicates
        which elements of the residual correlation matrices are
        significant at the 5% level in either a positive or negative
        direction; i.e., if r^_{ij}(l) > 1.96*se(r^_{ij}(l)) then a '+'
        is printed, if r^_{ij}(l) < -1.96*se(r^_{ij}(l)) then a '-' is
        printed, otherwise a fullstop (.) is printed. The summary table
        is only printed if k <= 6 on entry.

        The residual cross-correlation matrices, their standard errors
        and the modified portmanteau statistic with its significance are
        available also as output variables in `r`, `rcm`, `chi`, `idf`
        and `siglev`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    qq : float, ndarray, shape (k, k)
        If `errno` != 1, then the upper triangle is set equal to the
        lower triangle.

    r0 : float, ndarray, shape (k, k)
        If i != j, then `r0`[i-1,j-1] contains an estimate of the
        (i,j)th element of the residual cross-correlation matrix at lag
        zero, R^_0. When i = j, `r0`[i-1,j-1] contains the standard
        deviation of the ith residual series. If `errno` = 3 on exit
        then the first k rows and columns of `r0` are set to zero.

    r : float, ndarray, shape (k, k, `m`)
        `r`[l-1,i-1,j-1] is an estimate of the (i,j)th element of the
        residual cross-correlation matrix at lag l, for l = 1,2,...,m,
        for j = 1,2,...,k, for i = 1,2,...,k. If `errno` = 3 on exit
        then all elements of `r` are set to zero.

    rcm : float, ndarray, shape (`m`*k*k, `m`*k*k)
        The estimated standard errors and correlations of the elements
        in the array `r`. The correlation between `r`[l-1,i-1,j-1] and
        `r`[l_2-1,i_2-1,j_2-1] is returned as `rcm`[s,t] where s =
        (l-1)*k*k+(j-1)*k+i and t = (l_2-1)*k*k+(j_2-1)*k+i_2 except
        that if s = t, then `rcm`[s,t] contains the standard error of
        `r`[l-1,i-1,j-1]. If on exit, `errno` >= 5, then all
        off-diagonal elements of `rcm` are set to zero and all diagonal
        elements are set to 1/sqrt(n).

    chi : float
        The value of the modified portmanteau statistic, Q_(m)^*. If
        `errno` = 3 on exit then `chi` is returned as zero.

    idf : int
        The number of degrees of freedom of `chi`.

    siglev : float
        The significance level of `chi` based on `idf` degrees of
        freedom. If `errno` = 3 on exit, `siglev` is returned as one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ldrcm = *<value>*, `m` = *<value>* and k =
            *<value>*.

            Constraint: ldrcm >= `m`*k*k.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` < n.

        (`errno` 1)
            On entry, `m` = *<value>*, `ip` = *<value>* and `iq` =
            *<value>*.

            Constraint: `m` > `ip`+`iq`.

        (`errno` 1)
            On entry, `ip` = 0 and `iq` = 0.

            Constraint: `ip` = `iq` = 0 must not hold.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, the MA parameter matrices are outside the
            invertibility region.

        (`errno` 2)
            On entry, the AR parameter matrices are outside the
            stationarity region.

        (`errno` 2)
            On entry, the covariance matrix `qq` is not positive
            definite.

        (`errno` 4)
            Excessive iterations needed to find zeros of determinental
            polynomials.

        (`errno` 5)
            On entry, the AR operator has a factor in common with the MA
            operator.

        (`errno` 6)
            The matrix `rcm` could not be computed because one of its
            diagonal elements was found to be non-positive.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one of the residual series in the array
            `v` has near-zero variance.

        (`errno` 3)
            On entry, at least two of the residual series are identical.

    Notes
    -----
    Let W_t = (w_{1t},w_{2t},...,w_{kt})^T, for t = 1,2,...,n, denote a
    vector of k time series which is assumed to follow a multivariate
    ARMA model of the form

        [table omitted]

    where epsilon_t = (epsilon_{1t},epsilon_{2t},...,epsilon_{kt})^T,
    for t = 1,2,...,n, is a vector of k residual series assumed to be
    Normally distributed with zero mean and positive definite covariance
    matrix Sigma.
    The components of epsilon_t are assumed to be uncorrelated at
    non-simultaneous lags.
    The phi_i and theta_j are k*k matrices of parameters. {phi_i}, for i
    = 1,2,...,p, are called the autoregressive (AR) parameter matrices,
    and {theta_i}, for i = 1,2,...,q, the moving average (MA) parameter
    matrices.
    The parameters in the model are thus the p (k*k) phi-matrices, the q
    (k*k) theta-matrices, the mean vector mu and the residual error
    covariance matrix Sigma.
    Let

        A(phi) = [[table omitted]]_{pk*pk} and B(theta) = [[table
        omitted]]_{qk*qk}

    where I denotes the k*k identity matrix.

    The ARMA model (1) is said to be stationary if the eigenvalues of
    A(phi) lie inside the unit circle, and invertible if the eigenvalues
    of B(theta) lie inside the unit circle.
    The ARMA model is assumed to be both stationary and invertible.
    Note that some of the elements of the phi- and/or theta-matrices may
    have been fixed at pre-specified values (for example by calling
    ``multi_varma_estimate``).

    The estimated residual cross-correlation matrix at lag l is defined
    to the k*k matrix R^_l whose (i,j)th element is computed as

        r^_{ij}(l) = sum_t = l+1^n(epsilon^_{it-l} -
        epsilon__i)(epsilon^_{jt} - epsilon__j)/(sqrt(sum_t =
        1^n(epsilon^_{it} - epsilon__i)^2sum_t = 1^n(epsilon^_{jt} -
        epsilon__j)^2)), l = 0,1,...,i and j = 1,2,...,k,

    where epsilon^_{it} denotes an estimate of the tth residual for the
    ith series epsilon_{it} and epsilon__i = sum_t = 1^n
    epsilon^_{it}/n. (Note that R^_l is an estimate of
    E(epsilon_{t-l}epsilon_t^T), where E is the expected value.)

    A modified portmanteau statistic, Q_(m)^*, is calculated from the
    formula (see Li and McLeod (1981))

        Q_(m)^* = k^2m(m+1)/2n+nsum_l =
        1^mr^(l)^T(R^_0^-1⊗R^_0^-1)r^(l),

    where ⊗ denotes Kronecker product, R^_0 is the estimated residual
    cross-correlation matrix at lag zero and r^(l) = vec(R^_l^T), where
    vec of a k*k matrix is a vector with the (i,j)th element in position
    (i-1)k+j. m denotes the number of residual cross-correlation
    matrices computed. (Advice on the choice of m is given in [Further
    Comments].) Let l_C denote the total number of 'free' parameters in
    the ARMA model excluding the mean, mu, and the residual error
    covariance matrix Sigma.
    Then, under the hypothesis of model adequacy, Q_(m)^*, has an
    asymptotic chi^2-distribution on mk^2-l_C degrees of freedom.

    Let _r^_= (vec(R_1^T),vec(R_2^T),...,vec(R_m^T)) then the covariance
    matrix of _r^_ is given by

        Var(_r_^) = [Y-X(X^TGG^TX)^-1X^T]/n,

    where Y = I_m⊗(Delta⊗Delta) and G = I_m(GG^T). Delta is the
    dispersion matrix Sigma in correlation form and G a nonsingular k*k
    matrix such that GG^T = Delta^-1 and G Delta G^T = I_k.
    The construction of the matrix X is discussed in Li and McLeod
    (1981). (Note that the mean, mu, plays no part in calculating
    Var(_r_^) and, therefore, is not required as input to
    ``multi_varma_diag``.)

    References
    ----------
    Li, W K and McLeod, A I, 1981, `Distribution of the residual
    autocorrelations in multivariate ARMA time series models`, J. Roy.
    Statist. Soc. Ser. B (43), 231--239
    """

    fun_name = 'naginterfaces.library.tsa.multi_varma_diag'

    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _qq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qq'},
        qq,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_v, 'v'), (_qq, 'qq')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _parhld = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parhld'},
        parhld,
        copy=False,
    )
    _r0 = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py),
        sorder=_m_sorder,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _rcm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py*_k_py*_k_py, _m_py*_k_py*_k_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = _b_tsa.multi_varma_diag(
            _k,
            _n,
            _v,
            ip,
            iq,
            _m,
            _par,
            _parhld,
            _qq,
            ishow,
            _r0,
            _r,
            _rcm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _qq_py = _qq
    _r0_py = _r0
    _r_py = _r
    _rcm_py = _rcm
    _chi_py = _return_scalars.chi
    _idf_py = _return_scalars.idf
    _siglev_py = _return_scalars.siglev
    _return_data = MultiVarmaDiagReturnData(
        _qq_py,
        _r0_py,
        _r_py,
        _rcm_py,
        _chi_py,
        _idf_py,
        _siglev_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniArmaRootsReturnData = _collections.namedtuple(
    'UniArmaRootsReturnData',
    ['rr', 'ri', 'rmod'],
)

def uni_arma_roots(k, ip, par):
    u"""
    Calculates the zeros of a vector autoregressive (or moving average)
    operator.

    ``uni_arma_roots`` calculates the zeros of a vector autoregressive
    (or moving average) operator. This function is likely to be used in
    conjunction with ``rand.times_mv_varma``, ``uni_arima_resid``,
    ``multi_varma_estimate`` or ``multi_varma_diag``.

    For full information please refer to the NAG Library document for
    g13dx

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13dxf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    ip : int
        The number of AR (or MA) parameter matrices, p (or q).

    par : float, array-like, shape (`ip`*`k`*`k`)
        The AR (or MA) parameter matrices read in row by row in the
        order phi_1,phi_2,...,phi_p (or theta_1,theta_2,...,theta_q).
        That is, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to the
        (i,j)th element of phi_l, for l = 1,2,...,p (or the (i,j)th
        element of theta_l, for l = 1,2,...,q).

    Returns
    -------
    rr : float, ndarray, shape (`k`*`ip`)
        The real parts of the eigenvalues.

    ri : float, ndarray, shape (`k`*`ip`)
        The imaginary parts of the eigenvalues.

    rmod : float, ndarray, shape (`k`*`ip`)
        The moduli of the eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            An excessive number of iterations have been required to
            calculate the eigenvalues.

    Notes
    -----
    Consider the vector autoregressive moving average (VARMA) model

        W_t-mu =
        phi_1(W_{t-1}-mu)+phi_2(W_{t-2}-mu)+...+phi_p(W_{t-p}-mu)+epsilon_t-theta_1epsilon_{t-1}-theta_2epsilon_{t-2}-...-theta_qepsilon_{t-q},

    where W_t denotes a vector of k time series and epsilon_t is a
    vector of k residual series having zero mean and a constant
    variance-covariance matrix.
    The components of epsilon_t are also assumed to be uncorrelated at
    non-simultaneous lags. phi_1,phi_2,...,phi_p denotes a sequence of
    k*k matrices of autoregressive (AR) parameters and
    theta_1,theta_2,...,theta_q denotes a sequence of k*k matrices of
    moving average (MA) parameters. mu is a vector of length k
    containing the series means.
    Let

        A(phi) = [[table omitted]]_{pk*pk}

    where I denotes the k*k identity matrix.

    The model (1) is said to be stationary if the eigenvalues of A(phi)
    lie inside the unit circle.
    Similarly let

        B(theta) = [[table omitted]]_{qk*qk}.

    Then the model is said to be invertible if the eigenvalues of
    B(theta) lie inside the unit circle.

    ``uni_arma_roots`` returns the pk eigenvalues of A(phi) (or the qk
    eigenvalues of B(theta)) along with their moduli, in descending
    order of magnitude.
    Thus to check for stationarity or invertibility you should check
    whether the modulus of the largest eigenvalue is less than 1.

    References
    ----------
    Wei, W W S, 1990, `Time Series Analysis: Univariate and Multivariate
    Methods`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.tsa.uni_arma_roots'

    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k = _b_utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    _ip = _b_utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ip_py = _b_utils._EngineIntScalarType.to_py(_ip)
    _rr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py*_ip_py,),
    )
    _ri = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py*_ip_py,),
    )
    _rmod = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py*_ip_py,),
    )
    _b_tsa.uni_arma_roots(
        _k,
        _ip,
        _par,
        _rr,
        _ri,
        _rmod,
    )
    _rr_py = _rr
    _ri_py = _ri
    _rmod_py = _rmod
    _return_data = UniArmaRootsReturnData(
        _rr_py,
        _ri_py,
        _rmod_py,
    )
    return _return_data

MultiKalmanSqrtVarReturnData = _collections.namedtuple(
    'MultiKalmanSqrtVarReturnData',
    ['s', 'k', 'h'],
)

def multi_kalman_sqrt_var(a, b, stq, q, c, r, s, tol=0.0):
    u"""
    Combined measurement and time update, one iteration of Kalman
    filter, time-varying, square root covariance filter.

    ``multi_kalman_sqrt_var`` performs a combined measurement and time
    update of one iteration of the time-varying Kalman filter using a
    square root covariance filter.

    For full information please refer to the NAG Library document for
    g13ea

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13eaf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The state transition matrix, A_i.

    b : float, array-like, shape (n, l)
        The noise coefficient matrix B_i.

    stq : bool
        If `stq` = True, the state noise covariance matrix Q_i is
        assumed to be the identity matrix. Otherwise the lower
        triangular Cholesky factor, Q_i^1/2, must be provided in `q`.

    q : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`stq`): l; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if not((`stq`): l; otherwise: 0.

        If `stq` = False, `q` must contain the lower triangular Cholesky
        factor of the state noise covariance matrix, Q_i^1/2. Otherwise
        `q` is not referenced.

    c : float, array-like, shape (m, n)
        The measurement coefficient matrix, C_i.

    r : float, array-like, shape (m, m)
        The lower triangular Cholesky factor of the measurement noise
        covariance matrix R_i^1/2.

    s : float, array-like, shape (n, n)
        The lower triangular Cholesky factor of the state covariance
        matrix, S_i.

    tol : float, optional
        The tolerance used to test for the singularity of H_i^1/2. If
        0.0 <= `tol` < m^2*machine precision, then m^2*machine precision
        is used instead. The inverse of the condition number of H^(1/2)
        is estimated by a call to ``lapacklin.dtrcon``. If this estimate
        is less than `tol` then H^(1/2) is assumed to be singular.

    Returns
    -------
    s : float, ndarray, shape (n, n)
        The lower triangular Cholesky factor of the state covariance
        matrix, S_{i+1}.

    k : float, ndarray, shape (n, m)
        The Kalman gain matrix, K_i, premultiplied by the state
        transition matrix, A_i, A_i K_i.

    h : float, ndarray, shape (m, m)
        The lower triangular matrix H_i^1/2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, l = *<value>*.

            Constraint: l >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The matrix H_i^1/2 is singular.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The Kalman filter arises from the state space model given by:

        [table omitted]

    where X_i is the state vector of length n at time i, Y_i is the
    observation vector of length m at time i, and W_i of length l and
    V_i of length m are the independent state noise and measurement
    noise respectively.

    The estimate of X_i given observations Y_1 to Y_{i-1} is denoted by
    X^_{i | i-1} with state covariance matrix Var(X^_{i | i-1}) = P_{i |
    i-1} = S_i S_i^T, while the estimate of X_i given observations Y_1
    to Y_i is denoted by X^_{i | i} with covariance matrix Var(X^_{i |
    i}) = P_{i | i}.
    The update of the estimate, X^_{i | i-1}, from time i to time (i+1),
    is computed in two stages.
    First, the measurement-update is given by

        X^_{i | i} = X^_{i | i-1}+K_i[Y_i-C_i X^_{i | i-1}]

    and

        P_{i | i} = [I-K_i C_i]P_{i | i-1}

    where K_i = P_{i | i-1}C_i^T[C_i P_{i | i-1}C_i^T+R_i]^-1 is the
    Kalman gain matrix.
    The second stage is the time-update for X which is given by

        X^_{i+1 | i} = A_i X^_{i | i}+D_i U_i

    and

        P_{i+1 | i} = A_i P_{i | i}A_i^T+B_i Q_i B_i^T

    where D_i U_i represents any deterministic control used.

    The square root covariance filter algorithm provides a stable method
    for computing the Kalman gain matrix and the state covariance
    matrix.
    The algorithm can be summarised as

        [table omitted]U = [table omitted]

    where U is an orthogonal transformation triangularizing the
    left-hand pre-array to produce the right-hand post-array.
    The relationship between the Kalman gain matrix, K_i, and G_i is
    given by

        A_i K_i = G_i(H_i^1/2)^-1.

    ``multi_kalman_sqrt_var`` requires the input of the lower triangular
    Cholesky factors of the noise covariance matrices R_i^1/2 and,
    optionally, Q_i^1/2 and the lower triangular Cholesky factor of the
    current state covariance matrix, S_i, and returns the product of the
    matrices A_i and K_i, A_i K_i, the Cholesky factor of the updated
    state covariance matrix S_{i+1} and the matrix H_i^1/2 used in the
    computation of the likelihood for the model.

    References
    ----------
    Vanbegin, M, van Dooren, P and Verhaegen, M H G, 1989, `Algorithm
    675: FORTRAN subroutines for computing the square root covariance
    filter and square root information filter in dense or Hessenberg
    forms`, ACM Trans. Math. Software (15), 243--256

    Verhaegen, M H G and van Dooren, P, 1986, `Numerical aspects of
    different Kalman filter implementations`, IEEE Trans. Auto. Contr.
    (AC-31), 907--917
    """

    fun_name = 'naginterfaces.library.tsa.multi_kalman_sqrt_var'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_q, 'q'), (_c, 'c'), (_r, 'r'), (_s, 's')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _l_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _l = _b_utils._EngineIntScalarType(
        _l_py, {'fun_name': fun_name, 'entity': '_l'},
    ).to_ctypes()
    _k = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    _b_tsa.multi_kalman_sqrt_var(
        _n,
        _m,
        _l,
        _a,
        _b,
        stq,
        _q,
        _c,
        _r,
        _s,
        _k,
        _h,
        tol,
    )
    _s_py = _s
    _k_py = _k
    _h_py = _h
    _return_data = MultiKalmanSqrtVarReturnData(
        _s_py,
        _k_py,
        _h_py,
    )
    return _return_data

MultiKalmanSqrtInvarReturnData = _collections.namedtuple(
    'MultiKalmanSqrtInvarReturnData',
    ['a', 'b', 'c', 's', 'k', 'h', 'u'],
)

def multi_kalman_sqrt_invar(transf, a, b, stq, q, c, r, s, tol=0.0):
    u"""
    Combined measurement and time update, one iteration of Kalman
    filter, time-invariant, square root covariance filter.

    ``multi_kalman_sqrt_invar`` performs a combined measurement and time
    update of one iteration of the time-invariant Kalman filter using a
    square root covariance filter.

    For full information please refer to the NAG Library document for
    g13eb

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ebf.html

    Parameters
    ----------
    transf : str, length 1
        Indicates whether to transform the input matrix pair (A,C) to
        lower observer Hessenberg form. The transformation will only be
        required on the first call to ``multi_kalman_sqrt_invar``.

        `transf` = 'T'

            The matrices in arrays `a` and `c` are transformed to lower
            observer Hessenberg form and the matrices in `b` and `s` are
            transformed as described in [Notes].

        `transf` = 'H'

            The matrices in arrays `a`, `c` and `b` should be as
            returned from a previous call to ``multi_kalman_sqrt_invar``
            with `transf` = 'T'.

    a : float, array-like, shape (n, n)
        If `transf` = 'T', the state transition matrix, A.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

    b : float, array-like, shape (n, l)
        If `transf` = 'T', the noise coefficient matrix B.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

    stq : bool
        If `stq` = True, the state noise covariance matrix Q_i is
        assumed to be the identity matrix. Otherwise the lower
        triangular Cholesky factor, Q_i^1/2, must be provided in `q`.

    q : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`stq`): l; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if not((`stq`): l; otherwise: 0.

        If `stq` = False, `q` must contain the lower triangular Cholesky
        factor of the state noise covariance matrix, Q_i^1/2. Otherwise
        `q` is not referenced.

    c : float, array-like, shape (m, n)
        If `transf` = 'T', the measurement coefficient matrix, C.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

    r : float, array-like, shape (m, m)
        The lower triangular Cholesky factor of the measurement noise
        covariance matrix R_i^1/2.

    s : float, array-like, shape (n, n)
        If `transf` = 'T' the lower triangular Cholesky factor of the
        state covariance matrix, S_i.

        If `transf` = 'H' the lower triangular Cholesky factor of the
        covariance matrix of the transformed state vector S_i^* as
        returned from a previous call to ``multi_kalman_sqrt_invar``
        with `transf` = 'T'.

    tol : float, optional
        The tolerance used to test for the singularity of H_i^1/2. If
        0.0 <= `tol` < m^2*machine precision, then m^2*machine precision
        is used instead. The inverse of the condition number of H^(1/2)
        is estimated by a call to ``lapacklin.dtrcon``. If this estimate
        is less than `tol` then H^(1/2) is assumed to be singular.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If `transf` = 'T', the transformed matrix, U*AU^(*T), otherwise
        `a` is unchanged.

    b : float, ndarray, shape (n, l)
        If `transf` = 'T', the transformed matrix, U*B, otherwise `b` is
        unchanged.

    c : float, ndarray, shape (m, n)
        If `transf` = 'T', the transformed matrix, CU^(*T), otherwise
        `c` is unchanged.

    s : float, ndarray, shape (n, n)
        The lower triangular Cholesky factor of the transformed state
        covariance matrix, S_i+1^*.

    k : float, ndarray, shape (n, m)
        The Kalman gain matrix for the transformed state vector
        premultiplied by the state transformed transition matrix,
        U*AK_i.

    h : float, ndarray, shape (m, m)
        The lower triangular matrix H_i^1/2.

    u : float, ndarray, shape (n, :)
        If `transf` = 'T' the n*n transformation matrix U*, otherwise
        `u` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `transf` = *<value>*.

            Constraint: `transf` = 'T' or 'H'.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, l = *<value>*.

            Constraint: l >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The matrix H_i^1/2 is singular.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The Kalman filter arises from the state space model given by

        [table omitted]

    where X_i is the state vector of length n at time i, Y_i is the
    observation vector of length m at time i and W_i of length l and V_i
    of length m are the independent state noise and measurement noise
    respectively.
    The matrices A,B and C are time invariant.

    The estimate of X_i given observations Y_1 to Y_{i-1} is denoted by
    X^_{i | i-1} with state covariance matrix Var(X^_{i | i-1}) = P_{i |
    i-1} = S_i S_i^T while the estimate of X_i given observations Y_1 to
    Y_i is denoted by X^_{i | i} with covariance matrix Var(X^_{i | i})
    = P_{i | i}.
    The update of the estimate, X^_{i | i-1}, from time i to time (i+1)
    is computed in two stages.
    First, the measurement-update is given by

        X^_{i | i} = X^_{i | i-1}+K_i[Y_i-CX^_{i | i-1}]

    where K_i = P_{i | i}C^T[CP_{i | i}C^T+R_i]^-1 is the Kalman gain
    matrix.
    The second stage is the time-update for X, which is given by

        X^_{i+1 | i} = AX^_{i | i}+D_i U_i

    where D_i U_i represents any deterministic control used.

    The square root covariance filter algorithm provides a stable method
    for computing the Kalman gain matrix and the state covariance
    matrix.
    The algorithm can be summarised as

        [table omitted]U = [table omitted]

    where U is an orthogonal transformation triangularizing the
    left-hand pre-array to produce the right-hand post-array.
    The triangularization is carried out via Householder transformations
    exploiting the zero pattern of the pre-array.
    The relationship between the Kalman gain matrix K_i and G_i is given
    by

        AK_i = G_i(H_i^1/2)^-1.

    In order to exploit the invariant parts of the model to simplify the
    computation of U the results for the transformed state space U*X are
    computed where U* is the transformation that reduces the matrix pair
    (A,C) to lower observer Hessenberg form.
    That is, the matrix U* is computed such that the compound matrix

        [CU^(*T);U*AU^(*T)]

    is a lower trapezoidal matrix.
    Further the matrix B is transformed to U*B.
    These transformations need only be computed once at the start of a
    series, and ``multi_kalman_sqrt_invar`` will, optionally, compute
    them. ``multi_kalman_sqrt_invar`` returns transformed matrices
    U*AU^(*T), U*B, CU^(*T) and U*AK_i, the Cholesky factor of the
    updated transformed state covariance matrix S_i+1^* (where U*P_{i+1
    | i}U^(*T) = S_i+1^*S_i+1^*T) and the matrix H_i^1/2, valid for both
    transformed and original models, which is used in the computation of
    the likelihood for the model.
    Note that the covariance matrices Q_i and R_i can be time-varying.

    References
    ----------
    Vanbegin, M, van Dooren, P and Verhaegen, M H G, 1989, `Algorithm
    675: FORTRAN subroutines for computing the square root covariance
    filter and square root information filter in dense or Hessenberg
    forms`, ACM Trans. Math. Software (15), 243--256

    Verhaegen, M H G and van Dooren, P, 1986, `Numerical aspects of
    different Kalman filter implementations`, IEEE Trans. Auto. Contr.
    (AC-31), 907--917
    """

    fun_name = 'naginterfaces.library.tsa.multi_kalman_sqrt_invar'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_q, 'q'), (_c, 'c'), (_r, 'r'), (_s, 's')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _l_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _l = _b_utils._EngineIntScalarType(
        _l_py, {'fun_name': fun_name, 'entity': '_l'},
    ).to_ctypes()
    _k = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    _transf = _b_utils._EngineCharScalarType(
        transf, {'fun_name': fun_name, 'entity': 'transf'}, exp_chrlen=1,
    ).to_ctypes()
    _transf_py = _b_utils._EngineCharScalarType.to_py(_transf)
    if _transf_py in ('T', 't'):
        _u_dim2 = _n_py
    else:
        _u_dim2 = 0
    exp_shape_u = (_n_py, _u_dim2)
    _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_u,
        sorder=_m_sorder,
    )
    _b_tsa.multi_kalman_sqrt_invar(
        transf,
        _n,
        _m,
        _l,
        _a,
        _b,
        stq,
        _q,
        _c,
        _r,
        _s,
        _k,
        _h,
        _u,
        tol,
    )
    _a_py = _a
    _b_py = _b
    _c_py = _c
    _s_py = _s
    _k_py = _k
    _h_py = _h
    _u_py = _u
    _return_data = MultiKalmanSqrtInvarReturnData(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
        _k_py,
        _h_py,
        _u_py,
    )
    return _return_data

KalmanUnscentedStateRevcomReturnData = _collections.namedtuple(
    'KalmanUnscentedStateRevcomReturnData',
    ['irevcm'],
)

def kalman_unscented_state_revcom(irevcm, y, ly, x, st, xt, fxt, comm, lx=None, ropt=None):
    u"""
    Combined time and measurement update, one iteration of the Unscented
    Kalman Filter for a nonlinear state space model, with additive noise
    (reverse communication).

    ``kalman_unscented_state_revcom`` applies the Unscented Kalman
    Filter to a nonlinear state space model, with additive noise.

    ``kalman_unscented_state_revcom`` uses reverse communication for
    evaluating the nonlinear functionals of the state space model.

    For full information please refer to the NAG Library document for
    g13ej

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ejf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0 or 3.

        If `irevcm` = 0, it is assumed that t = 0, otherwise it is
        assumed that t != 0 and that ``kalman_unscented_state_revcom``
        has been called at least once before at an earlier time step.

        `On intermediate entry`: `irevcm` must remain unchanged.

    y : float, array-like, shape (my)
        y_t, the observed data at the current time point.

    ly : float, array-like, shape (my, my)
        L_y, such that L_y L_y^T = Sigma_y, i.e., the lower triangular
        part of a Cholesky decomposition of the observation noise
        covariance structure. Only the lower triangular part of `ly` is
        referenced.

        If Sigma_y is time dependent, the value supplied should be for
        time t.

    x : float, ndarray, shape (mx), modified in place
        `On initial entry`: x^_{t-1} the state vector for the previous
        time point.

        `On intermediate exit`: when

        `irevcm` = 1

            `x` is unchanged.

        `irevcm` = 2

            x^_t_.

        `On intermediate entry`: `x` must remain unchanged.

        `On final exit`: x^_t the updated state vector.

    st : float, ndarray, shape (mx, mx), modified in place
        `On initial entry`: S_t, such that S_{t-1}S_{t-1}^T = P_{t-1},
        i.e., the lower triangular part of a Cholesky decomposition of
        the state covariance matrix at the previous time point. Only the
        lower triangular part of `st` is referenced.

        `On intermediate exit`: when

        `irevcm` = 1

            `st` is unchanged.

        `irevcm` = 2

            S_t_, the lower triangular part of a Cholesky factorization
            of P_t_.

        `On intermediate entry`: `st` must remain unchanged.

        `On final exit`: S_t, the lower triangular part of a Cholesky
        factorization of the updated state covariance matrix.

    xt : float, ndarray, shape (mx, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `irevcm`=0: max(my, n); if `irevcm` in
        (1, 2, 3) and licomm>=10: max(my, `comm`['icomm'][9]);
        otherwise: 0.

        `On initial entry`: need not be set.

        `On intermediate exit`: X_t when `irevcm` = 1, otherwise Y_t.

        For the jth sigma point, the value for the ith parameter is held
        in `xt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,mx.

        `On intermediate entry`: `xt` must remain unchanged.

        `On final exit`: the contents of `xt` are undefined.

    fxt : float, ndarray, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `irevcm` in (1, 2): max(mx, my);
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `irevcm`=1: n+mx; if `irevcm`=2: n+my;
        otherwise: 0.

        `On initial entry`: need not be set.

        `On intermediate exit`: the contents of `fxt` are undefined.

        `On intermediate entry`: F(X_t) when `irevcm` = 1, otherwise
        H(Y_t) for the values of X_t and Y_t held in `xt`.

        For the jth sigma point the value for the ith parameter should
        be held in `fxt`[i-1,j-1], for j = 1,2,...,n.

        When `irevcm` = 1, i = 1,2,...,mx and when `irevcm` = 2, i =
        1,2,...,my.

        `On final exit`: the contents of `fxt` are undefined.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    lx : None or float, array-like, shape (:, mx), optional
        L_x, such that L_x L_x^T = Sigma_x, i.e., the lower triangular
        part of a Cholesky decomposition of the process noise covariance
        structure. Only the lower triangular part of `lx` is referenced.

        If ldlx = 0, there is no process noise (v_t = 0 for all t) and
        `lx` is not referenced.

        If Sigma_x is time dependent, the value supplied should be for
        time t.

    ropt : None or float, array-like, shape (lropt), optional
        Options. The default value will be used for `ropt`[i-1] if lropt
        < i. Setting lropt = 0 will use the default values for all
        options and `ropt` need not be set.

        `ropt`[0]

            If set to 1 then the second set of sigma points are redrawn,
            as given by equation [equation]. If set to 2 then the second
            set of sigma points are generated via augmentation, as given
            by equation [equation].

            Default is for the sigma points to be redrawn (i.e.,
            `ropt`[0] = 1)

        `ropt`[1]

            kappa_x, value of kappa used when constructing the first set
            of sigma points, X_t.

            Defaults to 3-mx.

        `ropt`[2]

            alpha_x, value of alpha used when constructing the first set
            of sigma points, X_t.

            Defaults to 1.

        `ropt`[3]

            beta_x, value of beta used when constructing the first set
            of sigma points, X_t.

            Defaults to 2.

        `ropt`[4]

            Value of kappa used when constructing the second set of
            sigma points, Y_t.

            Defaults to 3-2*mx when ldlx != 0 and the second set of
            sigma points are augmented and kappa_x otherwise.

        `ropt`[5]

            Value of alpha used when constructing the second set of
            sigma points, Y_t.

            Defaults to alpha_x.

        `ropt`[6]

            Value of beta used when constructing the second set of sigma
            points, Y_t.

            Defaults to beta_x.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: `irevcm` = 1 or 2. The value of `irevcm`
        specifies what intermediate values are returned by this function
        and what values the calling program must assign to arguments of
        ``kalman_unscented_state_revcom`` before re-entering the
        routine. Details of the output and required input are given in
        the individual argument descriptions.

        `On final exit`: `irevcm` = 3

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1, 2 or 3.

        (`errno` 21)
            On entry, mx = *<value>*.

            Constraint: mx >= 1.

        (`errno` 22)
            mx has changed between calls.

            On intermediate entry, mx = *<value>*.

            On initial entry, mx = *<value>*.

        (`errno` 31)
            On entry, my = *<value>*.

            Constraint: my >= 1.

        (`errno` 32)
            my has changed between calls.

            On intermediate entry, my = *<value>*.

            On initial entry, my = *<value>*.

        (`errno` 121)
            On entry, augmented sigma points requested, n = *<value>*
            and mx = *<value>*.

            Constraint: n >= *<value>*.

        (`errno` 122)
            On entry, redrawn sigma points requested, n = *<value>* and
            mx = *<value>*.

            Constraint: n >= *<value>*.

        (`errno` 123)
            n has changed between calls.

            On intermediate entry, n = *<value>*.

            On intermediate exit, n = *<value>*.

        (`errno` 171)
            On entry, `ropt`[0] = *<value>*.

            Constraint: `ropt`[0] = 1 or 2.

        (`errno` 172)
            On entry, `ropt`[*<value>*] = *<value>*.

            Constraint: kappa > *<value>*.

        (`errno` 173)
            On entry, `ropt`[*<value>*] = *<value>*.

            Constraint: alpha > 0.

        (`errno` 181)
            On entry, lropt = *<value>*.

            Constraint: 0 <= lropt <= 7.

        (`errno` 191)
            `comm`['icomm'] has been corrupted between calls.

        (`errno` 211)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 301)
            A weight was negative and it was not possible to downdate
            the Cholesky factorization.

        (`errno` 302)
            Unable to calculate the Kalman gain matrix.

        (`errno` 303)
            Unable to calculate the Cholesky factorization of the
            updated state covariance matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 202)
            On entry, licomm = *<value>* and lrcomm = *<value>*.

            Constraint: licomm >= 30 and lrcomm >= 30+my+mx*my+2*max(mx,
            my).

            The minimum required values for licomm and lrcomm are
            returned in `comm`['icomm'][0] and `comm`['icomm'][1]
            respectively.

    Notes
    -----
    ``kalman_unscented_state_revcom`` applies the Unscented Kalman
    Filter (UKF), as described in Julier and Uhlmann (1997b) to a
    nonlinear state space model, with additive noise, which, at time t,
    can be described by:

        [table omitted]

    where x_t represents the unobserved state vector of length m_x and
    y_t the observed measurement vector of length m_y.
    The process noise is denoted v_t, which is assumed to have mean zero
    and covariance structure Sigma_x, and the measurement noise by u_t,
    which is assumed to have mean zero and covariance structure Sigma_y.

    Unscented Kalman Filter Algorithm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Given x^_0, an initial estimate of the state and P_0 and initial
    estimate of the state covariance matrix, the UKF can be described as
    follows:

    (a) Generate a set of sigma points (see [Sigma Points]):

            X_t = [ x^_{t-1} x^_{t-1}+γ*sqrt(P_{t-1})
            x^_{t-1}-γ*sqrt(P_{t-1})]

    (#) Evaluate the known model function F:

            F_t = F(X_t)

        The function F is assumed to accept the m_x*n matrix, X_t and
        return an m_x*n matrix, F_t. The columns of both X_t and F_t
        correspond to different possible states. The notation F_{t,i} is
        used to denote the ith column of F_t, hence the result of
        applying F to the ith possible state.

    (#) Time Update:

            [table omitted]

    (#) Redraw another set of sigma points (see [Sigma Points]):

            Y_t = [x^_t^ x^_t^+γ*sqrt(P_t_) x^_t^-γ*sqrt(P_t_)]

    (#) Evaluate the known model function H:

            H_t = H(Y_t)

        The function H is assumed to accept the m_x*n matrix, Y_t and
        return an m_y*n matrix, H_t. The columns of both Y_t and H_t
        correspond to different possible states. As above H_{t,i} is
        used to denote the ith column of H_t.

    (#) Measurement Update:

            [table omitted]

    Here K_t is the Kalman gain matrix, x^_t is the estimated state
    vector at time t and P_t the corresponding covariance matrix.
    Rather than implementing the standard UKF as stated above
    ``kalman_unscented_state_revcom`` uses the square-root form
    described in the Haykin (2001).

    Sigma Points
    ~~~~~~~~~~~~
    A nonlinear state space model involves propagating a vector of
    random variables through a nonlinear system and we are interested in
    what happens to the mean and covariance matrix of those variables.
    Rather than trying to directly propagate the mean and covariance
    matrix, the UKF uses a set of carefully chosen sample points,
    referred to as sigma points, and propagates these through the system
    of interest.
    An estimate of the propagated mean and covariance matrix is then
    obtained via the weighted sample mean and covariance matrix.

    For a vector of m random variables, x, with mean mu and covariance
    matrix Sigma, the sigma points are usually constructed as:

        X_t = [mu mu+γ*sqrt(Sigma) mu-γ*sqrt(Sigma)]

    When calculating the weighted sample mean and covariance matrix two
    sets of weights are required, one used when calculating the weighted
    sample mean, denoted W^m and one used when calculating the weighted
    sample covariance matrix, denoted W^c.
    The weights and multiplier, gamma, are constructed as follows:

        [table omitted]

    where, usually L = m and alpha,beta and kappa are constants.
    The total number of sigma points, n, is given by 2*L+1.
    The constant alpha is usually set to somewhere in the range 10^-4 <=
    alpha <= 1 and for a Gaussian distribution, the optimal values of
    kappa and beta are 3-L and 2 respectively.

    Rather than redrawing another set of sigma points in (d) of the UKF
    an alternative method can be used where the sigma points used in (a)
    are augmented to take into account the process noise.
    This involves replacing equation [equation] with:

        Y_t = [X_t X_{t,1}+γ*sqrt(Sigma_x) X_{t,1}-γ*sqrt(Sigma_x)]

    Augmenting the sigma points in this manner requires setting L to 2*L
    (and hence n to 2*n-1) and recalculating the weights.
    These new values are then used for the rest of the algorithm.
    The advantage of augmenting the sigma points is that it keeps any
    odd-moments information captured by the original propagated sigma
    points, at the cost of using a larger number of points.

    References
    ----------
    Haykin, S, 2001, `Kalman Filtering and Neural Networks`, John Wiley
    and Sons

    Julier, S J, 2002, `The scaled unscented transformation`,
    Proceedings of the 2002 American Control Conference (Volume 6),
    4555--4559

    Julier, S J and Uhlmann, J K, 1997, `A consistent, debiased method
    for converting between polar and Cartesian coordinate systems`,
    Proceedings of AeroSense '97, International Society for Optics and
    Phonotonics, 110--121

    Julier, S J and Uhlmann, J K, 1997, `A new extension of the Kalman
    Filter to nonlinear systems`, International Symposium for
    Aerospace/Defense, Sensing, Simulation and Controls (Volume 3) (26)
    """

    fun_name = 'naginterfaces.library.tsa.kalman_unscented_state_revcom'

    _lx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lx'},
        lx,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mx = _b_utils._EngineIntScalarType(
        1 if lx is None else _b_utils._EngineFloat64ArrayType.get_shape(_lx, {'fun_name': fun_name, 'entity': 'lx'}, exp_rank=2, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _ly = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ly'},
        ly,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'xt'},
        xt,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(xt, {'fun_name': fun_name, 'entity': 'xt'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fxt'},
        fxt,
    )
    _ropt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ropt'},
        ropt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lropt_py = _b_utils._EngineFloat64ArrayType.get_shape(_ropt, {'fun_name': fun_name, 'entity': 'ropt'}, exp_rank=1, allow_none=True)[0]
    _lropt = _b_utils._EngineIntScalarType(
        _lropt_py, {'fun_name': fun_name, 'entity': '_lropt'},
    ).to_ctypes()
    try:
        _return_scalars = _b_tsa.kalman_unscented_state_revcom(
            irevcm,
            _mx,
            _my,
            _y,
            _lx,
            _ly,
            x,
            st,
            _n,
            xt,
            fxt,
            _ropt,
            _lropt,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _irevcm_py = _return_scalars.irevcm
    _return_data = KalmanUnscentedStateRevcomReturnData(
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _irevcm_py

KalmanUnscentedStateReturnData = _collections.namedtuple(
    'KalmanUnscentedStateReturnData',
    ['x', 'st'],
)

def kalman_unscented_state(y, lx, ly, f, h, x, st, data=None, spiked_sorder='C'):
    u"""
    Combined time and measurement update, one iteration of the Unscented
    Kalman Filter for a nonlinear state space model, with additive
    noise.

    ``kalman_unscented_state`` applies the Unscented Kalman Filter (UKF)
    to a nonlinear state space model, with additive noise.

    ``kalman_unscented_state`` uses direct communication for evaluating
    the nonlinear functionals of the state space model.

    For full information please refer to the NAG Library document for
    g13ek

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ekf.html

    Parameters
    ----------
    y : float, array-like, shape (my)
        y_t, the observed data at the current time point.

    lx : float, array-like, shape (mx, mx)
        L_x, such that L_x L_x^T = Sigma_x, i.e., the lower triangular
        part of a Cholesky decomposition of the process noise covariance
        structure. Only the lower triangular part of `lx` is referenced.

        If Sigma_x is time dependent, the value supplied should be for
        time t.

    ly : float, array-like, shape (my, my)
        L_y, such that L_y L_y^T = Sigma_y, i.e., the lower triangular
        part of a Cholesky decomposition of the observation noise
        covariance structure. Only the lower triangular part of `ly` is
        referenced.

        If Sigma_y is time dependent, the value supplied should be for
        time t.

    f : callable fxt = f(xt, data=None)
        The state function, F.

        Parameters
        ~~~~~~~~~~
        xt : float, ndarray, shape (mx, n)
            X_t, the sigma points generated in (a). For the jth sigma
            point, the value for the ith parameter is held in
            `xt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,m_x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fxt : float, array-like, shape (mx, n)
            F(X_t).

            For the jth sigma point the value for the ith parameter
            should be held in `fxt`[i-1,j-1], for j = 1,2,...,n, for i =
            1,2,...,m_x.

    h : callable hyt = h(my, yt, data=None)
        The measurement function, H as described in (e).

        Parameters
        ~~~~~~~~~~
        my : int
            m_y, the number of observed variables.

        yt : float, ndarray, shape (mx, n)
            Y_t, the sigma points generated in (d). For the jth sigma
            point, the value for the ith parameter is held in
            `yt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,m_x, where
            m_x is the number of state variables and n is the number of
            sigma points.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hyt : float, array-like, shape (`my`, n)
            H(Y_t).

            For the jth sigma point the value for the ith parameter
            should be held in `hyt`[i-1,j-1], for j = 1,2,...,n, for i =
            1,2,...,m_y.

    x : float, array-like, shape (mx)
        x^_{t-1} the state vector for the previous time point.

    st : float, array-like, shape (mx, mx)
        S_t, such that S_{t-1}S_{t-1}^T = P_{t-1}, i.e., the lower
        triangular part of a Cholesky decomposition of the state
        covariance matrix at the previous time point. Only the lower
        triangular part of `st` is referenced.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `lx` and `st` are spiked (i.e., have unit extent in all but
        one dimension, or have size 1), `spiked_sorder` selects the
        storage order to associate with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    x : float, ndarray, shape (mx)
        x^_t the updated state vector.

    st : float, ndarray, shape (mx, mx)
        S_t, the lower triangular part of a Cholesky factorization of
        the updated state covariance matrix.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, mx = *<value>*.

            Constraint: mx >= 1.

        (`errno` 21)
            On entry, my = *<value>*.

            Constraint: my >= 1.

        (`errno` 301)
            A weight was negative and it was not possible to downdate
            the Cholesky factorization.

        (`errno` 302)
            Unable to calculate the Kalman gain matrix.

        (`errno` 303)
            Unable to calculate the Cholesky factorization of the
            updated state covariance matrix.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 61)
            User requested termination in `f`.

        (`errno` 71)
            User requested termination in `h`.

    Notes
    -----
    ``kalman_unscented_state`` applies the Unscented Kalman Filter
    (UKF), as described in Julier and Uhlmann (1997b) to a nonlinear
    state space model, with additive noise, which, at time t, can be
    described by:

        [table omitted]

    where x_t represents the unobserved state vector of length m_x and
    y_t the observed measurement vector of length m_y.
    The process noise is denoted v_t, which is assumed to have mean zero
    and covariance structure Sigma_x, and the measurement noise by u_t,
    which is assumed to have mean zero and covariance structure Sigma_y.

    Unscented Kalman Filter Algorithm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Given x^_0, an initial estimate of the state and P_0 and initial
    estimate of the state covariance matrix, the UKF can be described as
    follows:

    (a) Generate a set of sigma points (see [Sigma Points]):

            X_t = [ x^_{t-1} x^_{t-1}+γ*sqrt(P_{t-1})
            x^_{t-1}-γ*sqrt(P_{t-1})]

    (#) Evaluate the known model function F:

            F_t = F(X_t)

        The function F is assumed to accept the m_x*n matrix, X_t and
        return an m_x*n matrix, F_t. The columns of both X_t and F_t
        correspond to different possible states. The notation F_{t,i} is
        used to denote the ith column of F_t, hence the result of
        applying F to the ith possible state.

    (#) Time Update:

            [table omitted]

    (#) Redraw another set of sigma points (see [Sigma Points]):

            Y_t = [x^_t^ x^_t^+γ*sqrt(P_t_) x^_t^-γ*sqrt(P_t_)]

    (#) Evaluate the known model function H:

            H_t = H(Y_t)

        The function H is assumed to accept the m_x*n matrix, Y_t and
        return an m_y*n matrix, H_t. The columns of both Y_t and H_t
        correspond to different possible states. As above H_{t,i} is
        used to denote the ith column of H_t.

    (#) Measurement Update:

            [table omitted]

    Here K_t is the Kalman gain matrix, x^_t is the estimated state
    vector at time t and P_t the corresponding covariance matrix.
    Rather than implementing the standard UKF as stated above
    ``kalman_unscented_state`` uses the square-root form described in
    the Haykin (2001).

    Sigma Points
    ~~~~~~~~~~~~
    A nonlinear state space model involves propagating a vector of
    random variables through a nonlinear system and we are interested in
    what happens to the mean and covariance matrix of those variables.
    Rather than trying to directly propagate the mean and covariance
    matrix, the UKF uses a set of carefully chosen sample points,
    referred to as sigma points, and propagates these through the system
    of interest.
    An estimate of the propagated mean and covariance matrix is then
    obtained via the weighted sample mean and covariance matrix.

    For a vector of m random variables, x, with mean mu and covariance
    matrix Sigma, the sigma points are usually constructed as:

        X_t = [mu mu+γ*sqrt(Sigma) mu-γ*sqrt(Sigma)]

    When calculating the weighted sample mean and covariance matrix two
    sets of weights are required, one used when calculating the weighted
    sample mean, denoted W^m and one used when calculating the weighted
    sample covariance matrix, denoted W^c.
    The weights and multiplier, gamma, are constructed as follows:

        [table omitted]

    where, usually L = m and alpha,beta and kappa are constants.
    The total number of sigma points, n, is given by 2*L+1.
    The constant alpha is usually set to somewhere in the range 10^-4 <=
    alpha <= 1 and for a Gaussian distribution, the optimal values of
    kappa and beta are 3-L and 2 respectively.

    The constants, kappa, alpha and beta are given by kappa = 3-m_x,
    alpha = 1.0 and beta = 2.
    If more control is required over the construction of the sigma
    points then the reverse communication function,
    ``kalman_unscented_state_revcom``, can be used instead.

    References
    ----------
    Haykin, S, 2001, `Kalman Filtering and Neural Networks`, John Wiley
    and Sons

    Julier, S J, 2002, `The scaled unscented transformation`,
    Proceedings of the 2002 American Control Conference (Volume 6),
    4555--4559

    Julier, S J and Uhlmann, J K, 1997, `A consistent, debiased method
    for converting between polar and Cartesian coordinate systems`,
    Proceedings of AeroSense '97, International Society for Optics and
    Phonotonics, 110--121

    Julier, S J and Uhlmann, J K, 1997, `A new extension of the Kalman
    Filter to nonlinear systems`, International Symposium for
    Aerospace/Defense, Sensing, Simulation and Controls (Volume 3) (26)

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.kalman_unscented_state_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.kalman_unscented_state'

    cb_f_name = fun_name + ':f'

    def cb_f(mx, n, xt, fxt, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fxt_f = f(
                xt,
                data,
            )
        else:
            _fxt_f = f(
                xt,
            )
        _fxt_f_np = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fxt'},
            _fxt_f,
            copy=False,
            exp_shape=(mx, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_f_name},
            ((_fxt_f_np, 'fxt'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fxt[:, :] = _fxt_f_np

    cb_h_name = fun_name + ':h'

    def cb_h(mx, my, n, yt, hyt, data=None):
        """Wrapping user's h. Called as base h."""
        if data is not None:
            _hyt_h = h(
                my,
                yt,
                data,
            )
        else:
            _hyt_h = h(
                my,
                yt,
            )
        _hyt_h_np = _utils._to_ndarray(
            {'fun_name': cb_h_name, 'returned': 'hyt'},
            _hyt_h,
            copy=False,
            exp_shape=(my, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_h_name},
            ((_hyt_h_np, 'hyt'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        hyt[:, :] = _hyt_h_np

    _lx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lx'},
        lx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ly = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ly'},
        ly,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _st = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_lx, 'lx'), (_ly, 'ly'), (_st, 'st')),
        spiked_sorder=spiked_sorder,
    )
    _mx_py = _b_utils._EngineFloat64ArrayType.get_shape(_lx, {'fun_name': fun_name, 'entity': 'lx'}, exp_rank=2)[0]
    _mx = _b_utils._EngineIntScalarType(
        _mx_py, {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    _b_utils._check_callable(h, {'fun_name': fun_name, 'entity': 'h'})
    _h = cb_h
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_tsa.kalman_unscented_state(
            _mx,
            _my,
            _y,
            _lx,
            _ly,
            _f,
            _h,
            _x,
            _st,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _st_py = _st
    _return_data = KalmanUnscentedStateReturnData(
        _x_py,
        _st_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

UniGarchAsym1EstimReturnData = _collections.namedtuple(
    'UniGarchAsym1EstimReturnData',
    ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
)

def uni_garch_asym1_estim(dist, yt, x, ip, iq, mn, isym, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for either a symmetric
    GARCH process or a GARCH process with asymmetry of the form
    (epsilon_{t-1}+gamma)^2.

    ``uni_garch_asym1_estim`` estimates the parameters of either a
    standard univariate regression GARCH process, or a univariate
    regression-type I AGARCH(p,q) process (see Engle and Ng (1993)).

    For full information please refer to the NAG Library document for
    g13fa

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13faf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    isym : int
        If `isym` = 1, the asymmetry term gamma will be included in the
        model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        If `isym` = 1, the next element must contain the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element must contain df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts`[1] = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    copts : bool, array-like, shape (2)
        The options to be used by ``uni_garch_asym1_estim``.

        `copts`[0] = True

            Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

            The function provides initial parameter estimates of the
            regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the coefficients beta_j, for j =
        1,2,...,p.

        If `isym` = 1, the next element contains the estimate for the
        asymmetry parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o.

        The next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        If `isym` = 1, the next element contains the standard error for
        gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the score for alpha_o.

        The next `iq` elements contain the score for alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the scores for beta_j, for j =
        1,2,...,p.

        If `isym` = 1, the next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            1+`iq`+`ip`+`isym`+`mn`+nreg, else npar =
            2+`iq`+`ip`+`isym`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, `isym` = *<value>*.

            Constraint: `isym` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    A univariate regression-type I AGARCH(p,q) process, with q
    coefficients alpha_i, for i = 1,2,...,q, p coefficients beta_i, for
    i = 1,2,...,p, and k linear regression coefficients b_i, for i =
    1,2,...,k, can be represented by:

        y_t = b_o+x_t^Tb+epsilon_t

        h_t = alpha_0+sum_i = 1^q alpha_i(epsilon_{t-i}+gamma)^2+sum_i =
        1^p beta_i h_{t-i}, t = 1,2,...,T

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of terms in the
    sequence, y_t denotes the endogenous variables, x_t the exogenous
    variables, b_o the regression mean, b the regression coefficients,
    epsilon_t the residuals, h_t the conditional variance, df the number
    of degrees of freedom of the Student's t-distribution, and psi_t the
    set of all information up to time t.

    ``uni_garch_asym1_estim`` provides an estimate for theta^, the
    parameter vector theta = (b_o,b^T,omega^T) where b^T =
    (b_1,...,b_k), omega^T =
    (alpha_0,alpha_1,...,alpha_q,beta_1,...,beta_p,gamma) when `dist` =
    'N' and omega^T =
    (alpha_0,alpha_1,...,alpha_q,beta_1,...,beta_p,gamma,df) when `dist`
    = 'T'.

    `isym`, `mn` and nreg can be used to simplify the GARCH(p,q)
    expression in (1) as follows:

    **No Regression and No Mean**

        y_t = epsilon_t,

        `isym` = 0,

        `mn` = 0,

        nreg = 0 and

        theta is a (p+q+1) vector when `dist` = 'N' and a (p+q+2) vector
        when `dist` = 'T'.

    **No Regression**

        y_t = b_o+epsilon_t,

        `isym` = 0,

        `mn` = 1,

        nreg = 0 and

        theta is a (p+q+2) vector when `dist` = 'N' and a (p+q+3) vector
        when `dist` = 'T'.

    **Note:** if the y_t = mu+epsilon_t, where mu is known (not to be
    estimated by ``uni_garch_asym1_estim``) then (1) can be written as
    y_t^mu = epsilon_t, where y_t^mu = y_t-mu.
    This corresponds to the case **No Regression and No Mean**, with y_t
    replaced by y_t-mu.

    **No Mean**

        y_t = x_t^Tb+epsilon_t,

        `isym` = 0,

        `mn` = 0,

        nreg = k and

        theta is a (p+q+k+1) vector when `dist` = 'N' and a (p+q+k+2)
        vector when `dist` = 'T'.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_garch_asym1_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym1_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _copts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'copts'},
        copts,
        copy=False,
    )
    try:
        _return_scalars = _b_tsa.uni_garch_asym1_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            isym,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            _copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = UniGarchAsym1EstimReturnData(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_asym1_forecast(nt, ip, iq, theta, gamma, ht, et):
    u"""
    Univariate time series, forecast function for either a symmetric
    GARCH process or a GARCH process with asymmetry of the form
    (epsilon_{t-1}+gamma)^2.

    ``uni_garch_asym1_forecast`` forecasts the conditional variances
    h_t, for t = T+1,...,T+xi, from a type I AGARCH(p,q) sequence, where
    xi is the forecast horizon and T is the current time (see Engle and
    Ng (1993)).

    For full information please refer to the NAG Library document for
    g13fb

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fbf.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Assume the GARCH(p,q) process can be represented by:

        h_t = alpha_0+sum_i = 1^q alpha_i(epsilon_{t-i}+gamma)^2+sum_i =
        1^p beta_i h_{t-i}, t = 1,2,...,T

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t), has been modelled by ``uni_garch_asym1_estim`` and the
    estimated conditional variances and residuals are contained in the
    arrays `ht` and `et` respectively.

    ``uni_garch_asym1_forecast`` will then use the last max(p, q)
    elements of the arrays `ht` and `et` to estimate the conditional
    variance forecasts, h_t | psi_T, where t = T+1,...,T+xi and xi is
    the forecast horizon.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_garch_asym1_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym1_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_tsa.uni_garch_asym1_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        gamma,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

UniGarchAsym2EstimReturnData = _collections.namedtuple(
    'UniGarchAsym2EstimReturnData',
    ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
)

def uni_garch_asym2_estim(dist, yt, x, ip, iq, mn, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for a GARCH process
    with asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``uni_garch_asym2_estim`` estimates the parameters of a univariate
    regression-type II AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fc

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fcf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        The next element must contain the asymmetry parameter gamma.

        If `dist` = 'T', the next element must contain df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the mean term b_o.

        If `copts`[1] = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    copts : bool, array-like, shape (2)
        The options to be used by ``uni_garch_asym2_estim``.

        `copts`[0] = True

            Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

            The function provides initial parameter estimates of the
            regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the coefficients beta_j, for j =
        1,2,...,p.

        The next element contains the estimate for the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        The next element contains the standard error for gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the score for alpha_o and the next
        `iq` elements contain the score for alpha_i, for i = 1,2,...,q.

        The next `ip` elements are the scores for beta_j, for j =
        1,2,...,p.

        The next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            2+`iq`+`ip`+`mn`+nreg, else npar = 3+`iq`+`ip`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    A univariate regression-type II AGARCH(p,q) process, with q
    coefficients alpha_i, for i = 1,2,...,q, p coefficients, beta_i, for
    i = 1,2,...,p, and k linear regression coefficients b_i, for i =
    1,2,...,k, can be represented by:

        y_t = b_o+x_t^Tb+epsilon_t

        h_t = alpha_0+sum_i = 1^q
        alpha_i(|epsilon_{t-i}|+gammaepsilon_{t-i})^2+sum_i = 1^p beta_i
        h_{t-i}, t = 1,2,...,T.

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of terms in the
    sequence, y_t denotes the endogenous variables, x_t the exogenous
    variables, b_o the regression mean, b the regression coefficients,
    epsilon_t the residuals, h_t the conditional variance, and psi_t the
    set of all information up to time t.

    ``uni_garch_asym2_estim`` provides an estimate for the parameter
    vector theta = (b_o,b^T,omega^T) where b^T = (b_1,...,b_k), omega^T
    = (alpha_0,alpha_1,...,alpha_q,beta_1,...,beta_p,gamma) when `dist`
    = 'N' and omega^T =
    (alpha_0,alpha_1,...,alpha_q,beta_1,...,beta_p,gamma,df) when `dist`
    = 'T'.

    `mn` and nreg can be used to simplify the GARCH(p,q) expression in
    (1) as follows:

    **No Regression and No Mean**

        y_t = epsilon_t,

        `mn` = 0,

        nreg = 0 and

        theta is a (p+q+2) vector when `dist` = 'N' and a (p+q+3) vector
        when `dist` = 'T'.

    **No Regression**

        y_t = b_o+epsilon_t,

        `mn` = 1,

        nreg = 0 and

        theta is a (p+q+3) vector when `dist` = 'N' and a (p+q+4)*1
        vector when `dist` = 'T'.

    **Note:** if the y_t = mu+epsilon_t, where mu is known (not to be
    estimated by ``uni_garch_asym2_estim``) then (1) can be written as
    y_t^mu = epsilon_t, where y_t^mu = y_t-mu.
    This corresponds to the case **No Regression and No Mean**, with y_t
    replaced by y_t-mu.

    **No Mean**

        y_t = x_t^Tb+epsilon_t,

        `mn` = 0,

        nreg = k and

        theta is a (p+q+k+2) vector when `dist` = 'N' and a (p+q+k+3)
        vector when `dist` = 'T'.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym2_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _copts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'copts'},
        copts,
        copy=False,
    )
    try:
        _return_scalars = _b_tsa.uni_garch_asym2_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            _copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = UniGarchAsym2EstimReturnData(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_asym2_forecast(nt, ip, iq, theta, gamma, ht, et):
    u"""
    Univariate time series, forecast function for a GARCH process with
    asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``uni_garch_asym2_forecast`` forecasts the conditional variances,
    h_t,t = T+1,...,T+xi from a type II AGARCH(p,q) sequence, where xi
    is the forecast horizon and T is the current time (see Engle and Ng
    (1993)).

    For full information please refer to the NAG Library document for
    g13fd

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fdf.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Assume the GARCH(p,q) process can be represented by:

        h_t = alpha_0+sum_i = 1^q
        alpha_i(|epsilon_{t-i}|+gammaepsilon_{t-i})^2+sum_i = 1^p beta_i
        h_{t-i}, t = 1,2,...,T.

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t), has been modelled by ``uni_garch_asym2_estim`` and the
    estimated conditional variances and residuals are contained in the
    arrays `ht` and `et` respectively.

    ``uni_garch_asym2_forecast`` will then use the last max(p, q)
    elements of the arrays `ht` and `et` to estimate the conditional
    variance forecasts, h_t | psi_T, where t = T+1,...,T+xi and xi is
    the forecast horizon.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym2_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_tsa.uni_garch_asym2_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        gamma,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

UniGarchGjrEstimReturnData = _collections.namedtuple(
    'UniGarchGjrEstimReturnData',
    ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
)

def uni_garch_gjr_estim(dist, yt, x, ip, iq, mn, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for an asymmetric
    Glosten, Jagannathan and Runkle (GJR) GARCH process.

    ``uni_garch_gjr_estim`` estimates the parameters of a univariate
    regression-GJR GARCH(p,q) process (see Glosten `et al.` (1993)).

    For full information please refer to the NAG Library document for
    g13fe

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fef.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        The next element must contain the asymmetry parameter gamma.

        If `dist` = 'T', the next element contains df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts`[1] = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    copts : bool, array-like, shape (2)
        The options to be used by ``uni_garch_gjr_estim``.

        `copts`[0] = True

            Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

            The function provides initial parameter estimates of the
            regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the moving average coefficients
        beta_j, for j = 1,2,...,p.

        The next element contains the estimate for the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        The next element contains the standard error for gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the score for alpha_o, the next `iq`
        elements contain the scores for alpha_i, for i = 1,2,...,q.

        The next `ip` elements are the score for beta_j, for j =
        1,2,...,p.

        The next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            2+`iq`+`ip`+`mn`+nreg, else npar = 3+`iq`+`ip`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    A univariate regression-GJR GARCH(p,q) process, with q coefficients
    alpha_i, for i = 1,2,...,q, p coefficients beta_i, for i =
    1,2,...,p, and k linear regression coefficients b_i, for i =
    1,2,...,k, can be represented by:

        y_t = b_o+x_t^Tb+epsilon_t

        h_t = alpha_0+sum_i = 1^q(alpha_i+gamma
        I_{t-i})epsilon_t-i^2+sum_i = 1^p beta_i h_{t-i}, t = 1,2,...,T

    where I_t = 1, if epsilon_t < 0, I_t = 0, if epsilon_t >= 0, and
    epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of terms in the
    sequence, y_t denotes the endogenous variables, x_t the exogenous
    variables, b_o the regression mean, b the regression coefficients,
    epsilon_t the residuals, h_t is the conditional variance, and psi_t
    the set of all information up to time t.

    ``uni_garch_gjr_estim`` provides an estimate for theta^, the
    parameter vector theta = (b_o,b^T,omega^T) where b^T =
    (b_1,...,b_k), omega^T =
    (alpha_0,alpha_1,...,alpha_q,beta_1,...,beta_p,gamma) when `dist` =
    'N' and omega^T =
    (alpha_0,alpha_1,...,alpha_q,beta_1,...,beta_p,gamma,df) when `dist`
    = 'T'.

    `mn`, nreg can be used to simplify the GARCH(p,q) expression in (1)
    as follows:

    **No Regression and No Mean**

        y_t = epsilon_t,

        `mn` = 0,

        nreg = 0 and

        theta is a (p+q+2) vector when `dist` = 'N', and a (p+q+3)
        vector when `dist` = 'T'.

    **No Regression**

        y_t = b_o+epsilon_t,

        `mn` = 1,

        nreg = 0 and

        theta is a (p+q+3) vector when `dist` = 'N', and a (p+q+4)
        vector when `dist` = 'T'.

    **Note:** if the y_t = mu+epsilon_t, where mu is known (not to be
    estimated by ``uni_garch_gjr_estim``) then (1) can be written as
    y_t^mu = epsilon_t, where y_t^mu = y_t-mu.
    This corresponds to the case **No Regression and No Mean**, with y_t
    replaced by y_t-mu.

    **No Mean**

        y_t = x_t^Tb+epsilon_t,

        `mn` = 0,

        nreg = k and

        theta is a (p+q+k+2) vector when `dist` = 'N', and a (p+q+k+3)
        vector when `dist` = 'T'.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Glosten, L, Jagannathan, R and Runkle, D, 1993, `Relationship
    between the expected value and the volatility of nominal excess
    return on stocks`, Journal of Finance (48), 1779--1801

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_gjr_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _copts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'copts'},
        copts,
        copy=False,
    )
    try:
        _return_scalars = _b_tsa.uni_garch_gjr_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            _copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = UniGarchGjrEstimReturnData(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_gjr_forecast(nt, ip, iq, theta, gamma, ht, et):
    u"""
    Univariate time series, forecast function for an asymmetric Glosten,
    Jagannathan and Runkle (GJR) GARCH process.

    ``uni_garch_gjr_forecast`` forecasts the conditional variances, h_t,
    for t = T+1,...,T+xi from a GJR GARCH(p,q) sequence, where xi is the
    forecast horizon and T is the current time (see Glosten `et al.`
    (1993)).

    For full information please refer to the NAG Library document for
    g13ff

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fff.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Assume the GARCH(p,q) process can be represented by:

        h_t = alpha_0+sum_i = 1^q(alpha_i+gamma
        I_{t-i})epsilon_t-i^2+sum_i = 1^p beta_i h_{t-i}, t = 1,2,...,T.

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t), and I_t = 1, if epsilon_t < 0, or I_t = 0, if epsilon_t
    >= 0, has been modelled by ``uni_garch_gjr_estim``, and the
    estimated conditional variances and residuals are contained in the
    arrays `ht` and `et` respectively.

    ``uni_garch_gjr_forecast`` will then use the last max(p, q) elements
    of the arrays `ht` and `et` to estimate the conditional variance
    forecasts, h_t | psi_T, where t = T+1,...,T+xi and xi is the
    forecast horizon.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Glosten, L, Jagannathan, R and Runkle, D, 1993, `Relationship
    between the expected value and the volatility of nominal excess
    return on stocks`, Journal of Finance (48), 1779--1801

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_gjr_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_tsa.uni_garch_gjr_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        gamma,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

UniGarchExpEstimReturnData = _collections.namedtuple(
    'UniGarchExpEstimReturnData',
    ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
)

def uni_garch_exp_estim(dist, yt, x, ip, iq, mn, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for an exponential
    GARCH (EGARCH) process.

    ``uni_garch_exp_estim`` estimates the parameters of a univariate
    regression-exponential GARCH(p,q) process (see Engle and Ng (1993)).

    For full information please refer to the NAG Library document for
    g13fg

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fgf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the autoregressive coefficients
        alpha_i, for i = 1,2,...,q.

        The next `iq` elements contain the coefficients phi_i, for i =
        1,2,...,q.

        The next `ip` elements must contain the moving average
        coefficients beta_i, for i = 1,2,...,p.

        If `dist` = 'T', the next element must contain an estimate for
        df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts` = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts` = False then `pht` is the value to be used for the
        pre-observed conditional variance, otherwise `pht` is not
        referenced.

    copts : bool
        If `copts` = True, the function provides initial parameter
        estimates of the regression terms, otherwise these are provided
        by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o and the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `iq` elements contain the coefficients phi_i, for i =
        1,2,...,q.

        The next `ip` elements are the moving average coefficients
        beta_i, for i = 1,2,...,p.

        If `dist` = 'T', the next element contains an estimate for df
        then the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `iq` elements contain the standard errors for phi_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the scores for alpha_o, the next `iq`
        elements contain the scores for alpha_i, for i = 1,2,...,q, the
        next `iq` elements contain the scores for phi_i, for i =
        1,2,...,q, the next `ip` elements are the scores for beta_j, for
        j = 1,2,...,p.

        If `dist` = 'T', the next element contains the scores for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts` = True then `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            1+2*`iq`+`ip`+`mn`+nreg, else npar =
            2+2*`iq`+`ip`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    A univariate regression-exponential GARCH(p,q) process, with q
    coefficients alpha_i, for i = 1,2,...,q, q coefficients phi_i, for i
    = 1,2,...,q, p coefficients, beta_i, for i = 1,2,...,p, and k linear
    regression coefficients b_i, for i = 1,2,...,k, can be represented
    by:

        y_t = b_o + x_t^T b + epsilon_t ; ln(h_t) = alpha_0 + sum_i =
        1^q alpha_i z_{t-i} + sum_i = 1^q phi_i (|z_{t-i}|-E[|z_{t-i}|])
        + sum_i = 1^p beta_i ln (h_{t-i}), t = 1,2,...,T

    where z_t = (epsilon_t)/sqrt(h_t), E[|z_{t-i}|] denotes the expected
    value of |z_{t-i}| and epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t
    | psi_{t-1} = S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of terms in the
    sequence, y_t denotes the endogenous variables, x_t the exogenous
    variables, b_o the regression mean, b the regression coefficients,
    epsilon_t the residuals, h_t the conditional variance, df the number
    of degrees of freedom of the Student's t-distribution, and psi_t the
    set of all information up to time t.

    ``uni_garch_exp_estim`` provides an estimate theta^, for the vector
    theta = (b_o,b^T,omega^T) where b^T = (b_1,...,b_k), omega^T =
    (alpha_0,alpha_1,...,alpha_q,phi_1,...,phi_q,beta_1,...,beta_p,gamma)
    when `dist` = 'N', and omega^T =
    (alpha_0,alpha_1,...,alpha_q,phi_1,...,phi_q,beta_1,...,beta_p,gamma,df)
    when `dist` = 'T'.

    `mn`, nreg can be used to simplify the GARCH(p,q) expression in (1)
    as follows:

    **No Regression and No Mean**

        y_t = epsilon_t,

        `mn` = 0,

        nreg = 0 and

        theta is a (2*q+p+1) vector when `dist` = 'N', and a (2*q+p+2)
        vector, when `dist` = 'T'.

    **No Regression**

        y_t = b_o+epsilon_t,

        `mn` = 1,

        nreg = 0 and

        theta is a (2*q+p+2) vector when `dist` = 'N' and a (2*q+p+3)
        vector, when `dist` = 'T'.

    **Note:** if the y_t = mu+epsilon_t, where mu is known (not to be
    estimated by ``uni_garch_exp_estim``) then (1) can be written as
    y_t^mu = epsilon_t, where y_t^mu = y_t-mu.
    This corresponds to the case **No Regression and No Mean**, with y_t
    replaced by y_t-mu.

    **No Mean**

        y_t = x_t^Tb+epsilon_t,

        `mn` = 0,

        nreg = k and

        theta is a (2*q+p+1+k) vector when `dist` = 'N' and a
        (2*q+p+2+k) vector, when `dist` = 'T'.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Glosten, L, Jagannathan, R and Runkle, D, 1993, `Relationship
    between the expected value and the volatility of nominal excess
    return on stocks`, Journal of Finance (48), 1779--1801

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_exp_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    try:
        _return_scalars = _b_tsa.uni_garch_exp_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = UniGarchExpEstimReturnData(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_exp_forecast(nt, ip, iq, theta, ht, et):
    u"""
    Univariate time series, forecast function for an exponential GARCH
    (EGARCH) process.

    ``uni_garch_exp_forecast`` forecasts the conditional variances,
    h_t,t = T+1,...,T+xi from an exponential GARCH(p,q) sequence, where
    xi is the forecast horizon and T is the current time (see Engle and
    Ng (1993)).

    For full information please refer to the NAG Library document for
    g13fh

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13fhf.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (2*`iq`+`ip`+1)
        The initial parameter estimates for the vector theta. The first
        element must contain the coefficient alpha_o and the next `iq`
        elements must contain the autoregressive coefficients alpha_i,
        for i = 1,2,...,q. The next `iq` elements must contain the
        coefficients phi_i, for i = 1,2,...,q. The next `ip` elements
        must contain the moving average coefficients beta_j, for j =
        1,2,...,p.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    Assume the GARCH(p,q) process represented by:

        ln(h_t) = alpha_0+sum_i = 1^q alpha_i z_{t-i}+sum_j = 1^q
        phi_i(|z_{t-j}|-E[|z_{t-i}|])+sum_j = 1^p beta_i ln(h_{t-j}), t
        = 1,2,...,T.

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t), and z_t = (epsilon_t)/sqrt(h_t), E[|z_{t-i}|] denotes
    the expected value of |z_{t-i}|, has been modelled by
    ``uni_garch_exp_estim``, and the estimated conditional variances and
    residuals are contained in the arrays `ht` and `et` respectively.

    ``uni_garch_exp_forecast`` will then use the last max(p, q) elements
    of the arrays `ht` and `et` to estimate the conditional variance
    forecasts, h_t | psi_T, where t = T+1,...,T+xi and xi is the
    forecast horizon.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Glosten, L, Jagannathan, R and Runkle, D, 1993, `Relationship
    between the expected value and the volatility of nominal excess
    return on stocks`, Journal of Finance (48), 1779--1801

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    fun_name = 'naginterfaces.library.tsa.uni_garch_exp_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_tsa.uni_garch_exp_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

InhomIemaReturnData = _collections.namedtuple(
    'InhomIemaReturnData',
    ['iema', 'pn'],
)

def inhom_iema(iema, t, tau, inter, sinit=None, pn=0, comm=None):
    u"""
    Computes the iterated exponential moving average for a univariate
    inhomogeneous time series.

    ``inhom_iema`` calculates the iterated exponential moving average
    for an inhomogeneous time series.

    For full information please refer to the NAG Library document for
    g13me

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13mef.html

    Parameters
    ----------
    iema : float, array-like, shape (nb)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    t : float, array-like, shape (nb)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 31 will be returned, but
        ``inhom_iema`` will continue as if t was strictly increasing by
        using the absolute value.

    tau : float
        tau, the parameter controlling the rate of decay, which must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau
        can be calculated without overflowing, for all i.

    inter : int, array-like, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

            Previous point, with nu = 1.

        `inter`[i] = 2

            Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

            Next point, nu = mu.

        Zumbach and Müller (2001) recommend that linear interpolation is
        used in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    sinit : None or float, array-like, shape (m+2), optional
        If `pn` = 0, the values used to start the iterative process,
        with

            `sinit`[0] = t_0,

            `sinit`[1] = z_0,

            `sinit`[j+1] = EMA[tau,j;z](t_0), for j = 1,2,...,m.

        If `pn` != 0, `sinit` is not referenced.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``inhom_iema``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to ``inhom_iema``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    iema : float, ndarray, shape (nb)
        The iterated EMA, with `iema`[i-1] = EMA[tau,m;z](t_i).

    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 32)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 41)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 42)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 52)
            On entry, m = *<value>*.

            On entry at previous call, m = *<value>*.

            Constraint: if `pn` > 0 then m must be unchanged since
            previous call.

        (`errno` 71)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 72)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 73)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the previous call.

        (`errno` 81)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 82)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 91)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 101)
            On entry, `pn` = 0, lrcomm = *<value>* and m = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= m+20.

        (`errno` 102)
            On entry, `pn` != 0, lrcomm = *<value>* and m = *<value>*.

            Constraint: if `pn` != 0, lrcomm >= m+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `iema` or for `tau`.

    Notes
    -----
    ``inhom_iema`` calculates the iterated exponential moving average
    for an inhomogeneous time series.
    The time series is represented by two vectors of length n; a vector
    of times, t; and a vector of values, z.
    Each element of the time series is, therefore, composed of the pair
    of scalar values (t_i,z_i), for i = 1,2,...,n.
    Time can be measured in any arbitrary units, as long as all elements
    of t use the same units.

    The exponential moving average (EMA), with parameter tau, is an
    average operator, with the exponentially decaying kernel given by

        (e^(-t_i/tau))/tau.

    The exponential form of this kernel gives rise to the following
    iterative formula for the EMA operator (see Zumbach and Müller
    (2001)):

        EMA[tau;z](t_i) =
        mu*EMA[tau;z](t_{i-1})+(nu-mu)*z_{i-1}+(1-nu)*z_i

    where

        mu = e^(-alpha) and alpha = (t_i-t_{i-1})/tau.

    The value of nu depends on the method of interpolation chosen.
    ``inhom_iema`` gives the option of three interpolation methods:

    [table omitted]

    The m-iterated exponential moving average, EMA[tau,m;z](t_i), m > 1,
    is defined using the recursive formula:

        EMA[tau,m;z] = EMA[tau;EMA[tau,m-1;z]]

    with

        EMA[tau,1;z] = EMA[tau;z].

    For large datasets or where all the data is not available at the
    same time, z and t can be split into arbitrary sized blocks and
    ``inhom_iema`` called multiple times.

    References
    ----------
    Dacorogna, M M, Gencay, R, Müller, U, Olsen, R B and Pictet, O V,
    2001, `An Introduction to High-frequency Finance`, Academic Press

    Zumbach, G O and Müller, U A, 2001, `Operators on inhomogeneous time
    series`, International Journal of Theoretical and Applied Finance
    (4(1)), 147--178

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.inhom_iema_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.inhom_iema'

    _iema = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iema'},
        iema,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_iema, {'fun_name': fun_name, 'entity': 'iema'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    if _pn_py != 0:
        _comm_needs_init = _b_utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('rcomm',),
        )
    else:
        _comm_needs_init = _b_utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('rcomm',),
            allow_none=True,
        )
    _sinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sinit'},
        sinit,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if sinit is None:
        _m = _b_utils._EngineIntScalarType(
            int(_b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)), {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
        _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    else:
        _m = _b_utils._EngineIntScalarType(
            _b_utils._EngineFloat64ArrayType.get_shape(_sinit, {'fun_name': fun_name, 'entity': 'sinit'}, exp_rank=1, allow_none=True)[0]-2, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
        _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _inter = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inter'},
        inter,
        copy=False,
    )
    try:
        _return_scalar = _b_tsa.inhom_iema(
            _nb,
            _iema,
            _t,
            tau,
            _m,
            _sinit,
            _inter,
            _pn,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _iema_py = _iema
    _pn_py = _return_scalar
    _return_data = InhomIemaReturnData(
        _iema_py,
        _pn_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InhomIemaAllReturnData = _collections.namedtuple(
    'InhomIemaAllReturnData',
    ['iema', 'p', 'pn'],
)

def inhom_iema_all(z, t, tau, m1, inter, ftype, p, sorder=1, sinit=None, x=None, pn=0, comm=None):
    u"""
    Computes the iterated exponential moving average for a univariate
    inhomogeneous time series, intermediate results are also returned.

    ``inhom_iema_all`` calculates the iterated exponential moving
    average for an inhomogeneous time series, returning the intermediate
    results.

    For full information please refer to the NAG Library document for
    g13mf

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13mff.html

    Parameters
    ----------
    z : float, array-like, shape (nb)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    t : float, array-like, shape (nb)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 61 will be returned, but
        ``inhom_iema_all`` will continue as if t was strictly increasing
        by using the absolute value.

    tau : float
        tau, the parameter controlling the rate of decay. tau must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau
        can be calculated without overflowing, for all i.

    m1 : int
        The minimum number of times the EMA operator is to be iterated.

    inter : int, array-like, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

            Previous point, with nu = 1.

        `inter`[i] = 2

            Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

            Next point, nu = mu.

        Zumbach and Müller (2001) recommend that linear interpolation is
        used in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    ftype : int
        The function type used to define the relationship between y and
        z when calculating EMA[tau,1;y]. Three functions are provided:

        `ftype` = 1

            The identity function, with y_i = z_i^[p].

        `ftype` = 2

            The absolute value, with y_i = |z_i|^p.

        `ftype` = 3

            The absolute difference, with y_i = |z_i-x_i|^p, where the
            vector x is supplied in `x`.

    p : float
        p, the power used in the transformation function.

    sorder : int, optional
        Determines the storage order of output returned in `iema`.

    sinit : None or float, array-like, shape (m2+2), optional
        If `pn` = 0, the values used to start the iterative process,
        with

            `sinit`[0] = t_0,

            `sinit`[1] = y_0,

            `sinit`[j+1] = EMA[tau,j;y](t_0), j = 1,2,...,m2.

        If `pn` != 0 then `sinit` is not referenced.

    x : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ftype`=3: nb; otherwise: 0.

        If `ftype` = 3, x_i, the vector used to shift the current block
        of observations, for i = k+1,...,k+b, where k is the number of
        observations processed so far, i.e., the value supplied in `pn`
        on entry.

        If `ftype` != 3 then `x` is not referenced.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``inhom_iema_all``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to
        ``inhom_iema_all``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    iema : float, ndarray, shape (:, :)
        The iterated exponential moving average.

        If `sorder` = 1, `iema`[i-1,j-1] = EMA[tau,j+`m1`-1;y](t_{i+k}).

        If `sorder` = 2, `iema`[j-1,i-1] = EMA[tau,j+`m1`-1;y](t_{i+k}).

        For i = 1,2,...,nb, j = 1,2,...,m2-`m1`+1 and k is the number of
        observations processed so far, i.e., the value supplied in `pn`
        on entry.

    p : float
        If `ftype` = 1, then [p], the actual power used in the
        transformation function is returned, otherwise `p` is unchanged.

    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 51)
            On entry, `sorder` = 1, ldiema = *<value>* and nb =
            *<value>*.

            Constraint: ldiema >= nb.

        (`errno` 51)
            On entry, `sorder` = 2, ldiema = *<value>* and m2-`m1`+1 =
            *<value>*.

            Constraint: ldiema >= m2-`m1`+1.

        (`errno` 62)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 71)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 72)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 81)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 82)
            On entry, `m1` = *<value>*.

            On entry at previous call, `m1` = *<value>*.

            Constraint: if `pn` > 0 then `m1` must be unchanged since
            previous call.

        (`errno` 91)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: m2 >= `m1`.

        (`errno` 92)
            On entry, m2 = *<value>*.

            On entry at previous call, m2 = *<value>*.

            Constraint: if `pn` > 0 then m2 must be unchanged since
            previous call.

        (`errno` 101)
            On entry, `ftype` != 1, j = *<value>* and `sinit`[j-1] =
            *<value>*.

            Constraint: if `ftype` != 1, `sinit`[j-1] >= 0.0, for j =
            2,3,...,m2+2.

        (`errno` 111)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 112)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 113)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the last call.

        (`errno` 121)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 1, 2 or 3.

        (`errno` 122)
            On entry, `ftype` = *<value>*, On entry at previous call,
            `ftype` = *<value>*.

            Constraint: if `pn` != 0, `ftype` must be unchanged since
            the previous call.

        (`errno` 131)
            On entry, `p` = *<value>*.

            Constraint: absolute value of `p` must be representable as
            an integer.

        (`errno` 132)
            On entry, `p` = *<value>*.

            Constraint: if `ftype` != 1, `p` != 0.0. If `ftype` = 1, the
            nearest integer to `p` must not be 0.

        (`errno` 133)
            On entry, i = *<value>*, `z`[i-1] = *<value>* and `p` =
            *<value>*.

            Constraint: if `ftype` = 1 or 2 and `z`[i] = 0 for any i
            then `p` > 0.0.

        (`errno` 134)
            On entry, i = *<value>*, `z`[i-1] = *<value>*, `x`[i-1] =
            *<value>* and `p` = *<value>*.

            Constraint: if `ftype` = 3 and `z`[i] = `x`[i] for any i
            then `p` > 0.0.

        (`errno` 135)
            On entry, `p` = *<value>*.

            On exit from previous call, `p` = *<value>*.

            Constraint: if `pn` > 0 then `p` must be unchanged since
            previous call.

        (`errno` 151)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 152)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 161)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 171)
            On entry, `pn` = 0, lrcomm = *<value>* and m2 = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= m2+20.

        (`errno` 172)
            On entry, `pn` != 0, lrcomm = *<value>* and m2 = *<value>*.

            Constraint: if `pn` != 0 then lrcomm >= m2+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `z`, `x` or for `tau`.

    Notes
    -----
    ``inhom_iema_all`` calculates the iterated exponential moving
    average for an inhomogeneous time series.
    The time series is represented by two vectors of length n: a vector
    of times, t; and a vector of values, z.
    Each element of the time series is, therefore, composed of the pair
    of scalar values (t_i,z_i), for i = 1,2,...,n.
    Time can be measured in any arbitrary units, as long as all elements
    of t use the same units.

    The exponential moving average (EMA), with parameter tau, is an
    average operator, with the exponentially decaying kernel given by

        (e^(-t_i/tau))/tau.

    The exponential form of this kernel gives rise to the following
    iterative formula (Zumbach and Müller (2001)) for the EMA operator:

        EMA[tau;y](t_i) =
        mu*EMA[tau;y](t_{i-1})+(nu-mu)*y_{i-1}+(1-nu)*y_i

    where

        mu = e^(-alpha) and alpha = (t_i-t_{i-1})/tau.

    The value of nu depends on the method of interpolation chosen and
    the relationship between y and the input series z depends on the
    transformation function chosen. ``inhom_iema_all`` gives the option
    of three interpolation methods:

    [table omitted]

    and three transformation functions:

    [table omitted]

    where the notation [p] is used to denote the integer nearest to p.
    In the case of the absolute difference x is a user-supplied vector
    of length n and, therefore, each element of the time series is
    composed of the triplet of scalar values, (t_i,z_i,x_i).

    The m-iterated exponential moving average, EMA[tau,m;y](t_i), is
    defined using the recursive formula:

        EMA[tau,m;y](t_i) = EMA[tau;EMA[tau,m-1;y](t_i)](t_i)

    with

        EMA[tau,1;y](t_i) = EMA[tau;y](t_i).

    For large datasets or where all the data is not available at the
    same time, z,t and, where required, x can be split into arbitrary
    sized blocks and ``inhom_iema_all`` called multiple times.

    References
    ----------
    Dacorogna, M M, Gencay, R, Müller, U, Olsen, R B and Pictet, O V,
    2001, `An Introduction to High-frequency Finance`, Academic Press

    Zumbach, G O and Müller, U A, 2001, `Operators on inhomogeneous time
    series`, International Journal of Theoretical and Applied Finance
    (4(1)), 147--178
    """

    fun_name = 'naginterfaces.library.tsa.inhom_iema_all'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _sinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sinit'},
        sinit,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if sinit is None:
        _m2 = 1
        _m2_py = _m2
    else:
        _m2 = _b_utils._EngineIntScalarType(
            _b_utils._EngineFloat64ArrayType.get_shape(_sinit, {'fun_name': fun_name, 'entity': 'sinit'}, exp_rank=1, allow_none=True)[0]-2, {'fun_name': fun_name, 'entity': '_m2'},
        ).to_ctypes()
        _m2_py = _b_utils._EngineIntScalarType.to_py(_m2)
    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _m1 = _b_utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m1_py = _b_utils._EngineIntScalarType.to_py(_m1)
    if _sorder_py == 1:
        _iema_dim1 = _nb_py
    else:
        _iema_dim1 = _m2_py-_m1_py+1
    if _sorder_py == 1:
        _iema_dim2 = _m2_py-_m1_py+1
    elif _sorder_py == 2:
        _iema_dim2 = _nb_py
    else:
        _iema_dim2 = 0
    exp_shape_iema = (_iema_dim1, _iema_dim2)
    _iema = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_iema,
        sorder=_b_utils._EngineData.default_sorder,
    )
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inter = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inter'},
        inter,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    try:
        _return_scalars = _b_tsa.inhom_iema_all(
            _sorder,
            _nb,
            _z,
            _iema,
            _t,
            tau,
            _m1,
            _m2,
            _sinit,
            _inter,
            ftype,
            p,
            _x,
            _pn,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _p_py = _return_scalars.p
    _pn_py = _return_scalars.pn
    _iema_py = _iema
    _return_data = InhomIemaAllReturnData(
        _iema_py,
        _p_py,
        _pn_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InhomMaReturnData = _collections.namedtuple(
    'InhomMaReturnData',
    ['ma', 'p', 'pn', 'wma'],
)

def inhom_ma(ma, t, tau, m1, m2, inter, ftype, p, sinit=None, pn=0, comm=None):
    u"""
    Computes the exponential moving average for a univariate
    inhomogeneous time series.

    ``inhom_ma`` provides a moving average, moving norm, moving variance
    and moving standard deviation operator for an inhomogeneous time
    series.

    For full information please refer to the NAG Library document for
    g13mg

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13mgf.html

    Parameters
    ----------
    ma : float, array-like, shape (nb)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    t : float, array-like, shape (nb)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 31 will be returned, but
        ``inhom_ma`` will continue as if t was strictly increasing by
        using the absolute value.

        The lagged difference, t_i-t_{i-1} must be sufficiently small
        that e^(-alpha), alpha = (t_i-t_{i-1})/tau~ can be calculated
        without overflowing, for all i.

    tau : float
        tau, the parameter controlling the rate of decay. tau must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau~
        can be calculated without overflowing, for all i, where tau~ =
        2*tau/(m_2+m_1).

    m1 : int
        m_1, the iteration of the EMA operator at which the sum is
        started.

    m2 : int
        m_2, the iteration of the EMA operator at which the sum is
        ended.

    inter : int, array-like, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

            Previous point, with nu = 1.

        `inter`[i] = 2

            Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

            Next point, nu = mu.

        Zumbach and Müller (2001) recommend that linear interpolation is
        used in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    ftype : int
        The function type used to define the relationship between y and
        z when calculating EMA[tau,1;y]. Three functions are provided:

        `ftype` = 1

            The identity function, with y_i = z_i^[p].

        `ftype` = 2 or 4

            The absolute value, with y_i = |z_i|^p.

        `ftype` = 3 or 5

            The absolute difference, with y_i =
            |z_i-MA[tau,m;y](t_i)|^p.

        If `ftype` = 4 or 5 then the resulting vector of averages is
        scaled by p^-1 as described in `ma`.

    p : float
        p, the power used in the transformation function.

    sinit : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ftype` in (3, 5): 2*`m2`+3; otherwise: `m2`+2.

        If `pn` = 0, the values used to start the iterative process,
        with

            `sinit`[0] = t_0,

            `sinit`[1] = y_0,

            `sinit`[j+1] = EMA[tau,j;y](t_0), for i = 1,2,...,`m2`.

        In addition, if `ftype` = 3 or 5 then

            `sinit`[`m2`+2] = z_0,

            `sinit`[`m2`+j+1] = EMA[tau,j;z](t_0), for j = 1,2,...,`m2`.

        i.e., initial values based on the original data z as opposed to
        the transformed data y.

        If `pn` != 0, `sinit` is not referenced.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``inhom_ma``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to ``inhom_ma``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    ma : float, ndarray, shape (nb)
        The moving average:

        if `ftype` = 4 or 5

            `ma`[i-1] = {MA[tau,m_1,m_2;y](t_i)}^(1/p),

        otherwise

            `ma`[i-1] = MA[tau,m_1,m_2;y](t_i).

    p : float
        If `ftype` = 1, then [p], the actual power used in the
        transformation function is returned, otherwise `p` is unchanged.

    pn : int
        k+b, the updated number of observations processed so far.

    wma : float, ndarray, shape (nb)
        Either the moving average or exponential moving average,
        depending on the value of `ftype`.

        if `ftype` = 3 or 5

            `wma`[i-1] = MA[tau;y](t_i)

        otherwise

            `wma`[i-1] = EMA[tau~;y](t_i).

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 32)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 41)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 42)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 52)
            On entry, `m1` = *<value>*.

            On entry at previous call, `m1` = *<value>*.

            Constraint: if `pn` > 0 then `m1` must be unchanged since
            previous call.

        (`errno` 61)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m2` >= `m1`.

        (`errno` 62)
            On entry, `m2` = *<value>*.

            On entry at previous call, `m2` = *<value>*.

            Constraint: if `pn` > 0 then `m2` must be unchanged since
            previous call.

        (`errno` 71)
            On entry, `ftype` != 1, j = *<value>* and `sinit`[j-1] =
            *<value>*.

            Constraint: if `ftype` != 1, `sinit`[j-1] >= 0.0, for j =
            2,3,...,`m2`+2.

        (`errno` 81)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 82)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 83)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the last call.

        (`errno` 91)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 1, 2, 3, 4 or 5.

        (`errno` 92)
            On entry, `ftype` = *<value>*, On entry at previous call,
            `ftype` = *<value>*.

            Constraint: if `pn` != 0, `ftype` must be unchanged since
            the previous call.

        (`errno` 101)
            On entry, `p` = *<value>*.

            Constraint: absolute value of `p` must be representable as
            an integer.

        (`errno` 102)
            On entry, `p` = *<value>*.

            Constraint: if `ftype` != 1, `p` != 0.0. If `ftype` = 1, the
            nearest integer to `p` must not be 0.

        (`errno` 103)
            On entry, i = *<value>*, `ma`[i-1] = *<value>* and `p` =
            *<value>*.

            Constraint: if `ftype` = 1, 2 or 4 and `ma`[i-1] = 0 for any
            i then `p` > 0.0.

        (`errno` 104)
            On entry, i = *<value>*, `ma`[i-1] = *<value>*, `wma`[i-1] =
            *<value>* and `p` = *<value>*.

            Constraint: if `p` < 0.0, `ma`[i-1]-`wma`[i-1] != 0.0, for
            any i.

        (`errno` 105)
            On entry, `p` = *<value>*.

            On exit from previous call, `p` = *<value>*.

            Constraint: if `pn` > 0 then `p` must be unchanged since
            previous call.

        (`errno` 111)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 112)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 131)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 141)
            On entry, `pn` = 0, lrcomm = *<value>* and `m2` = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= 2*`m2`+20.

        (`errno` 142)
            On entry, `pn` != 0, lrcomm = *<value>* and `m2` =
            *<value>*.

            Constraint: if `pn` != 0, lrcomm >= 2*`m2`+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `ma` or for `tau`.

    Notes
    -----
    ``inhom_ma`` provides a number of operators for an inhomogeneous
    time series.
    The time series is represented by two vectors of length n; a vector
    of times, t; and a vector of values, z.
    Each element of the time series is, therefore, composed of the pair
    of scalar values (t_i,z_i), for i = 1,2,...,n.
    Time t can be measured in any arbitrary units, as long as all
    elements of t use the same units.

    The main operator available, the moving average (MA), with parameter
    tau is defined as

        MA[tau,m_1,m_2;y](t_i) = 1/(m_2-m_1+1)sum_j =
        m_1^m_2EMA[tau~,j;y](t_i)

    where tau~ = 2*tau/(m_2+m_1), m_1 and m_2 are user-supplied integers
    controlling the amount of lag and smoothing respectively, with m_2
    >= m_1 and EMA[·] is the iterated exponential moving average
    operator.

    The iterated exponential moving average, EMA[tau~,m;y](t_i), is
    defined using the recursive formula:

        EMA[tau~,m;y](t_i) = EMA[tau~;EMA[tau~,m-1;y](t_i)](t_i)

    with

        EMA[tau~,1;y](t_i) = EMA[tau~;y](t_i)

    and

        EMA[tau~;y](t_i) =
        mu*EMA[tau~;y](t_{i-1})+(nu-mu)*y_{i-1}+(1-nu)*y_i

    where

        mu = e^(-alpha) and alpha = (t_i-t_{i-1})/(tau~).

    The value of nu depends on the method of interpolation chosen and
    the relationship between y and the input series z depends on the
    transformation function chosen. ``inhom_ma`` gives the option of
    three interpolation methods:

    [table omitted]

    and three transformation functions:

    [table omitted]

    where the notation [p] is used to denote the integer nearest to p.
    In addition, if either the absolute value or absolute difference
    transformation are used then the resulting moving average can be
    scaled by p^-1.

    The various parameter options allow a number of different operators
    to be applied by ``inhom_ma``, a few of which are:

    (i) **Moving Average (MA)**, as defined in [equation] (obtained by
        setting `ftype` = 1 and `p` = 1).

    (#) **Moving Norm (MNorm)**, defined as

            MNorm(tau,m,p;z) = MA[tau,1,m;|z|^p]^(1/p)

        (obtained by setting `ftype` = 4, `m1` = 1 and `m2` = m).

    (#) **Moving Variance (MVar)**, defined as

            MVar(tau,m,p;z) = MA[tau,1,m;|z-MA[tau,1,m;z]|^p]

        (obtained by setting `ftype` = 3, `m1` = 1 and `m2` = m).

    (#) **Moving Standard Deviation (MSD)**, defined as

            MSD(tau,m,p;z) = MA[tau,1,m;|z-MA[tau,1,m;z]|^p]^(1/p)

        (obtained by setting `ftype` = 5, `m1` = 1 and `m2` = m).

    For large datasets or where all the data is not available at the
    same time, z and t can be split into arbitrary sized blocks and
    ``inhom_ma`` called multiple times.

    References
    ----------
    Dacorogna, M M, Gencay, R, Müller, U, Olsen, R B and Pictet, O V,
    2001, `An Introduction to High-frequency Finance`, Academic Press

    Zumbach, G O and Müller, U A, 2001, `Operators on inhomogeneous time
    series`, International Journal of Theoretical and Applied Finance
    (4(1)), 147--178
    """

    fun_name = 'naginterfaces.library.tsa.inhom_ma'

    _ma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ma'},
        ma,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_ma, {'fun_name': fun_name, 'entity': 'ma'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sinit'},
        sinit,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inter = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inter'},
        inter,
        copy=False,
    )
    _wma = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nb_py,),
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    try:
        _return_scalars = _b_tsa.inhom_ma(
            _nb,
            _ma,
            _t,
            tau,
            m1,
            m2,
            _sinit,
            _inter,
            ftype,
            p,
            _pn,
            _wma,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ma_py = _ma
    _p_py = _return_scalars.p
    _pn_py = _return_scalars.pn
    _wma_py = _wma
    _return_data = InhomMaReturnData(
        _ma_py,
        _p_py,
        _pn_py,
        _wma_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

CpPeltReturnData = _collections.namedtuple(
    'CpPeltReturnData',
    ['ntau', 'tau', 'sparam'],
)

def cp_pelt(ctype, y, beta=None, minss=2, param=None):
    u"""
    Change point detection, using the PELT algorithm.

    ``cp_pelt`` detects change points in a univariate time series, that
    is, the time points at which some feature of the data, for example
    the mean, changes.
    Change points are detected using the PELT (Pruned Exact Linear Time)
    algorithm using one of a provided set of cost functions.

    For full information please refer to the NAG Library document for
    g13na

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13naf.html

    Parameters
    ----------
    ctype : int
        A flag indicating the assumed distribution of the data and the
        type of change point being looked for.

        `ctype` = 1

            Data from a Normal distribution, looking for changes in the
            mean, mu.

        `ctype` = 2

            Data from a Normal distribution, looking for changes in the
            standard deviation sigma.

        `ctype` = 3

            Data from a Normal distribution, looking for changes in the
            mean, mu and standard deviation sigma.

        `ctype` = 4

            Data from a Gamma distribution, looking for changes in the
            scale parameter b.

        `ctype` = 5

            Data from an exponential distribution, looking for changes
            in lambda.

        `ctype` = 6

            Data from a Poisson distribution, looking for changes in
            lambda.

    y : float, array-like, shape (n)
        y, the time series.

        If `ctype` = 6, that is the data is assumed to come from a
        Poisson distribution, floor(y+0.5) is used in all calculations.

    beta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if `ctype` = 3: 2*log(`n`);
        otherwise: log(`n`).

        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

            beta = p*log(n)

        AIC

            beta = 2*p

        Hannan-Quinn

            beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        This is usually set to 2 when `ctype` = 3 and 1 otherwise.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    param : None or float, array-like, shape (1), optional
        phi, values for the parameters that will be treated as fixed. If
        `ctype` = 4 then `param` must be supplied.

        If supplied, then when

        `ctype` = 1

            `param`[0] = sigma, the standard deviation of the normal
            distribution. If not supplied then sigma is estimated from
            the full input data,

        `ctype` = 2

            `param`[0] = mu, the mean of the normal distribution. If not
            supplied then mu is estimated from the full input data,

        `ctype` = 4

            `param`[0] must hold the shape, a, for the gamma
            distribution,

        otherwise

            `param` is not referenced.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    sparam : float, ndarray, shape (:) or (:, :)
        The estimated values of the distribution parameters in each
        segment

        `ctype` = 1, 2 or 3

            `sparam`[0,i-1] = mu_i and `sparam`[1,i-1] = sigma_i for i =
            1,2,...,m, where mu_i and sigma_i is the mean and standard
            deviation, respectively, of the values of y in the ith
            segment.

            It should be noted that sigma_i = sigma_j when `ctype` = 1
            and mu_i = mu_j when `ctype` = 2, for all i and j.

        `ctype` = 4

            `sparam`[0,i-1] = a_i and `sparam`[1,i-1] = b_i for i =
            1,2,...,m, where a_i and b_i are the shape and scale
            parameters, respectively, for the values of y in the ith
            segment. It should be noted that a_i = `param`[0] for all i.

        `ctype` = 5 or 6

            `sparam`[i-1] = lambda_i for i = 1,2,...,m, where lambda_i
            is the mean of the values of y in the ith segment.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ctype` = *<value>*.

            Constraint: `ctype` = 1, 2, 3, 4, 5 or 6.

        (`errno` 21)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 31)
            On entry, `ctype` = *<value>* and `y`[*<value>*] =
            *<value>*.

            Constraint: if `ctype` = 4, 5 or 6 then `y`[i-1] >= 0.0, for
            i = 1,2,...,n.

        (`errno` 32)
            On entry, `y`[*<value>*] = *<value>*, is too large.

        (`errno` 51)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 71)
            On entry, `ctype` = *<value>* and `param`[0] = *<value>*.

            Constraint: if `ctype` = 1 or 4 and `param` has been
            supplied, then `param`[0] > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 200)
            To avoid overflow some truncation occurred when calculating
            the cost function, C. All output is returned as normal.

        (`errno` 201)
            To avoid overflow some truncation occurred when calculating
            the parameter estimates returned in `sparam`. All output is
            returned as normal.

    Notes
    -----
    Let y_{1:n} = {y_j:j = 1,2,...,n} denote a series of data and tau =
    {tau_i:i = 1,2,...,m} denote a set of m ordered (strictly monotonic
    increasing) indices known as change points with 1 <= tau_i <= n and
    tau_m = n.
    For ease of notation we also define tau_0 = 0.
    The m change points, tau, split the data into m segments, with the
    ith segment being of length n_i and containing
    y_{tau_{i-1}+1:tau_i}.

    Given a cost function, C(y_{tau_{i-1}+1:tau_i}) ``cp_pelt`` solves

        minimize_{m,tau}sum_i = 1^m(C(y_{tau_{i-1}+1:tau_i})+beta)

    where beta is a penalty term used to control the number of change
    points.
    This minimization is performed using the PELT algorithm of Killick
    `et al.` (2012).
    The PELT algorithm is guaranteed to return the optimal solution to
    [equation] if there exists a constant K such that

        C(y_{(u+1):v})+C(y_{(v+1):w})+K <= C(y_{(u+1):w})

    for all u < v < w.

    ``cp_pelt`` supplies four families of cost function.
    Each cost function assumes that the series, y, comes from some
    distribution, D(Theta).
    The parameter space, Theta = {theta,phi} is subdivided into theta
    containing those parameters allowed to differ in each segment and
    phi those parameters treated as constant across all segments.
    All four cost functions can then be described in terms of the
    likelihood function, L and are given by:

        C(y_{(tau_{i-1}+1):tau_i}) =
        -2*log(L(theta^_i,phi|y_{(tau_{i-1}+1):tau_i}))

    where theta^_i is the maximum likelihood estimate of theta within
    the ith segment.
    In all four cases setting K = 0 satisfies equation [equation].
    Four distributions are available: Normal, Gamma, Exponential and
    Poisson.
    Letting

        S_i = sum_j = tau_{i-1}^tau_iy_j

    the log-likelihoods and cost functions for the four distributions,
    and the available subdivisions of the parameter space are:

        **Normal distribution:** Theta = {mu,sigma^2}

                -2*log(L) = sum_i = 1^msum_j =
                tau_{i-1}^tau_ilog(2*pi)+log(sigma_i^2)+((y_j-mu_i)^2)/(sigma_i^2)

            `Mean changes:` theta = {mu}

                    C(y_{tau_{i-1}+1:tau_i}) = sum_j =
                    tau_{i-1}^tau_i((y_j-n_i^-1*S_i)^2)/(sigma^2)

            `Variance changes:` theta = {sigma^2}

                    C(y_{tau_{i-1}+1:tau_i}) = n_i*(log(sum_j =
                    tau_{i-1}^tau_i(y_j-mu)^2)-log(n_i))

            `Both mean and variance change:` theta = {mu,sigma^2}

                    C(y_{tau_{i-1}+1:tau_i}) = n_i*(log(sum_j =
                    tau_{i-1}^tau_i(y_j-n_i^-1*S_i)^2)-log(n_i))

        **Gamma distribution:** Theta = {a,b}

                -2*log(L) = 2*sum_i = 1^msum_j =
                tau_{i-1}^tau_ilog(Gamma(a_i))+a_i*log(b_i)+(1-a_i)*log(y_j)+(y_j)/(b_i)

            `Scale changes:` theta = {b}

                    C(y_{tau_{i-1}+1:tau_i}) =
                    2*a*n_i(log(S_i)-log(a*n_i))

        **Exponential Distribution:** Theta = {lambda}

                -2*log(L) = 2*sum_i = 1^msum_j =
                tau_{i-1}^tau_ilog(lambda_i)+(y_j)/(lambda_i)

            `Mean changes:` theta = {lambda}

                    C(y_{tau_{i-1}+1:tau_i}) = 2*n_i(log(S_i)-log(n_i))

        **Poisson distribution:** Theta = {lambda}

                -2*log(L) = 2*sum_i = 1^msum_j = tau_{i-1}^tau_i
                lambda_i-floor(y_j+0.5)*log(lambda_i)+log(Gamma(floor(y_j+0.5)+1))

            `Mean changes:` theta = {lambda}

                    C(y_{tau_{i-1}+1:tau_i}) = 2*S_i*(log(n_i)-log(S_i))

            when calculating S_i for the Poisson distribution, the sum
            is calculated for floor(y_i+0.5) rather than y_i.

    References
    ----------
    Chen, J and Gupta, A K, 2010, `Parametric Statistical Change Point
    Analysis With Applications to Genetics`, Medicine and Finance
    (Second Edition), Birkhäuser

    Killick, R, Fearnhead, P and Eckely, I A, 2012, `Optimal detection
    of changepoints with a linear computational cost`, Journal of the
    American Statistical Association (107:500), 1590--1598

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.cp_pelt_ex.main`
    """

    fun_name = 'naginterfaces.library.tsa.cp_pelt'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ctype = _b_utils._EngineIntScalarType(
        ctype, {'fun_name': fun_name, 'entity': 'ctype'},
    ).to_ctypes()
    _ctype_py = _b_utils._EngineIntScalarType.to_py(_ctype)
    if _ctype_py == 3:
        _beta = 2*log(_n_py)
    else:
        _beta = log(_n_py)
    _beta = _beta if beta is None else beta
    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _sparam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py+2,),
    )
    try:
        _return_scalar = _b_tsa.cp_pelt(
            _ctype,
            _n,
            _y,
            _beta,
            minss,
            _param,
            _tau,
            _sparam,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    if _ctype_py in (5, 6):
        _sparam_dim1 = _ntau_py
    else:
        _sparam_dim1 = 2
    if _ctype_py in (5, 6):
        _sparam_dim2 = None
    else:
        _sparam_dim2 = _ntau_py
    if _sparam_dim2 is None:
        _sparam_py = _sparam[:_sparam_dim1]
    else:
        _sparam_py = _ascontiguousarray(_sparam[:_sparam_dim1*_sparam_dim2].reshape((_sparam_dim1, _sparam_dim2), order='F'))
    _return_data = CpPeltReturnData(
        _ntau_py,
        _tau_py,
        _sparam_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

CpPeltUserReturnData = _collections.namedtuple(
    'CpPeltUserReturnData',
    ['ntau', 'tau'],
)

def cp_pelt_user(n, beta, k, costfn, minss=2, data=None):
    u"""
    Change points detection using the PELT algorithm, user supplied cost
    function.

    ``cp_pelt_user`` detects change points in a univariate time series,
    that is, the time points at which some feature of the data, for
    example the mean, changes.
    Change points are detected using the PELT (Pruned Exact Linear Time)
    algorithm and a user-supplied cost function.

    For full information please refer to the NAG Library document for
    g13nb

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13nbf.html

    Parameters
    ----------
    n : int
        n, the length of the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

            beta = p*log(n)

        AIC

            beta = 2*p

        Hannan-Quinn

            beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        The value of p will depend on the cost function being used.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    k : float
        K, the constant value that satisfies equation [equation]. If K
        exists, it is unlikely to be unique in such cases, it is
        recommened that the largest value of K, that satisfies equation
        [equation], is chosen. No check is made that K is the correct
        value for the supplied cost function.

    costfn : callable c = costfn(ts, r, data=None)
        The cost function, C. `costfn` must calculate a vector of costs
        for a number of segments.

        Parameters
        ~~~~~~~~~~
        ts : int
            A reference time point.

        r : int, ndarray, shape (nr)
            Time points which, along with `ts`, define the segments
            being considered, 0 <= `r`[i-1] <= n for i = 1,2,...nr.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (nr)
            On exit, the cost function, C.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 51)
            User requested termination.

    Notes
    -----
    Let y_{1:n} = {y_j:j = 1,2,...,n} denote a series of data and tau =
    {tau_i:i = 1,2,...,m} denote a set of m ordered (strictly monotonic
    increasing) indices known as change points with 1 <= tau_i <= n and
    tau_m = n.
    For ease of notation we also define tau_0 = 0.
    The m change points, tau, split the data into m segments, with the
    ith segment being of length n_i and containing
    y_{tau_{i-1}+1:tau_i}.

    Given a user-supplied cost function, C(y_{tau_{i-1}+1:tau_i})
    ``cp_pelt_user`` solves

        minimize_{m,tau}sum_i = 1^m(C(y_{tau_{i-1}+1:tau_i})+beta)

    where beta is a penalty term used to control the number of change
    points.
    This minimization is performed using the PELT algorithm of Killick
    `et al.` (2012).
    The PELT algorithm is guaranteed to return the optimal solution to
    [equation] if there exists a constant K such that

        C(y_{(u+1):v})+C(y_{(v+1):w})+K <= C(y_{(u+1):w})

    for all u < v < w

    References
    ----------
    Chen, J and Gupta, A K, 2010, `Parametric Statistical Change Point
    Analysis With Applications to Genetics`, Medicine and Finance
    (Second Edition), Birkhäuser

    Killick, R, Fearnhead, P and Eckely, I A, 2012, `Optimal detection
    of changepoints with a linear computational cost`, Journal of the
    American Statistical Association (107:500), 1590--1598
    """

    fun_name = 'naginterfaces.library.tsa.cp_pelt_user'

    cb_costfn_name = fun_name + ':costfn'

    def cb_costfn(ts, nr, r, c, data=None):
        """Wrapping user's costfn. Called as base costfn."""
        if data is not None:
            _c_costfn = costfn(
                ts,
                r,
                data,
            )
        else:
            _c_costfn = costfn(
                ts,
                r,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_costfn_name, 'returned': 'c'},
            _c_costfn,
            copy=False,
            exp_shape=(nr,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(costfn, {'fun_name': fun_name, 'entity': 'costfn'})
    _costfn = cb_costfn
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_tsa.cp_pelt_user(
            _n,
            beta,
            minss,
            k,
            _costfn,
            _tau,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    _return_data = CpPeltUserReturnData(
        _ntau_py,
        _tau_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

CpBinaryReturnData = _collections.namedtuple(
    'CpBinaryReturnData',
    ['ntau', 'tau', 'sparam'],
)

def cp_binary(ctype, y, beta=None, minss=2, param=None, mdepth=0):
    u"""
    Change point detection, using binary segmentation.

    ``cp_binary`` detects change points in a univariate time series,
    that is, the time points at which some feature of the data, for
    example the mean, changes.
    Change points are detected using binary segmentation using one of a
    provided set of cost functions.

    For full information please refer to the NAG Library document for
    g13nd

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13ndf.html

    Parameters
    ----------
    ctype : int
        A flag indicating the assumed distribution of the data and the
        type of change point being looked for.

        `ctype` = 1

            Data from a Normal distribution, looking for changes in the
            mean, mu.

        `ctype` = 2

            Data from a Normal distribution, looking for changes in the
            standard deviation sigma.

        `ctype` = 3

            Data from a Normal distribution, looking for changes in the
            mean, mu and standard deviation sigma.

        `ctype` = 4

            Data from a Gamma distribution, looking for changes in the
            scale parameter b.

        `ctype` = 5

            Data from an exponential distribution, looking for changes
            in lambda.

        `ctype` = 6

            Data from a Poisson distribution, looking for changes in
            lambda.

    y : float, array-like, shape (n)
        y, the time series.

        If `ctype` = 6, that is the data is assumed to come from a
        Poisson distribution, floor(y+0.5) is used in all calculations.

    beta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if `ctype` = 3: 2*log(`n`);
        otherwise: log(`n`).

        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

            beta = p*log(n)

        AIC

            beta = 2*p

        Hannan-Quinn

            beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        This is usually set to 2 when `ctype` = 3 and 1 otherwise.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    param : None or float, array-like, shape (1), optional
        phi, values for the parameters that will be treated as fixed. If
        `ctype` = 4 then `param` must be supplied.

        If supplied, then when

        `ctype` = 1

            `param`[0] = sigma, the standard deviation of the normal
            distribution. If not supplied then sigma is estimated from
            the full input data,

        `ctype` = 2

            `param`[0] = mu, the mean of the normal distribution. If not
            supplied then mu is estimated from the full input data,

        `ctype` = 4

            `param`[0] must hold the shape, a, for the gamma
            distribution,

        otherwise

            `param` is not referenced.

    mdepth : int, optional
        K, the maximum depth for the iterative process, which in turn
        puts an upper limit on the number of change points with m <=
        2^K.

        If K <= 0 then no limit is put on the depth of the iterative
        process and no upper limit is put on the number of change
        points, other than that inherent in the length of the series and
        the value of `minss`.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    sparam : float, ndarray, shape (:) or (:, :)
        The estimated values of the distribution parameters in each
        segment

        `ctype` = 1, 2 or 3

            `sparam`[0,i-1] = mu_i and `sparam`[1,i-1] = sigma_i for i =
            1,2,...,m, where mu_i and sigma_i is the mean and standard
            deviation, respectively, of the values of y in the ith
            segment.

            It should be noted that sigma_i = sigma_j when `ctype` = 1
            and mu_i = mu_j when `ctype` = 2, for all i and j.

        `ctype` = 4

            `sparam`[0,i-1] = a_i and `sparam`[1,i-1] = b_i for i =
            1,2,...,m, where a_i and b_i are the shape and scale
            parameters, respectively, for the values of y in the ith
            segment. It should be noted that a_i = `param`[0] for all i.

        `ctype` = 5 or 6

            `sparam`[i-1] = lambda_i for i = 1,2,...,m, where lambda_i
            is the mean of the values of y in the ith segment.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ctype` = *<value>*.

            Constraint: `ctype` = 1, 2, 3, 4, 5 or 6.

        (`errno` 21)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 31)
            On entry, `ctype` = *<value>* and `y`[*<value>*] =
            *<value>*.

            Constraint: if `ctype` = 4, 5 or 6 then `y`[i-1] >= 0.0, for
            i = 1,2,...,n.

        (`errno` 32)
            On entry, `y`[*<value>*] = *<value>*, is too large.

        (`errno` 51)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 71)
            On entry, `ctype` = *<value>* and `param`[0] = *<value>*.

            Constraint: if `ctype` = 1 or 4 and `param` has been
            supplied, then `param`[0] > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 200)
            To avoid overflow some truncation occurred when calculating
            the cost function, C. All output is returned as normal.

        (`errno` 201)
            To avoid overflow some truncation occurred when calculating
            the parameter estimates returned in `sparam`. All output is
            returned as normal.

    Notes
    -----
    Let y_{1:n} = {y_j:j = 1,2,...,n} denote a series of data and tau =
    {tau_i:i = 1,2,...,m} denote a set of m ordered (strictly monotonic
    increasing) indices known as change points, with 1 <= tau_i <= n and
    tau_m = n.
    For ease of notation we also define tau_0 = 0.
    The m change points, tau, split the data into m segments, with the
    ith segment being of length n_i and containing
    y_{tau_{i-1}+1:tau_i}.

    Given a cost function, C(y_{tau_{i-1}+1:tau_i}), ``cp_binary`` gives
    an approximate solution to

        minimize_{m,tau}sum_i = 1^m(C(y_{tau_{i-1}+1:tau_i})+beta)

    where beta is a penalty term used to control the number of change
    points.
    The solution is obtained in an iterative manner as follows:

    (1) Set u = 1, w = n and k = 0

    (#) Set k = k+1. If k > K, where K is a user-supplied control
        parameter, then terminate the process for this segment.

    (#) Find v that minimizes

            C(y_{u:v})+C(y_{v+1:w})

    (#) Test

            C(y_{u:v})+C(y_{v+1:w})+beta < C(y_{u:w})

    (#) If inequality [equation] is false then the process is terminated
        for this segment.

    (#) If inequality [equation] is true, then v is added to the set of
        change points, and the segment is split into two subsegments,
        y_{u:v} and y_{v+1:w}. The whole process is repeated from step 2
        independently on each subsegment, with the relevant changes to
        the definition of u and w (i.e., w is set to v when processing
        the left-hand subsegment and u is set to v+1 when processing the
        right-hand subsegment.

    The change points are ordered to give tau.

    ``cp_binary`` supplies four families of cost function.
    Each cost function assumes that the series, y, comes from some
    distribution, D(Theta).
    The parameter space, Theta = {theta,phi} is subdivided into theta
    containing those parameters allowed to differ in each segment and
    phi those parameters treated as constant across all segments.
    All four cost functions can then be described in terms of the
    likelihood function, L and are given by:

        C(y_{(tau_{i-1}+1):tau_i}) =
        -2*log(L(theta^_i,phi|y_{(tau_{i-1}+1):tau_i}))

    where the theta^_i is the maximum likelihood estimate of theta
    within the ith segment.
    Four distributions are available; Normal, Gamma, Exponential and
    Poisson distributions.
    Letting

        S_i = sum_j = tau_{i-1}^tau_iy_j

    the log-likelihoods and cost functions for the four distributions,
    and the available subdivisions of the parameter space are:

        **Normal distribution:** Theta = {mu,sigma^2}

                -2*log(L) = sum_i = 1^msum_j =
                tau_{i-1}^tau_ilog(2*pi)+log(sigma_i^2)+((y_j-mu_i)^2)/(sigma_i^2)

            `Mean changes:` theta = {mu}

                    C(y_{tau_{i-1}+1:tau_i}) = sum_j =
                    tau_{i-1}^tau_i((y_j-n_i^-1*S_i)^2)/(sigma^2)

            `Variance changes:` theta = {sigma^2}

                    C(y_{tau_{i-1}+1:tau_i}) = n_i*(log(sum_j =
                    tau_{i-1}^tau_i(y_j-mu)^2)-log(n_i))

            `Both mean and variance change:` theta = {mu,sigma^2}

                    C(y_{tau_{i-1}+1:tau_i}) = n_i*(log(sum_j =
                    tau_{i-1}^tau_i(y_j-n_i^-1*S_i)^2)-log(n_i))

        **Gamma distribution:** Theta = {a,b}

                -2*log(L) = 2*sum_i = 1^msum_j =
                tau_{i-1}^tau_ilog(Gamma(a_i))+a_i*log(b_i)+(1-a_i)*log(y_j)+(y_j)/(b_i)

            `Scale changes:` theta = {b}

                    C(y_{tau_{i-1}+1:tau_i}) =
                    2*a*n_i(log(S_i)-log(a*n_i))

        **Exponential Distribution:** Theta = {lambda}

                -2*log(L) = 2*sum_i = 1^msum_j =
                tau_{i-1}^tau_ilog(lambda_i)+(y_j)/(lambda_i)

            `Mean changes:` theta = {lambda}

                    C(y_{tau_{i-1}+1:tau_i}) = 2*n_i(log(S_i)-log(n_i))

        **Poisson distribution:** Theta = {lambda}

                -2*log(L) = 2*sum_i = 1^msum_j = tau_{i-1}^tau_i
                lambda_i-floor(y_j+0.5)*log(lambda_i)+log(Gamma(floor(y_j+0.5)+1))

            `Mean changes:` theta = {lambda}

                    C(y_{tau_{i-1}+1:tau_i}) = 2*S_i*(log(n_i)-log(S_i))

            when calculating S_i for the Poisson distribution, the sum
            is calculated for floor(y_i+0.5) rather than y_i.

    References
    ----------
    Chen, J and Gupta, A K, 2010, `Parametric Statistical Change Point
    Analysis With Applications to Genetics`, Medicine and Finance
    (Second Edition), Birkhäuser

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    fun_name = 'naginterfaces.library.tsa.cp_binary'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ctype = _b_utils._EngineIntScalarType(
        ctype, {'fun_name': fun_name, 'entity': 'ctype'},
    ).to_ctypes()
    _ctype_py = _b_utils._EngineIntScalarType.to_py(_ctype)
    if _ctype_py == 3:
        _beta = 2*log(_n_py)
    else:
        _beta = log(_n_py)
    _beta = _beta if beta is None else beta
    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mdepth = _b_utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _mdepth_py = _b_utils._EngineIntScalarType.to_py(_mdepth)
    _minss = _b_utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    _minss_py = _b_utils._EngineIntScalarType.to_py(_minss)
    if _mdepth_py > 0:
        _tau_dim1 = min(int(ceil(float(_n_py)/float(_minss_py))), 2**_mdepth_py)
    else:
        _tau_dim1 = int(ceil(float(_n_py)/float(_minss_py)))
    exp_shape_tau = (_tau_dim1,)
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        exp_shape_tau,
    )
    _sparam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    try:
        _return_scalar = _b_tsa.cp_binary(
            _ctype,
            _n,
            _y,
            _beta,
            _minss,
            _param,
            _mdepth,
            _tau,
            _sparam,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    if _ctype_py in (5, 6):
        _sparam_dim1 = _ntau_py
    else:
        _sparam_dim1 = 2
    if _ctype_py in (5, 6):
        _sparam_dim2 = None
    else:
        _sparam_dim2 = _ntau_py
    if _sparam_dim2 is None:
        _sparam_py = _sparam[:_sparam_dim1]
    else:
        _sparam_py = _ascontiguousarray(_sparam[:_sparam_dim1*_sparam_dim2].reshape((_sparam_dim1, _sparam_dim2), order='F'))
    _return_data = CpBinaryReturnData(
        _ntau_py,
        _tau_py,
        _sparam_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

CpBinaryUserReturnData = _collections.namedtuple(
    'CpBinaryUserReturnData',
    ['ntau', 'tau'],
)

def cp_binary_user(n, beta, chgpfn, minss=2, mdepth=0, data=None):
    u"""
    Change point detection, using binary segmentation, user supplied
    cost function.

    ``cp_binary_user`` detects change points in a univariate time
    series, that is, the time points at which some feature of the data,
    for example the mean, changes.
    Change points are detected using binary segmentation for a
    user-supplied cost function.

    For full information please refer to the NAG Library document for
    g13ne

    https://www.nag.com/numeric/nl/nagdoc_29.3/flhtml/g13/g13nef.html

    Parameters
    ----------
    n : int
        n, the length of the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

            beta = p*log(n).

        AIC

            beta = 2*p.

        Hannan-Quinn

            beta = 2*p*log(log(n)).

        where p is the number of parameters being treated as estimated
        in each segment.

        The value of p will depend on the cost function being used.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    chgpfn : callable (v, cost, info) = chgpfn(side, u, w, minss, info,
        data=None)
        `chgpfn` must calculate a proposed change point, and the
        associated costs, within a specified segment.

        Parameters
        ~~~~~~~~~~
        side : int
            Flag indicating what `chgpfn` must calculate and at which
            point of the Binary Segmentation it has been called.

            `side` = -1

                only C(y_{u:w}) need be calculated and returned in
                `cost`[0], neither `v` nor the other elements of `cost`
                need be set. In this case, u = 1 and w = n.

            `side` = 0

                all elements of `cost` and `v` must be set. In this
                case, u = 1 and w = n.

            `side` = 1

                the segment, y_{u:w}, is a left-hand side subsegment
                from a previous iteration of the Binary Segmentation
                algorithm. All elements of `cost` and `v` must be set.

            `side` = 2

                the segment, y_{u:w}, is a right-hand side subsegment
                from a previous iteration of the Binary Segmentation
                algorithm. All elements of `cost` and `v` must be set.

            The distinction between `side` = 1 or 2 may allow for
            `chgpfn` to be implemented in a more efficient manner.

            The first call to `chgpfn` will always have `side` = -1 and
            the second call will always have `side` = 0.

            All subsequent calls will be made with `side` = 1 or 2.

        u : int
            u, the start of the segment of interest.

        w : int
            w, the end of the segment of interest.

        minss : int
            The minimum distance between two change points, as passed to
            ``cp_binary_user``.

        info : int
            `info` = 0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        v : int
            If `side` = -1 then `v` need not be set.

            if `side` != -1 then v, the proposed change point.

            That is, the value which minimizes

                minimize_v C(y_{u:v})+C(y_{v+1:w})

            for v = u+`minss`-1 to w-`minss`.

        cost : float, array-like, shape (3)
            Costs associated with the proposed change point, v.

            If `side` = -1 then `cost`[0] = C(y_{u:w}) and the remaining
            two elements of `cost` need not be set.

            If `side` != -1 then

                `cost`[0] = C(y_{u:v})+C(y_{v+1:w}).

                `cost`[1] = C(y_{u:v}).

                `cost`[2] = C(y_{v+1:w}).

        info : int
            In most circumstances `info` should remain unchanged.

            If `info` is set to a strictly positive value then
            ``cp_binary_user`` terminates with `errno` = 51.

            If `info` is set to a strictly negative value the current
            segment is skipped (i.e., no change points are considered in
            this segment) and ``cp_binary_user`` continues as normal.

            If `info` was set to a strictly negative value at any point
            and no other errors occur then ``cp_binary_user`` will
            terminate with `errno` = 52.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    mdepth : int, optional
        K, the maximum depth for the iterative process, which in turn
        puts an upper limit on the number of change points with m <=
        2^K.

        If K <= 0 then no limit is put on the depth of the iterative
        process and no upper limit is put on the number of change
        points, other than that inherent in the length of the series and
        the value of `minss`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 51)
            User requested termination by setting `info` = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 52)
            User requested a segment to be skipped by setting `info` =
            *<value>*.

    Notes
    -----
    Let y_{1:n} = {y_j:j = 1,2,...,n} denote a series of data and tau =
    {tau_i:i = 1,2,...,m} denote a set of m ordered (strictly monotonic
    increasing) indices known as change points with 1 <= tau_i <= n and
    tau_m = n.
    For ease of notation we also define tau_0 = 0.
    The m change points, tau, split the data into m segments, with the
    ith segment being of length n_i and containing
    y_{tau_{i-1}+1:tau_i}.

    Given a cost function, C(y_{tau_{i-1}+1:tau_i}), ``cp_binary_user``
    gives an approximate solution to

        minimize_{m,tau}sum_i = 1^m(C(y_{tau_{i-1}+1:tau_i})+beta)

    where beta is a penalty term used to control the number of change
    points.
    The solution is obtained in an iterative manner as follows:

    (1) Set u = 1, w = n and k = 0

    (#) Set k = k+1. If k > K, where K is a user-supplied control
        parameter, then terminate the process for this segment.

    (#) Find v that minimizes

            C(y_{u:v})+C(y_{v+1:w})

    (#) Test

            C(y_{u:v})+C(y_{v+1:w})+beta < C(y_{u:w})

    (#) If inequality [equation] is false then the process is terminated
        for this segment.

    (#) If inequality [equation] is true, then v is added to the set of
        change points, and the segment is split into two subsegments,
        y_{u:v} and y_{v+1:w}. The whole process is repeated from step 2
        independently on each subsegment, with the relevant changes to
        the definition of u and w (i.e., w is set to v when processing
        the left-hand subsegment and u is set to v+1 when processing the
        right-hand subsegment.

    The change points are ordered to give tau.

    References
    ----------
    Chen, J and Gupta, A K, 2010, `Parametric Statistical Change Point
    Analysis With Applications to Genetics`, Medicine and Finance
    (Second Edition), Birkhäuser
    """

    fun_name = 'naginterfaces.library.tsa.cp_binary_user'

    cb_chgpfn_name = fun_name + ':chgpfn'

    def cb_chgpfn(side, u, w, minss, cost, info, data=None):
        """Wrapping user's chgpfn. Called as base chgpfn."""
        if data is not None:
            _v_chgpfn, _cost_chgpfn, _info_chgpfn = chgpfn(
                side,
                u,
                w,
                minss,
                info,
                data,
            )
        else:
            _v_chgpfn, _cost_chgpfn, _info_chgpfn = chgpfn(
                side,
                u,
                w,
                minss,
                info,
            )
        cost[:] = _utils._to_ndarray(
            {'fun_name': cb_chgpfn_name, 'returned': 'cost'},
            _cost_chgpfn,
            copy=False,
            exp_shape=(3,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _info_chgpfn_base = _info_chgpfn
        _v_chgpfn_base = _v_chgpfn
        return (
            _v_chgpfn_base,
            _info_chgpfn_base,
        )

    _b_utils._check_callable(chgpfn, {'fun_name': fun_name, 'entity': 'chgpfn'})
    _chgpfn = cb_chgpfn
    _mdepth = _b_utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _mdepth_py = _b_utils._EngineIntScalarType.to_py(_mdepth)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _minss = _b_utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    _minss_py = _b_utils._EngineIntScalarType.to_py(_minss)
    if _mdepth_py > 0:
        _tau_dim1 = min(int(ceil(float(_n_py)/float(_minss_py))), 2**_mdepth_py)
    else:
        _tau_dim1 = int(ceil(float(_n_py)/float(_minss_py)))
    exp_shape_tau = (_tau_dim1,)
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        exp_shape_tau,
    )
    try:
        _return_scalar = _b_tsa.cp_binary_user(
            _n,
            beta,
            _minss,
            _mdepth,
            _chgpfn,
            _tau,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    _return_data = CpBinaryUserReturnData(
        _ntau_py,
        _tau_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data
